(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file coercion.ml contains all the coercion functions for values and opertors 
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Boolean ;;
open Integer ;;
open Float ;;
open Fixed ;;


let checkArrayDims a1 a2 =
	if ((Array.length a1)=(Array.length a2)) then
		if ((Array.length a1.(0))=(Array.length a2.(0))) then
			()
		else
			raise (Error ("checkArrayDims","??? Matrix dimensions must agree (columns).")) 
	else
		raise (Error ("checkArrayDims","??? Matrix dimensions must agree (lines).")) ;;

 
let castInt2Int32 i = match i with
	I32(j) -> i
|	I64(j) -> I32(Int64.to_int32 j)
| 	_ -> i ;;


let castInt2Int64 i = match i with
	I64(j) -> i
|	I32(j) -> I64(Int64.of_int32 j)
| 	_ -> i ;;


let castInt2Float32 i = match i with
	I32(j) -> M(FR.from_float_prec float32_mantissa GMP_RNDN (Int32.to_float j))
|	I64(j) -> M(FR.from_float_prec float32_mantissa GMP_RNDN (Int64.to_float j)) 
|	IPInfty -> MPInfty
|	IMInfty	->	MMInfty 
| 	IBottom -> MBottom ;;


let castInt2Float64 i = match i with
	I32(j) -> M(FR.from_float_prec float64_mantissa GMP_RNDN (Int32.to_float j))
|	I64(j) -> M(FR.from_float_prec float64_mantissa GMP_RNDN (Int64.to_float j)) 
|	IPInfty -> MPInfty
|	IMInfty	->	MMInfty 
|	IBottom -> MBottom ;;


let castInt2Float128 i = match i with
	I32(j) -> M(FR.from_float_prec float128_mantissa GMP_RNDN (Int32.to_float j))
|	I64(j) -> M(FR.from_float_prec float128_mantissa GMP_RNDN (Int64.to_float j)) 
|	IPInfty -> MPInfty
|	IMInfty	->	MMInfty 
|	IBottom -> MBottom ;;


let castInt2MP i prec = match i with
	I32(j) -> M(FR.from_float_prec prec GMP_RNDN (Int32.to_float j))
|	I64(j) -> M(FR.from_float_prec prec GMP_RNDN (Int64.to_float j)) 
|	IPInfty -> MPInfty
|	IMInfty	->	MMInfty 
|	IBottom -> MBottom ;;


let castInt2Fixed i ip fp = match i with
	I32(j) -> let (r,rl) = binStringOfInt64 (Int64.of_int32 j) 
				  in if (rl>ip) then XPInfty
				 	 else X((Int32.compare j Int32.zero)>=0,r,String.make fp '0')
|	I64(j) -> let (r,rl) = binStringOfInt64 j 
			  in if (rl>ip) then XPInfty
			 	 else X((Int64.compare j Int64.zero)>=0,r,String.make fp '0') 
|	IPInfty -> XPInfty
|	IMInfty	->	XMInfty 
|	IBottom -> XBottom ;;


let castF2X x p = match x with
	M(a) -> let s = FR.to_string_base_digits GMP_RNDN 10 p a in
			let ((f,ip,fp),e) = fixedOfString s
			in ((f,ip,fp),e) 
|   MPInfty -> ((XPInfty,1,1),!eZero)
|   MMInfty -> ((XMInfty,1,1),!eZero)
|   MBottom -> ((XBottom,1,1),!eZero) ;;  


let castFE2FX x ip fp = match x with
	FE((M(f1),M(f2)),e,p) -> 	let f1' = FR.to_string_base_digits GMP_RNDN 10 p f1 in
				 				let f2' = FR.to_string_base_digits GMP_RNDN 10 p f2 in
								let ((x1',ip1,fp1),e1') = fixedOfString f1' in
								let ((x2',ip2,fp2),e2') = fixedOfString f2' in
								let r =	FX((x1',x2'),max ip1 ip2,max fp1 fp2,mpfiPlus e (mpfiJoin e1' e2') !errPrec OUTSIDE) 
							  	in resizeFX r ip fp
|  FE((a,b),e,p) ->
	let ((a',ip1,fp1),_) = castF2X a p in
	let ((b',ip2,fp2),_) = castF2X b p in 
		FX((a',b'),max ip1 ip2,max fp1 fp2,e) ;;  
					

let castIntInterval2FE32 (i,j) = FE((castInt2Float32 i,castInt2Float32 j),!eZero,float32_mantissa) ;;
let castIntInterval2FE64 (i,j) = FE((castInt2Float64 i,castInt2Float64 j),!eZero,float64_mantissa) ;;
let castIntInterval2FE128 (i,j) = FE((castInt2Float128 i,castInt2Float128 j),!eZero,float128_mantissa) ;;


let castIntInterval2MP (i,j) prec = FE((castInt2MP i prec,castInt2MP j prec),!eZero,prec) ;;

 
let castIntInterval2FX (i,j) iPrec fPrec = 
	let a = castInt2Fixed i iPrec fPrec in
	let b = castInt2Fixed j iPrec fPrec in
	let iPrec' = (match (a,b) with
					(X(_,i1,_),X(_,i2,_)) -> (max (String.length i1) (String.length i2))
				  | _ -> iPrec
			     )
	in FX((a,b),iPrec',fPrec,!eZero) ;;


let castFE2FEPrec fe prec = match fe with
	FE((M(m1),M(m2)),e,_) -> 
		let m1' = mpFloatMult (M(m1)) (M(FR.from_float_prec prec GMP_RNDN 1.0)) prec GMP_RNDD in
		let m2' = mpFloatMult (M(m2)) (M(FR.from_float_prec prec GMP_RNDN 1.0)) prec GMP_RNDU in
		let e' = mpfiPlus (!fErrZero) e !errPrec OUTSIDE in
			FE((m1',m2'),mpfiPlus e' (mpFloatMinus m1' (M(m1)) !errPrec GMP_RNDD,
			                          mpFloatMinus m2' (M(m2)) !errPrec GMP_RNDU) !errPrec OUTSIDE,prec)
| _ -> !feTop ;;

 
let rec cast value resType = match resType with
  	Type_Int_32 -> (match value with
						IVal(v1,v2) -> IVal(castInt2Int32 v1,castInt2Int32 v2)	
					| 	FVal(FE(f,_,_)) -> IVal(mpfi2I32 f) 
					|   XVal(_) -> let r = cast value Type_Float_64 
								   in cast r resType
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast00","this cast is not supported."))
					|	_ -> raise (Error ("cast01","this cast is not supported."))
					)
| 	Type_Int_64 -> (match value with
						IVal(v1,v2) -> IVal(castInt2Int64 v1,castInt2Int64 v2)	
					| 	FVal(FE(f,_,_)) -> IVal(mpfi2I64 f)
					|   XVal(_) -> let r = cast value Type_Float_64 
								   in cast r resType
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast02","this cast is not supported."))
					|	_ -> raise (Error ("cast03","this cast is not supported."))
					)
|	Type_Float_32 -> (match value with
						IVal(i) -> FVal(castIntInterval2FE32 i)	
					| 	FVal(f) -> FVal(castFE2FEPrec f float32_mantissa)
					|   XVal(_) -> raise (Error ("cast04","casts from fixed to float not supported."))
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast05","this cast is not supported."))
					|	_ -> raise (Error ("cast06","this cast is not supported."))
					)
|	Type_Float_64 -> (match value with
						IVal(i) -> FVal(castIntInterval2FE64 i)	
					| 	FVal(f) -> FVal(castFE2FEPrec f float64_mantissa)
					|   XVal(_) -> raise (Error ("cast07","casts from fixed to float not supported."))
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast08","this cast is not supported."))
					|	_ -> raise (Error ("cast09","this cast is not supported."))
					)
|	Type_Float_128 -> (match value with
						IVal(i) -> FVal(castIntInterval2FE128 i)	
					| 	FVal(f) -> FVal(castFE2FEPrec f float128_mantissa)
					|   XVal(_) -> raise (Error ("cast10","casts from fixed to float not supported."))
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast11","this cast is not supported."))
					|	_ -> raise (Error ("cast12","this cast is not supported."))
					)
|	Type_Fixed(ip,fp) -> (match value with
						IVal(i) -> XVal(castIntInterval2FX i ip fp)	
					| 	FVal(f) -> XVal(castFE2FX f ip fp) 
					|   XVal(x) -> XVal(resizeFX x ip fp)
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast13","this cast is not supported."))
					|	_ -> raise (Error ("cast14","this cast is not supported."))
					)
|	Type_MP(p) ->   (match value with
				 			IVal(i) -> FVal(castIntInterval2MP i p)	
					  | 	FVal(f) -> FVal(castFE2FEPrec f p)
					  |     XVal(_) -> raise (Error ("cast15","casts from mp to fixed not supported."))
					  |     AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
					  					cast a.(0).(0) resType
								     else
										raise (Error ("cast16","this cast is not supported."))
					  |	    _ -> raise (Error ("cast17","this cast is not supported."))
					)
| 	Type_Array(v,u,t) -> 	let (x,y) = i2PairOfInt (u,v) in
							let a = Array.make_matrix y x ValBottom in
							let _ = for i=0 to (y-1) do
										for j=0 to (x-1) do
											a.(i).(j) <- cast value t
										done
									done
							in AVal(a) 
|	Type_Bool -> (match value with
						IVal(a,b) ->	
							let b1 = (iEq (a,b) !i64Zero) in
							let b2 = (iLt (a,b) !i64Zero) in
							let b3 = (iGt (a,b) !i64Zero) in
								(match b1 with
									B(true) -> BVal(B(false))
								| _ -> (match (b2,b3) with
											(B(bb1),B(bb2)) -> if (bb1 || bb2) then BVal(B(true)) else BVal(BTop)
										|	_ -> BVal(BTop)
										)
								)
					| 	FVal(f) -> (* control follows the float, not the error *)
							if (feEq f !fe128Zero) then BVal(B(false)) 
							else
								(match f with
									FE((a,b),_,p) -> if ((mpFloatGt a (M(FR.from_float_prec float128_mantissa GMP_RNDN 0.0))) || 
														 (mpFloatLt b (M(FR.from_float_prec float128_mantissa GMP_RNDN 0.0)))
														 ) then BVal(B(true))
													 else BVal(BTop)
								)
					| 	XVal(FX((X(s1,i1,f1),X(s2,i2,f2)),_,_,_)) -> (* control follows the fixed number not the error *)
							if (((int_of_string i1)=0)&&((int_of_string f1)=0)&&
							    ((int_of_string i2)=0)&&((int_of_string f2)=0)) then BVal(B(false)) 
							else
								if (s1 || (not s2)) then 
									BVal(B(true))
							 	else BVal(BTop)
					|	BVal(_) -> value
					|   AVal(a) -> if (((Array.length a)=1)&&((Array.length a.(0))=1)) then
										cast a.(0).(0) resType
								   else
										raise (Error ("cast18","this cast is not supported."))
					|	_ -> raise (Error ("cast19","this cast is not supported."))
					)
|	Type_Fun(_)	->  raise (Error ("cast20","??? Cannot cast to function."))
| 	UndefinedType ->  value
;;


(*****************************************************************************)
(*                 Coercions in arithmetic operations                        *)
(*****************************************************************************)

let rec valPlus v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) -> 	IVal(iPlus i1 i2)
|	(FVal(f1),FVal(f2)) -> 	FVal(fePlus f1 f2 (max (getFEPrec v1) (getFEPrec v2)))
|	(IVal(i1),FVal(f2)) -> 	let f1 = getFE (cast (IVal(i1)) (typeOfPrec (getFEPrec v2))) in
								FVal(fePlus f1 f2 (getFEPrec v2))
|	(FVal(f1),IVal(i2)) -> 	let f2 = getFE (cast (IVal(i2)) (typeOfPrec (getFEPrec v1))) in
								FVal(fePlus f1 f2 (getFEPrec v1))
|	(XVal(f1),XVal(f2)) -> 	XVal(fiPlus f1 f2)
| 	(XVal(f1),IVal(i2)) -> 	let f2 = castIntInterval2FX i2 99 (getFPrec f1) in
							let f2' = if ((getIPrec f2)=99) then 
										castIntInterval2FX i2 (getIPrec f1) (getFPrec f1) 
								      else f2 
							in XVal(fiPlus f1 f2')
| 	(IVal(i1),XVal(f2)) -> 	let f1 = castIntInterval2FX i1 99 (getFPrec f2)	in
							let f1' = if ((getIPrec f1)=99) then 
										castIntInterval2FX i1 (getIPrec f2) (getFPrec f2) 
								      else f1 
							in XVal(fiPlus f1' f2)
|   (XVal(f1),FVal(f2)) ->  let f2' = castFE2FX f2 (getIPrec f1) (getFPrec f1) 
							in XVal(fiPlus f1 f2')
|   (FVal(f1),XVal(f2)) ->  let f1' = castFE2FX f1 (getIPrec f2) (getFPrec f2) 
							in XVal(fiPlus f1' f2)
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valPlus a1.(i).(j) a2.(i).(j) 
									done
								done
							in AVal(b) 
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valPlus v1 a.(i).(j)
									done
								done
							in AVal(b) 							
|	(AVal(a),_) -> 			let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valPlus a.(i).(j) v2
									done
								done
							in AVal(b) 							
|	(BVal(_),_) -> 			raise (Error ("valPlus","??? Type mismatch in addition (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valPlus","??? Type mismatch in addition (boolean not expected)."))
|	(ValTop,_) -> 			ValTop
|	(_,ValTop) ->			ValTop
|	(ValBottom,_) ->		ValBottom
|	(_,ValBottom) ->		ValBottom
|   (ArrayLines(_),_) -> 	raise (Error ("valPlus","??? ArrayLines not expected in valPlus."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valPlus","??? ArrayLines not expected in valPlus."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valPlus","??? ArrayColumns not expected in valPlus."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valPlus","??? ArrayColumns not expected in valPlus."))
;;


let rec valMinus v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) -> IVal(iMinus i1 i2)
|	(FVal(f1),FVal(f2)) -> FVal(feMinus f1 f2 (max (getFEPrec v1) (getFEPrec v2)))
|	(IVal(i1),FVal(f2)) -> let f1 = getFE (cast (IVal(i1)) (typeOfPrec (getFEPrec v2))) in
							FVal(feMinus f1 f2 (getFEPrec v2))
|	(FVal(f1),IVal(i2)) -> let f2 = getFE (cast (IVal(i2)) (typeOfPrec (getFEPrec v1))) in
							FVal(feMinus f1 f2 (getFEPrec v1))
|	(XVal(f1),XVal(f2)) -> 	XVal(fiMinus f1 f2)
| 	(XVal(f1),IVal(i2)) -> 	let f2 = castIntInterval2FX i2 99 (getFPrec f1) in
							let f2' = if ((getIPrec f2)=99) then 
										castIntInterval2FX i2 (getIPrec f1) (getFPrec f1) 
								      else f2 
							in XVal(fiPlus f1 f2')
| 	(IVal(i1),XVal(f2)) -> 	let f1 = castIntInterval2FX i1 99 (getFPrec f2) in
							let f1' = if ((getIPrec f1)=99) then 
										castIntInterval2FX i1 (getIPrec f2) (getFPrec f2) 
								      else f1 
							in XVal(fiPlus f1' f2)
|   (XVal(f1),FVal(f2)) ->  let f2' = castFE2FX f2 (getIPrec f1) (getFPrec f1) 
							in XVal(fiMinus f1 f2')
|   (FVal(f1),XVal(f2)) ->  let f1' = castFE2FX f1 (getIPrec f2) (getFPrec f2) 
							in XVal(fiMinus f1' f2)
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMinus a1.(i).(j) a2.(i).(j) 
									done
								done
							in AVal(b) 
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMinus v1 a.(i).(j)
									done
								done
							in AVal(b) 							
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMinus a.(i).(j) v2
									done
								done
							in AVal(b) 							
|	(BVal(_),_) -> 			raise (Error ("valMinus","??? Type mismatch in subtraction (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valMinus","??? Type mismatch in subtraction (boolean not expected)."))
|	(ValTop,_) -> 			ValTop
|	(_,ValTop) ->			ValTop
|	(ValBottom,_) ->		ValBottom
|	(_,ValBottom) ->		ValBottom
|   (ArrayLines(_),_) -> 	raise (Error ("valMinus","??? ArrayLines not expected in valMinus."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valMinus","??? ArrayLines not expected in valMinus."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valMinus","??? ArrayColumns not expected in valMins."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valMinus","??? ArrayColumns not expected in valMinus."))
;;


let rec valPower v (m,n) = match v with
	IVal(i) ->	 		IVal(iPowInterval i (m,n))
|	FVal(f) ->	 		FVal(fePowInterval f (m,n))
|	XVal(FX(f,ip,fp,e)) ->	XVal(fiPowInterval (FX(f,ip,fp,e)) ip fp (m,n))
|	AVal(a) ->	 		let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valPower a.(i).(j) (m,n) 
									done
								done
							in AVal(b) 
|	BVal(_) -> 			raise (Error ("valUMinus","??? Type mismatch in power (boolean not expected)."))
|	ValTop -> 			ValTop
|	ValBottom ->		ValBottom
|   ArrayLines(_) ->	raise (Error ("valPower","??? ArrayLines not expected in valPower."))
|   ArrayColumns(_) -> 	raise (Error ("valPower","??? ArrayColumns not expected in valPower."))

;;


let rec valMult v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) -> IVal(iMult i1 i2)
|	(FVal(f1),FVal(f2)) -> FVal(feMult f1 f2 (max (getFEPrec v1) (getFEPrec v2)))
|	(IVal(i1),FVal(f2)) -> let f1 = getFE (cast (IVal(i1)) (typeOfPrec (getFEPrec v2))) in
							FVal(feMult f1 f2 (getFEPrec v2))
|	(FVal(f1),IVal(i2)) -> let f2 = getFE (cast (IVal(i2)) (typeOfPrec (getFEPrec v1))) in
								FVal(feMult f1 f2 (getFEPrec v1))
|	(XVal(f1),XVal(f2)) ->  XVal(fiMult f1 f2) 
| 	(XVal(f1),IVal(i2)) -> 	let f2 = castIntInterval2FX i2 99 (getFPrec f1) in
							let f2' = if ((getIPrec f2)=99) then 
										castIntInterval2FX i2 (getIPrec f1) (getFPrec f1) 
								      else f2 
							in XVal(fiPlus f1 f2')
| 	(IVal(i1),XVal(f2)) -> 	let f1 = castIntInterval2FX i1 99 (getFPrec f2) in
							let f1' = if ((getIPrec f1)=99) then 
										castIntInterval2FX i1 (getIPrec f2) (getFPrec f2) 
								      else f1 
							in XVal(fiPlus f1' f2)
|   (XVal(f1),FVal(f2)) ->  let f2' = castFE2FX f2 (getIPrec f1) (getFPrec f1) 
							in XVal(fiMult f1 f2')
|   (FVal(f1),XVal(f2)) ->  let f1' = castFE2FX f1 (getIPrec f2) (getFPrec f2) 
							in XVal(fiMult f1' f2)
|	(AVal(a1),AVal(a2)) ->	if (((Array.length a1)=1) && ((Array.length a2)=1) &&
								((Array.length a1.(0))=1) && ((Array.length a2.(0))=1)) then
									(* a1 and a2 are 1x1 *)
									valMult a1.(0).(0) a2.(0).(0)
							else
								if (((Array.length a1)=1) && ((Array.length a1.(0))=1)) then
									(* a1 is 1x1 *)
									valMult a1.(0).(0) v2
								else
									if (((Array.length a2)=1) && ((Array.length a2.(0))=1)) then
										(* a2 is 1x1 *)
										valMult v1 a2.(0).(0) 
							    	else
										if (((Array.length a1.(0))=1) && ((Array.length a2)=1) &&
										    ((Array.length a1)=(Array.length a2.(0)))) then
											(* a1 is 1xN and a2 is Nx1 *)
												let _ = if (not !warnMsgScalarProd) then
															(print_string("\n!!! Scalar product computed using left associativity.") ; 
															 warnMsgScalarProd := true
															)
												in scalarProd a2 a1
										else
											if (((Array.length a2.(0))=1) && ((Array.length a1)=1) &&
											    ((Array.length a2)=(Array.length a1.(0)))) then
												(* a1 is Nx1 and a2 is 1xN *)
												let _ = if (not !warnMsgScalarProd) then
															(print_string("\n!!! Scalar product computed using left associativity.") ; 
															 warnMsgScalarProd := true
															)
												
												in scalarProd a1 a2
											else
												let _ = if (not !warnMsgMatProd) then
															(print_string("\n!!! Matrix product computed using left associativity.") ; 
															 warnMsgMatProd := true
															)
												
												in matrixMult a1 a2  
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMult v1 a.(i).(j)
									done
								done
							in AVal(b) 							
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMult a.(i).(j) v2
									done
								done
							in AVal(b) 							
|	(BVal(_),_) -> 			raise (Error ("valMult","??? Type mismatch in multiplication (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valMult","??? Type mismatch in multiplication (boolean not expected)."))
|	(ValTop,_) -> 			ValTop
|	(_,ValTop) ->			ValTop
|	(ValBottom,_) ->		ValBottom
|	(_,ValBottom) ->		ValBottom
|   (ArrayLines(_),_) -> 	raise (Error ("valMult","??? ArrayLines not expected in valMult."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valMult","??? ArrayLines not expected in valMult."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valMult","??? ArrayColumns not expected in valMult."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valMult","??? ArrayColumns not expected in valMult."))


and scalarProd m1 m2 =
	let l1 = Array.length m1 in
	let l2 = Array.length m2 in
	let c1 = Array.length m1.(0) in
	let c2 = Array.length m2.(0) in
	let res = ref (IVal(!i32Zero)) in
		if ((c1=1) && (l2=1) && (l1=c2)) then
			let _ =	for i=0 to l1-1 do
						res := valPlus !res (valMult m1.(i).(0) m2.(0).(i))
					done
			in !res
		else
			if ((l1=1) && (c2=1) && (c1=l2)) then
				let _ =	for i=0 to l2-1 do
							res := valPlus !res (valMult m2.(i).(0) m1.(0).(i))
						done
				in !res
			else
				raise (Error ("scalarProd","??? Dimensions must agree, cannot perform scalar product.")) 


and matrixMult m1 m2 =
	let l1 = Array.length m1 in      
	let c1 = Array.length m1.(0) in 
	let l2 = Array.length m2 in     
	let c2 = Array.length m2.(0) in 
	let res = Array.make_matrix l1 c2 (IVal(!i32Zero)) in
		if (c1=l2) then
			let _ =	for i=0 to l1-1 do
						for j=0 to c2-1 do
							for k=0 to (c1-1) do
								res.(i).(j) <- valPlus (res.(i).(j)) (valMult m1.(i).(k) m2.(k).(j))
							done
						done
					done
			in AVal(res)
		else
			raise (Error ("matrixMult",("??? Matrix dimensions must agree, cannot perform product."^
			(string_of_int l1)^" "^(string_of_int c1)^" "^(string_of_int l2)^" "^(string_of_int c2)^" "))) ;;


let rec valDotMult v1 v2 = match (v1,v2) with
	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
									for i=0 to ((Array.length b)-1) do
										for j=0 to ((Array.length b.(i))-1) do
											b.(i).(j) <- valMult a1.(i).(j) a2.(i).(j) 
										done
									done
							in AVal(b) 
|	(ValTop,_) -> 			ValTop
|	(_,ValTop) ->			ValTop
|	(ValBottom,_) ->		ValBottom
|   (_,ValBottom) ->		ValBottom
| 	_ -> 					raise (Error ("valDotMult","??? Matrices expected as arguments of the .* operator .")) ;;



let rec valDiv v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) -> IVal(iDiv i1 i2)
|	(FVal(f1),FVal(f2)) -> FVal(feDiv f1 f2 (max (getFEPrec v1) (getFEPrec v2)))
|	(IVal(i1),FVal(f2)) -> let f1 = getFE (cast (IVal(i1)) (typeOfPrec (getFEPrec v2))) in
							FVal(feDiv f1 f2 (getFEPrec v2))
|	(FVal(f1),IVal(i2)) -> let f2 = getFE (cast (IVal(i2)) (typeOfPrec (getFEPrec v1))) in
							FVal(feDiv f1 f2 (getFEPrec v1))
|   (XVal(_),_) ->			raise (Error ("valDiv","??? Division with fixed-point operand not implemented."))
|   (_,XVal(_)) ->			raise (Error ("valDiv","??? Division with fixed-point operand not implemented."))
|	(FVal(_),AVal(a)) ->	let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valDiv v1 a.(i).(j)
									done
								done
							in AVal(b) 							
|	(AVal(a),FVal(_)) -> 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valDiv a.(i).(j) v2
									done
								done
							in AVal(b) 							
|	(IVal(_),AVal(a)) ->	let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valDiv v1 a.(i).(j)
									done
								done
								
							in AVal(b) 							
|	(AVal(a),IVal(_)) -> 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <-  valDiv a.(i).(j) v2
									done
								done
							in AVal(b) 	
(* |	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valDiv a1.(i).(j) a2.(i).(j) 
									done
								done
							in AVal(b) *)
|	(BVal(_),_) -> 			raise (Error ("valDiv","??? Type mismatch in division (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valDiv","??? Type mismatch in division (boolean not expected)."))
|	(ValTop,_) -> 			ValTop
|	(_,ValTop) ->			ValTop
|	(ValBottom,_) ->		ValBottom
|	(_,ValBottom) ->		ValBottom
|	(AVal(_),AVal(_)) ->	raise (Error ("ValDiv","??? Division of matrices not implemented."))  
|   (ArrayLines(_),_) -> 	raise (Error ("valDiv","??? ArrayLines not expected in valDiv."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valDiv","??? ArrayLines not expected in valDiv."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valDiv","??? ArrayColumns not expected in valDiv."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valDiv","??? ArrayColumns not expected in valDiv."))
;;


let rec valDotDiv v1 v2 = match (v1,v2) with
	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
									for i=0 to ((Array.length b)-1) do
										for j=0 to ((Array.length b.(i))-1) do
											b.(i).(j) <- valDiv a1.(i).(j) a2.(i).(j) 
										done
									done
							in AVal(b) 
|	(ValTop,_) -> 			ValTop
|	(_,ValTop) ->			ValTop
|	(ValBottom,_) ->		ValBottom
|   (_,ValBottom) ->		ValBottom
| 	_ -> 					raise (Error ("valDotDiv","??? Matrices expected as arguments of the ./ operator .")) ;;


let rec valNeg v = match v with
	IVal(i) -> 			IVal(iNeg i)
|	FVal(f) ->	 		FVal(feNeg f (getFEPrec v))
|	XVal(f) ->	 		XVal(fiNeg f)
|	AVal(a) ->	 		let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valNeg a.(i).(j) 
									done
								done
							in AVal(b) 
|	BVal(_) -> 			raise (Error ("valUMinus","??? Type mismatch in unary minus (boolean not expected)."))
|	ValTop -> 			ValTop
|	ValBottom ->		ValBottom
|   ArrayLines(_) ->	raise (Error ("valUMinus","??? ArrayLines not expected in valUMinus."))
|   ArrayColumns(_) -> 	raise (Error ("valUMinus","??? ArrayColumns not expected in valUMinus."))
;;


let rec valSqrt v = match v with 
	IVal(i) -> 			IVal(iSqrt i)
|	FVal(f) -> 			FVal(feSqrt f (getFEPrec v))
|   XVal(_) ->			raise (Error ("valSqrt","??? Square root of fixed point number not implemented."))
|	AVal(a) -> 			let b = copyArray a in
						let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <-  valSqrt a.(i).(j) 
									done	
								done
						in AVal(b) 	
|	BVal(_) -> 			raise (Error ("valSqrt","??? Type mismatch in square root (boolean not expected)."))
|	ValTop ->			ValTop
|	ValBottom ->		ValBottom
|   ArrayLines(_) ->	raise (Error ("valSqrt","??? ArrayLines not expected in valSqrt."))
|   ArrayColumns(_) -> 	raise (Error ("valSqrt","??? ArrayColumns not expected in valSqrt."))
;;


let rec valAbs v = match v with 
	IVal(i) -> 			IVal(iAbs i)
|	FVal(f) -> 			FVal(feAbs f (getFEPrec v))
|	XVal(f) -> 			XVal(fiAbs f)
|	AVal(a) ->		 	let b = copyArray a in
						let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <-  valAbs a.(i).(j) 
									done	
								done
						in AVal(b) 	
|	BVal(_) -> 			raise (Error ("valAbs","??? Type mismatch in square root (boolean not expected)."))
|	ValTop ->			ValTop
|	ValBottom ->		ValBottom
|   ArrayLines(_) ->	raise (Error ("valAbs","??? ArrayLines not expected in valAbs."))
|   ArrayColumns(_) -> 	raise (Error ("valAbs","??? ArrayColumns not expected in valAbs."))
;;


let rec valSseq v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) -> 			iSseq i1 i2
|	(FVal(f1),FVal(f2)) ->			feSseq f1 f2 
|	(XVal(f1),XVal(f2)) ->			fiSseq f1 f2 
|	(XVal(f1),IVal(_)) -> 			valSseq v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(IVal(_),XVal(f2)) -> 			valSseq (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2
|	(IVal(_),FVal(FE(_,_,p))) ->	valSseq (cast v1 (typeOfPrec p)) v2
|	(FVal(FE(_,_,p)),IVal(_)) ->	valSseq v1 (cast v2 (typeOfPrec p))
|   (XVal(f1),FVal(_)) ->			valSseq v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|   (FVal(_),XVal(f2)) ->			valSseq (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2
|	(AVal(a1),AVal(a2)) -> 	let x = ref false in
							let _ = checkArrayDims a1 a2  in
							let _ = x := true in
							let _ = 
								for i=0 to ((Array.length a1)-1) do
									for j=0 to ((Array.length a1.(i))-1) do
										x := (!x) && (valSseq a1.(i).(j) a2.(i).(j))
									done
								done
							in !x
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v1 
									done
								done
							in valSseq (AVal(b)) v2		
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v2
									done
								done
							in valSseq v1 (AVal(b)) 							
|	(ValBottom,_) ->				true
|	(_,ValBottom) ->				false
|	(_,ValTop) ->					true
|	(ValTop,_) ->					false
|	_ -> raise (Error ("valSseq","coercion problem")) ;;


let rec valSseqNeglectErrors v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) -> 				iSseq i1 i2
|	(FVal(f1),FVal(f2)) ->				feSseqNeglectErrors f1 f2 
|   (XVal(f1),XVal(f2)) -> 				fiSseqNeglectErrors f1 f2
|	(FVal(FE(_,_,p)),IVal(_)) ->		valSseqNeglectErrors v1 (cast v2 (typeOfPrec p)) 
|	(IVal(_),FVal(FE(_,_,p))) ->		valSseqNeglectErrors (cast v1 (typeOfPrec p)) v2
|	(XVal(FX(_,ip,fp,_)),IVal(_)) ->	valSseqNeglectErrors v1 (cast v2 (Type_Fixed(ip,fp))) 
|	(IVal(_),XVal(FX(_,ip,fp,_))) ->	valSseqNeglectErrors (cast v1 (Type_Fixed(ip,fp))) v2
|	(AVal(a1),AVal(a2)) -> 	let x = ref false in
							let _ = checkArrayDims a1 a2  in
							let _ = x := true in
							let _ = 
								for i=0 to ((Array.length a1)-1) do
									for j=0 to ((Array.length a1.(i))-1) do
										x := (!x) && (valSseqNeglectErrors a1.(i).(j) a2.(i).(j))
									done
								done
							in !x
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v1 
									done
								done
							in valSseqNeglectErrors (AVal(b)) v2		
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v2
									done
								done
							in valSseqNeglectErrors v1 (AVal(b)) 							
|	(ValBottom,_) ->					true
|	(_,ValBottom) ->					false
|	(_,ValTop) ->						true
|	(ValTop,_) ->						false
|	_ -> raise (Error ("valSseq","coercion problem")) ;;


let rec valMeet v1 v2 = match (v1,v2) with
	(IVal(i1),IVal(i2)) ->		 	IVal(iMeet i1 i2)
|	(FVal(f1),FVal(f2)) -> 			FVal(feMeet f1 f2) 
|	(XVal(f1),XVal(f2)) -> 			XVal(fiMeet f1 f2)
|	(XVal(f1),IVal(_)) -> 			valMeet v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(IVal(_),XVal(f2)) -> 			valMeet (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2
|	(XVal(f1),FVal(_)) -> 			valMeet v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(FVal(_),XVal(f2)) -> 			valMeet (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2
|	(FVal(FE(_,_,p)),IVal(_)) ->	valMeet v1 (cast v2 (typeOfPrec p))
|	(IVal(_),FVal(FE(_,_,p))) ->	valMeet (cast v1 (typeOfPrec p)) v2
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMeet a1.(i).(j) a2.(i).(j) 
									done
								done
							in AVal(b) 
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMeet v1 a.(i).(j)
									done
								done
							in AVal(b) 							
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valMeet a.(i).(j) v2
									done
								done
							in AVal(b) 							
|	(ValTop,_) ->					v2
|	(_,ValTop) ->					v1		
|   _ -> ValBottom  
;;
		
		
let rec valJoin x y = match (x,y) with
	(IVal(v1),IVal(v2)) -> 			IVal(iJoin v1 v2)
|	(FVal(v1),FVal(v2)) -> 			FVal(feJoin v1 v2)
|	(XVal(v1),XVal(v2)) -> 			XVal(fiJoin v1 v2)
|	(FVal(FE(_,_,p)),IVal(_)) ->	valJoin x (cast y (typeOfPrec p))
|	(IVal(_),FVal(FE(_,_,p))) ->	valJoin (cast x (typeOfPrec p)) y
|	(XVal(v1),IVal(_)) -> 			valJoin x (cast y (Type_Fixed(getIPrec v1,getFPrec v1)))
|	(IVal(_),XVal(v2)) -> 			valJoin (cast x (Type_Fixed(getIPrec v2,getFPrec v2))) y
|	(XVal(v1),FVal(_)) -> 			valJoin x (cast y (Type_Fixed(getIPrec v1,getFPrec v1)))
|	(FVal(_),XVal(v2)) -> 			valJoin (cast x (Type_Fixed(getIPrec v2,getFPrec v2))) y
| 	(BVal(v1),BVal(v2)) -> 			BVal(bJoin v1 v2)
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let b = copyArray a1 in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- valJoin a1.(i).(j) a2.(i).(j) 
									done
								done
							in AVal(b) 
|	(_,AVal(a2)) ->			let b1 = copyArray a2 in 
							let _ = 
								for i=0 to ((Array.length b1)-1) do
									for j=0 to ((Array.length b1.(i))-1) do
										b1.(i).(j) <- valJoin x a2.(i).(j)
									done
								done
							in AVal(b1) 							
|	(AVal(a1),_) ->		 	let b2 = copyArray a1 in
							let _ = 
								for i=0 to ((Array.length b2)-1) do
									for j=0 to ((Array.length b2.(i))-1) do
										b2.(i).(j) <- valJoin a1.(i).(j) y
									done
								done
							in AVal(b2) 							
|   (ValBottom,_) -> 				y
|	(_,ValBottom) -> 	 			x
|	(ValTop,_) -> 					ValTop
|	(_,ValTop) ->					ValTop
|   _ -> 							raise (Error ("valJoin","cannot coerce types"))  ;;


let rec valEq v1 v2 = match (v1,v2) with  (* only compare floats not errors terms *)
	(IVal(i1),IVal(i2)) -> 	iEq i1 i2
|	(FVal(FE(f1,_,_)),FVal(FE(f2,_,_))) -> 	mpfiEq f1 f2
|   (XVal(FX(x1,_,_,_)),XVal(FX(x2,_,_,_))) ->  fixedIntEq x1 x2 
|	(IVal(i1),FVal(f2)) -> 	valEq (cast v1 Type_Float_64) v2 
|	(FVal(f1),IVal(i2)) -> 	valEq v1 (cast v2 Type_Float_64) 
|	(IVal(i1),XVal(f2)) -> 	valEq (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2 
|	(XVal(f1),IVal(i2)) -> 	valEq v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(FVal(_),XVal(_)) -> 	raise (Error ("valEq","??? Type mismatch in equality test (float vs fixed)."))
|	(XVal(_),FVal(_)) -> 	raise (Error ("valEq","??? Type mismatch in equality test (float vs fixed)."))
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let x = ref (B(true)) in
							let _ = 
								for i=0 to ((Array.length a1)-1) do
									for j=0 to ((Array.length a1.(i))-1) do
										x := bAnd !x (valEq a1.(i).(j) a2.(i).(j)) 
									done
								done
							in !x

|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v1 
									done
								done
							in valEq (AVal(b)) v2		
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v2
									done
								done
							in valEq v1 (AVal(b)) 							
|	(BVal(_),_) -> 			raise (Error ("valEq","??? Type mismatch in comparison (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valEq","??? Type mismatch in comparison (boolean not expected)."))
|	(ValTop,_) -> 			BTop
|	(_,ValTop) ->			BTop
|	(ValBottom,_) ->		BBottom
|	(_,ValBottom) ->		BBottom
|   (ArrayLines(_),_) -> 	raise (Error ("valEq","??? ArrayLines not expected in valEq."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valEq","??? ArrayLines not expected in valEq."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valEq","??? ArrayColumns not expected in valEq."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valEq","??? ArrayColumns not expected in valEq."))
;;


let rec valLt v1 v2 = match (v1,v2) with  (* only compare floats not errors terms *)
	(IVal(i1),IVal(i2)) -> 	iLt i1 i2
|	(FVal(FE(f1,_,_)),FVal(FE(f2,_,_))) -> 	mpfiLt f1 f2
|   (XVal(FX(x1,_,_,_)),XVal(FX(x2,_,_,_))) ->  fixedIntLt x1 x2 
|	(IVal(i1),FVal(f2)) -> 	valLt (cast v1 Type_Float_64) v2 
|	(FVal(f1),IVal(i2)) -> 	valLt v1 (cast v2 Type_Float_64)
|	(IVal(i1),XVal(f2)) -> 	valLt (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2 
|	(XVal(f1),IVal(i2)) -> 	valLt v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(FVal(_),XVal(_)) -> 	raise (Error ("valLt","??? Type mismatch in comparison (float vs fixed)."))
|	(XVal(_),FVal(_)) -> 	raise (Error ("valLt","??? Type mismatch in comparison test (float vs fixed)."))
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let x = ref (B(true)) in
							let _ = 
								for i=0 to ((Array.length a1)-1) do
									for j=0 to ((Array.length a1.(i))-1) do
										x := bAnd !x (valLt a1.(i).(j) a2.(i).(j)) 
									done
								done
							in !x
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v1 
									done
								done
							in valLt (AVal(b)) v2		
|	(AVal(a),_) -> 			let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v2
									done
								done
							in valLt v1 (AVal(b)) 							
|	(BVal(_),_) -> 			raise (Error ("valLt","??? Type mismatch in comparison (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valLt","??? Type mismatch in comparison (boolean not expected)."))
|	(ValTop,_) -> 			BTop
|	(_,ValTop) ->			BTop
|	(ValBottom,_) ->		BBottom
|	(_,ValBottom) ->		BBottom
|   (ArrayLines(_),_) -> 	raise (Error ("valLt","??? ArrayLines not expected in valLt."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valLt","??? ArrayLines not expected in valLt."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valLt","??? ArrayColumns not expected in valLt."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valLt","??? ArrayColumns not expected in valLt."))
;;



let rec valLte v1 v2 = match (v1,v2) with  (* only compare floats not errors terms *)
	(IVal(i1),IVal(i2)) -> 	iLte i1 i2
|	(FVal(FE(f1,_,_)),FVal(FE(f2,_,_))) -> 	mpfiLte f1 f2
|   (XVal(FX(x1,_,_,_)),XVal(FX(x2,_,_,_))) ->  fixedIntLte x1 x2 
|	(IVal(i1),FVal(f2)) -> 	valLte (cast v1 Type_Float_64) v2 
|	(FVal(f1),IVal(i2)) -> 	valLte v1 (cast v2 Type_Float_64)
|	(IVal(i1),XVal(f2)) -> 	valLte (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2 
|	(XVal(f1),IVal(i2)) -> 	valLte v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(FVal(_),XVal(_)) -> 	raise (Error ("valLte","??? Type mismatch in comparison (float vs fixed)."))
|	(XVal(_),FVal(_)) -> 	raise (Error ("valLte","??? Type mismatch in comparison (float vs fixed)."))
|	(AVal(a1),AVal(a2)) -> 	let _ = checkArrayDims a1 a2  in
							let x = ref (B(true)) in
							let _ = 
								for i=0 to ((Array.length a1)-1) do
									for j=0 to ((Array.length a1.(i))-1) do
										x := bAnd !x (valLt a1.(i).(j) a2.(i).(j)) 
									done
								done
							in !x
|	(_,AVal(a)) ->			let b = copyArray a in 
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v1 
									done
								done
							in valLte (AVal(b)) v2		
|	(AVal(a),_) ->		 	let b = copyArray a in
							let _ = 
								for i=0 to ((Array.length b)-1) do
									for j=0 to ((Array.length b.(i))-1) do
										b.(i).(j) <- v2
									done
								done
							in valLte v1 (AVal(b)) 							
|	(BVal(_),_) -> 			raise (Error ("valLte","??? Type mismatch in comparison (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valLte","??? Type mismatch in comparison (boolean not expected)."))
|	(ValTop,_) -> 			BTop
|	(_,ValTop) ->			BTop
|	(ValBottom,_) ->		BBottom
|	(_,ValBottom) ->		BBottom
|   (ArrayLines(_),_) -> 	raise (Error ("valLte","??? ArrayLines not expected in valLte."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valLte","??? ArrayLines not expected in valLte."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valLte","??? ArrayColumns not expected in valLte."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valLte","??? ArrayColumns not expected in valLte."))
;;


let valNeq v1 v2 = bOr (valLt v1 v2) (valLt v2 v1) ;;


let isValNonNegative v = match v with
	IVal(I32(a),_) -> 	if ((Int32.compare a Int32.zero)>=0) then true else false
|	IVal(I64(a),_) -> 	if ((Int64.compare a Int64.zero)>=0) then true else false
|	FVal(FE((x,_),_,_)) -> mpFloatGte x (M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0))
|  	XVal(FX((x,_),_,_,_)) -> getFixedSign x 
|	_ -> raise (Error ("isValNonNegative","??? Unexpected type of value in comparison.")) ;;


let isValNonPositive v = match v with
	IVal(_,I32(a)) -> 	if ((Int32.compare a Int32.zero)<=0) then true else false
|	IVal(_,I64(a)) -> 	if ((Int64.compare a Int64.zero)<=0) then true else false
|	FVal(FE((_,x),_,_)) -> mpFloatLte x (M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0))
|  	XVal(FX((_,x),_,_,_)) -> (not (getFixedSign x)) || (isXZero x)
|	_ -> raise (Error ("isValNonNegative","??? Unexpected type of value in comparison.")) ;;


let rec valArrayMin v = match v with 
	IVal(_) -> v
|	FVal(_) -> v
|   XVal(_) -> v 
|	AVal(a) -> 	let b = ref a.(0).(0) in
				let _ = 
						for i=0 to ((Array.length a)-1) do
							for j=0 to ((Array.length a.(i))-1) do
								if ((valLt a.(i).(j) (!b))=(B(true)) || (valLt a.(i).(j) (!b))=BTop)  then
									b := a.(i).(j) 
							done	
						done
				in (!b) 	
|	BVal(_) -> 	v
|	ValTop ->	ValTop
|	ValBottom -> ValBottom
|   ArrayLines(_) -> raise (Error ("valArrayMin","Unexpected arrayLines"))
|   ArrayColumns(_) -> raise (Error ("valArrayMin","Unexpected arrayColumns")) 
;;


let rec valArrayMax v = match v with 
	IVal(_) -> v
|	FVal(_) -> v
| 	XVal(_) -> v
|	AVal(a) -> 	let b = ref a.(0).(0) in
				let _ = 
						for i=0 to ((Array.length a)-1) do
							for j=0 to ((Array.length a.(i))-1) do
								if ((valLt (!b) a.(i).(j))=(B(true)) || (valLt a.(i).(j) (!b))=BTop)  then
									b := a.(i).(j) 
							done	
						done
				in (!b) 	
|	BVal(_) -> 	v
|	ValTop ->	ValTop
|	ValBottom -> ValBottom
|   ArrayLines(_) -> raise (Error ("valArrayMax","Unexpected arrayLines"))
|   ArrayColumns(_) -> raise (Error ("valArrayMax","Unexpected arrayColumns")) 
;;

(*****************************************************************************)
(*                         Comparisons based on errors                       *)
(*****************************************************************************)

let errLt (v0Min,v0Max) (v1Min,v1Max) =
	let v0 = mpFloatMax (mpFloatAbs v0Min !errPrec GMP_RNDU) (mpFloatAbs v0Max !errPrec GMP_RNDU) in
	let v1 = mpFloatMax (mpFloatAbs v1Min !errPrec GMP_RNDU) (mpFloatAbs v1Max !errPrec GMP_RNDU) in
 		if (mpFloatLt v0 v1) then true
		else false ;;
		

let rec valArrayMaxErr v = match v with 
	IVal(_) -> v
|	FVal(_) -> v
| 	XVal(_) -> v
|	AVal(a) -> 	let b = ref a.(0).(0) in
				let _ = 
						for i=0 to ((Array.length a)-1) do
							for j=0 to ((Array.length a.(i))-1) do
								if (valErrLt (!b) a.(i).(j)) then
									b := a.(i).(j) 
							done	
						done
				in (!b) 	
|	BVal(_) -> 	v
|	ValTop ->	ValTop
|	ValBottom -> ValBottom
|   ArrayLines(_) -> raise (Error ("valArrayMax","Unexpected arrayLines"))
|   ArrayColumns(_) -> raise (Error ("valArrayMax","Unexpected arrayColumns")) 


and  valErrLt v1 v2 = match (v1,v2) with  (* only compare absolute val of errors not floats *)
	(IVal(i1),IVal(i2)) -> 	true
|	(FVal(FE(_,e1,_)),FVal(FE(_,e2,_))) -> 	errLt e1 e2
|   (XVal(FX(_,_,_,e1)),XVal(FX(_,_,_,e2))) ->  errLt e1 e2 
|	(IVal(i1),FVal(f2)) -> 	valErrLt (cast v1 Type_Float_64) v2 
|	(FVal(f1),IVal(i2)) -> 	valErrLt v1 (cast v2 Type_Float_64)
|	(IVal(i1),XVal(f2)) -> 	valErrLt (cast v1 (Type_Fixed(getIPrec f2,getFPrec f2))) v2 
|	(XVal(f1),IVal(i2)) -> 	valErrLt v1 (cast v2 (Type_Fixed(getIPrec f1,getFPrec f1)))
|	(FVal(_),XVal(_)) -> 	raise (Error ("valErrLt","??? Type mismatch in comparison (float vs fixed)."))
|	(XVal(_),FVal(_)) -> 	raise (Error ("valErrLt","??? Type mismatch in comparison test (float vs fixed)."))
|	(AVal(a1),AVal(a2)) -> 	let r1 = valArrayMaxErr v1 in
							let r2 = valArrayMaxErr v2 in valErrLt r1 r2
|	(_,AVal(a)) ->			let r2 = valArrayMaxErr v2 
							in valErrLt v1 r2
|	(AVal(a),_) -> 			let r1 = valArrayMaxErr v1
							in valErrLt r1 v2		
|	(BVal(_),_) -> 			raise (Error ("valLt","??? Type mismatch in comparison (boolean not expected)."))
|	(_,BVal(_)) -> 			raise (Error ("valLt","??? Type mismatch in comparison (boolean not expected)."))
|	(ValTop,_) -> 			false
|	(_,ValTop) ->			true
|	(ValBottom,_) ->		true
|	(_,ValBottom) ->		false
|   (ArrayLines(_),_) -> 	raise (Error ("valLt","??? ArrayLines not expected in valErrLt."))
|   (_,ArrayLines(_)) -> 	raise (Error ("valLt","??? ArrayLines not expected in valErrLt."))
|   (ArrayColumns(_),_) -> 	raise (Error ("valLt","??? ArrayColumns not expected in valErrLt."))
|   (_,ArrayColumns(_)) -> 	raise (Error ("valLt","??? ArrayColumns not expected in valErrLt."))
;;

