(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file environments.ml contains the functions to handle the environments during analysis
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Boolean ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Coercion ;;
open Matrix ;;
open Semiformal ;;
open Parser ;;
open Widenings ;;
open Slice ;;


(*****************************************************************************)
(*                                     Misc                                  *)
(*****************************************************************************)

let rec sseqEnv env1 env2 = match env1 with
	[] -> 				true
|	(id,v)::env1' ->	let v' = getEnvComplete id env2 ValBottom
						in if (valSseq v v') then sseqEnv env1' env2 else false ;;


let rec sseqEnvNeglectErrors env1 env2 = match env1 with
	[] -> 				true
|	(id,v)::env1' ->	let v' = getEnvComplete id env2 ValBottom
						in 	if (valSseqNeglectErrors v v') then sseqEnvNeglectErrors env1' env2 
							else false ;; 
		
		
let rec sseqEnvNeglectErrorsList env envList = match envList with
	[] -> 		(false,[])
|	_::[] -> 	(false,[])
|	e::es ->	if (sseqEnv (* NeglectErrors *) env e) then (true,e) else sseqEnvNeglectErrorsList env es ;;			
					
						
let eqEnv env1 env2 = (sseqEnv env1 env2) && (sseqEnv env2 env1) ;;

						
let completeEnv env varList = env@(List.map (fun id -> (id,ValBottom)) varList) ;;
let completeTEnv env varList = env@(List.map (fun id -> (id,UndefinedType)) varList) ;;

						
(*****************************************************************************)
(*                             Analyze of tests                              *)
(*****************************************************************************)
	
let xor a b = (a || b) && (not (a && b)) ;;	
	
let buildCondEnv varList env = List.map (fun id -> (id,getEnvComplete id env ValBottom)) varList ;;


let rec getVarList e res = match e with
	Cst(_) ->						res
|	Cast(_,e1,_) ->					getVarList e1 res
|	Id(v,_) ->						if (List.mem v res) then res else v::res
|	ArrayElt(v,e1,e2,_) ->			getVarList e2 (getVarList e1 (if (List.mem v res) then res else v::res))
|	CreateMatrix(el,_,_) ->			getVarListList el res
|	ArrayBlock(v,e1,e2,e3,e4,_) -> 	getVarList e4 (getVarList e3 (getVarList e2 (getVarList e1 (if (List.mem v res) then res else v::res))))	
|	ArrayExpr(a,_) ->				getVarListArray a 0 0 res
|	FunCall(_,ael,_) ->				getVarListList ael res 
|	UndefinedCall(v,ael,_) ->		getVarListList ael (if (List.mem v res) then res else v::res)
|	Plus(e1,e2,_) ->				getVarList e2 (getVarList e1 res)
|	Minus(e1,e2,_) ->				getVarList e2 (getVarList e1 res)
|	Times(e1,e2,_) ->				getVarList e2 (getVarList e1 res) 
|	DotTimes(e1,e2,_) ->			getVarList e2 (getVarList e1 res) 
|	DotDiv(e1,e2,_) ->				getVarList e2 (getVarList e1 res) 
|	Div(e1,e2,_) ->					getVarList e2 (getVarList e1 res) 
|	Power(e1,e2,_) ->				getVarList e2 (getVarList e1 res) 
|	UMinus(e1,_) ->					getVarList e1 res 
|	Sqrt(e1,_) ->					getVarList e1 res 
|	Abs(e1,_) ->					getVarList e1 res 
|	Min(e1,_) ->					getVarList e1 res 
|	Max(e1,_) ->					getVarList e1 res 
|	Size(e1,_) ->					getVarList e1 res 


and getVarListList ael res = match ael with
	[] 		-> res
|	e::es	-> getVarListList es (getVarList e res) 


and getVarListArray a i j res =
	if (i>=(Array.length a)) then res
	else
		if (j>=(Array.length a.(0))) then getVarListArray a (i+1) 0 res
		else
			getVarListArray a i (j+1) (getVarList a.(i).(j) res) ;;


let rec getCondVarList e = match e with
	Eq(e1,e2,_) -> getVarList e1 (getVarList e2 [])
| 	Neq(e1,e2,_) -> getVarList e1 (getVarList e2 []) 
| 	Lt(e1,e2,_) -> getVarList e1 (getVarList e2 []) 
| 	Lte(e1,e2,_) -> getVarList e1 (getVarList e2 []) 
| 	Gt(e1,e2,_) -> getVarList e1 (getVarList e2 []) 
| 	Gte(e1,e2,_) -> getVarList e1 (getVarList e2 []) 
| 	And(e1,e2,_) -> (getCondVarList e1)@(getCondVarList e2) 
| 	Or(e1,e2,_) -> (getCondVarList e1)@(getCondVarList e2)
| 	Xor(e1,e2,_) -> (getCondVarList e1)@(getCondVarList e2)
| 	Not(e1,_) -> getCondVarList e1 ;;


let negRelation r = match r with
	Relation_EQ -> Relation_NEQ
|	Relation_NEQ -> Relation_EQ
|	Relation_LT -> Relation_GTE
|	Relation_LTE -> Relation_GT
| 	Relation_GT -> Relation_LTE
|	Relation_GTE -> Relation_LT ;;


let rec insertEnv env1 env2 = match env1 with
	[] -> 				env2
|	(id,v)::env1' ->	insertEnv env1' (setEnv env2 id v) ;;


let rec envMeetAux env1 env2 = match env1 with
	[] -> []
|	(id,v)::env1' -> let v' =  (getEnvComplete id env2 ValBottom) in
						if (v'=ValBottom) then
							envMeetAux env1' env2
						else (id,(valMeet v v'))::(envMeetAux env1' env2) ;;


let envMeet env1 env2 = 
	let pureMeet = envMeetAux env1 env2 in (* variables defined in both envs *)
	let env' = insertEnv env1 env2 in      (* variables defined in either envs, bad values for vars in both *)
		insertEnv pureMeet env' ;;		   (* correct result *)


let rec isEmptyEnv env = match env with
	[] -> true
|	(_,ValBottom)::env' -> isEmptyEnv env'
|	(id,_)::env' -> false ;;


let rec envJoin env res = match env with
	[] -> res
|	(id,v)::env' -> envJoin env' (setEnv res id (valJoin v (getEnvComplete id res ValBottom))) ;;


let rec tEnvJoin env res = match env with
	[] -> res
|	(id,t)::env' -> let env'' = try (setEnv res id (maxType t (getEnvComplete id res UndefinedType)))
							    with _ -> setEnv res id t
	in tEnvJoin env' env'' ;;


let rec typedZero v = match v with
	IVal(I64(_),_) -> 		Cst(IVal(!i64Zero),Lab(0))
|	IVal(_,I64(_)) -> 		Cst(IVal(!i64Zero),Lab(0))
|	IVal(_) -> 				Cst(IVal(!i32Zero),Lab(0))
|	FVal(FE(_,_,p)) -> 		(Cst(FVal(FE((M(FR.from_float_prec p GMP_RNDN 0.0),
									  	  M(FR.from_float_prec p GMP_RNDN 0.0)),(MMInfty,MPInfty),p)),Lab(0)))
| 	XVal(FX(_,ip,fp,_)) -> 	(Cst(XVal(FX((xZero ip fp,xZero ip fp),ip,fp,!eZero)),Lab(0))) 
|   AVal(a) ->				typedZero a.(0).(0)
|	_ -> 					raise (Error ("typedZero","??? Unexpected type of value in condition")) ;;


let rec solveCmp lhs rel rhs = match rel with
	Relation_EQ ->	valMeet lhs rhs
|	Relation_NEQ ->	let v1 = solveCmp lhs Relation_LT rhs in
					let v2 = solveCmp lhs Relation_GT rhs in
						valJoin v1 v2
|	Relation_LT ->	(match rhs with
						IVal(_,I32(i2)) -> 			valMeet lhs (IVal(IMInfty,I32(Int32.sub i2 Int32.one)))
					|	IVal(_,I64(i2)) -> 			valMeet lhs (IVal(IMInfty,I64(Int64.sub i2 Int64.one)))					
 					|	FVal(FE((_,f2),_,p)) -> 	valMeet lhs (FVal(FE((MMInfty,mpFloatPredecessor f2 p),(MMInfty,MPInfty),p))) 
 					|	XVal(FX((_,f2),ip,fp,_)) -> valMeet lhs (XVal(FX((XMInfty,f2),ip,fp,(MMInfty,MPInfty))))
					|	IVal(_,IMInfty) -> 			valMeet lhs (IVal(IBottom,IBottom))		
					|	IVal(_,IBottom) ->			valMeet lhs (IVal(IBottom,IBottom)) 
					|	IVal(_,IPInfty) ->			lhs
					|	ValBottom ->				ValBottom
					|	ValTop ->					ValBottom
					|	BVal(_) -> 					raise (Error ("solveCmp1","??? Unxepected type of value in boolean expression B."))
					|	AVal(_) ->	 				raise (Error ("solveCmp1","??? Unxepected type of value in boolean expression A."))
					|   _ ->						raise (Error ("solveCmp1","??? Unxepected type of value in boolean expression C."))
					)
|	Relation_LTE ->	(match rhs with
						IVal(_,i) -> 				valMeet lhs (IVal(IMInfty,i))
					|	FVal(FE((_,f2),_,p)) ->	 	valMeet lhs (FVal(FE((MMInfty,f2),(MMInfty,MPInfty),p)))
					|	XVal(FX((_,f2),ip,fp,_)) ->	valMeet lhs (XVal(FX((XMInfty,f2),ip,fp,(MMInfty,MPInfty))))
					|	_ -> 						raise (Error ("solveCmp2","??? Unxepected type of value in condition."))
					)			
|	Relation_GT ->	(match rhs with
						IVal(I32(i1),_) -> 			valMeet lhs (IVal(I32(Int32.add i1 Int32.one),IPInfty))
					|	IVal(I64(i1),_) -> 			valMeet lhs (IVal(I64(Int64.add i1 Int64.one),IPInfty))					
 					|	FVal(FE((f1,_),_,p)) -> 	valMeet lhs (FVal(FE((mpFloatSuccessor f1 p,MPInfty),(MMInfty,MPInfty),p)))
 					|	XVal(FX((f1,_),ip,fp,_)) -> valMeet lhs (XVal(FX((f1,XPInfty),ip,fp,(MMInfty,MPInfty))))
					|	IVal(IMInfty,_) -> 			valMeet lhs (IVal(IBottom,IBottom))		
					|	IVal(IBottom,_) ->			valMeet lhs (IVal(IBottom,IBottom)) 
					|	IVal(IPInfty,_) ->			lhs
					|	ValBottom ->				ValBottom
					|	ValTop ->					ValBottom
					|	BVal(_) -> 					raise (Error ("solveCmp1","??? Unxepected type of value in condition B."))
					|	AVal(_) ->	 				raise (Error ("solveCmp1","??? Unxepected type of value in condition A."))
					|   _ ->						raise (Error ("solveCmp1","??? Unxepected type of value in condition C."))
					)
|	Relation_GTE ->	(match rhs with
						IVal(i,_) -> 				valMeet lhs (IVal(i,IPInfty))
					|	FVal(FE((f1,_),_,p)) ->	 	valMeet lhs (FVal(FE((f1,MPInfty),(MMInfty,MPInfty),p)))
					|	XVal(FX((f1,_),ip,fp,_)) ->	valMeet lhs (XVal(FX((f1,XPInfty),ip,fp,(MMInfty,MPInfty))))
					|	_ -> 						raise (Error ("solveCmp4","??? Unxepected type of value in condition."))
					)				
;;			
			

let rec evalCmp lhs rel rhs = match rel with
	Relation_EQ ->	valEq lhs rhs
|	Relation_NEQ ->	valNeq lhs rhs
|	Relation_LT ->	valLt lhs rhs
|	Relation_LTE ->	valLte lhs rhs			
|	Relation_GT ->	valLt rhs lhs			
|	Relation_GTE ->	valLte rhs lhs
;;			


(*****************************************************************************)
(*                        		Analyze of loops	                         *)
(*****************************************************************************)	
	

let rec foldTEnvList el res = match el with
	[] ->			res
|	env::envs ->	foldTEnvList envs (tEnvJoin env res) ;;


let rec foldEnvList el res = match el with
	[] ->			res
|	env::envs ->	foldEnvList envs (envJoin env res) ;;


let rec makeList elt size res = 
	if (size<=0) then res
	else makeList elt (size-1) (elt::res) ;;
	

(*****************************************************************************)
(*                        Analyze of function calls                          *)
(*****************************************************************************)	
	
let rec buildFunCallTEnv id varList valList tEnv = match varList with
	[] ->	(match valList with
				[] -> tEnv
			|	_ -> raise (Error ("buildFuncallAux","??? Error using ==> "^id^"\n"^"Too many input arguments."))
			)
|	var1::vars ->	(match valList with
						[] -> raise (Error ("buildFuncallAux","??? Error using ==> "^id^"\n"^"Not enough input arguments."))
					|	val1::vals -> buildFunCallTEnv id vars vals (setEnv tEnv var1 (typeOf val1))
					) ;;

 
let rec buildFunCallVEnv id varList valList vEnv = match varList with
[] ->	(match valList with
			[] -> vEnv
		|	_ -> raise (Error ("buildFuncallVEnv","??? Error using ==> "^id^"\n"^"Too many input arguments."))
		)
|	var1::vars -> (match valList with
					[] -> raise (Error ("buildFuncallVEnv","??? Error using ==> "^id^"\n"^"Not enough input arguments."))
				|	val1::vals -> buildFunCallVEnv id vars vals (setEnv vEnv var1 val1)
				) ;;
				
					
let buildFunCallEnv id valList tEnv vEnv =
	let Fun(varList,_,_,_) = getEnv id !funEnv 
	in (buildFunCallTEnv id varList valList tEnv,buildFunCallVEnv id varList valList vEnv) ;;	
	
	
let rec setUpFunCallAux id ael l ch =	
	let v = !verbose in
	let lexbuf = Lexing.from_channel ch in
	let (cmd,txt) = try
							Parser.interaction Lexer.token lexbuf 
					with	Error(f,e) ->		   close_in ch ; raise (Error ("setUpFunCallAux",("??? At line "^(string_of_int (!codeLine))^" of "^id^".m"^ (": "^e^" ("^f^")")))) 
						|   Parsing.Parse_error -> close_in ch ; raise (Error ("setUpFunCallAux",("??? At line "^(string_of_int (!codeLine))^" of "^id^".m"^ (": Expression or statement is incorrect."))))
						|	e ->				   close_in ch ; print_string ("At line "^(string_of_int !codeLine)^" of '"^id^".m': ") ; raise e 
	in
		if ((String.compare txt "FUNCTION")!=0) then
			raise (Error ("setUpFunCallAux","??? Function definition expected in file '"^id^".m'."))
		else
			let _ = verbose := v in FunCall(id,ael,l) ;;		

					
let setUpFunCall id ael l =		
	if (funExists id) then
		let ch = Pervasives.open_in (id^".m") in
		let _ = codeLine := 1 in
		let res = setUpFunCallAux id ael l ch in
		let _ = close_in ch in res 
	else
		raise (Error ("SetUpFunCall","??? Undefined identifier '"^id^"'.")) ;;	


let rec buildOutputVector outputs env res = match outputs with
	[] -> let l = List.length res in
			if (l=0) then
				raise (Error ("buildOutputVector","impossible case in buildoutputvector"))
			else
				if (l=1) then (List.hd res)
				else 
					let a = Array.make_matrix 1 l ValBottom in
					let t = try maxTypeOfValList res UndefinedType 
							with _ -> raise (Error ("buildOutputVectpr","??? Functions must return values of consistent types."))
					in
						(match t with
							Type_Array(_) -> raise (Error ("buildOutputVectpr","??? Functions cannot return more than one array."))
						 | _ ->	let _ =
										for i=0 to (l-1) do
						 					a.(0).(i) <- cast (List.nth res (l-1-i)) t
										done
								in AVal(a)
						)
|   id::ids -> 	let v = try
							getEnv id env
						with
							Error(_) -> raise (Error ("buildOutputVector","??? Output variable '"^id^"' is not defined in function."))
				in buildOutputVector ids env (v::res) ;;
	 
	
let rec valListJoin vl1 vl2 res = match vl1 with
	[] -> (match vl2 with
			  [] -> res
	    	| _ -> raise (Error ("valListJoin","??? lists do not have the same size."))
		  )
|   x::xs -> (match vl2 with
		      [] -> raise (Error ("valListJoin","??? lists do not have the same size."))
			| y::ys -> valListJoin xs ys ((valJoin x y)::res)
		  ) ;;


let rec computeMaxHDelta x = match x with
	FVal(FE((a,b),e,p)) -> 		let d = mpFloatMinus b a p GMP_RNDU 
								in FVal(FE((mpFloatNeg d p GMP_RNDU,d),e,p))
|	XVal(FX((a,b),ip,fp,e)) -> 	let (d,_) = fixedMinus b a ip fp 
								in XVal(FX((fixedNeg d,d),ip,fp,e))
|   IVal(a,b) -> 				let d = integerMinus b a 
								in IVal(integerNeg d,d)	
|	AVal(a) ->					let b = copyArray a in
								let _ = 
									for i=0 to ((Array.length b)-1) do
										for j=0 to ((Array.length b.(i))-1) do
											b.(i).(j) <- computeMaxHDelta a.(i).(j)
										done
									done
								in AVal(b)
| 	ValBottom -> 				ValBottom
|   ValTop ->					ValTop
|   _ -> 						raise (Error ("computeDelta","??? should not happen in computeDelta")) ;;
	
	
let computeDeltaParams v1 v2 = match (v1,v2) with
	(FVal(FE((a1,b1),e1,p1)),FVal(FE((a2,b2),e2,p2))) -> 	let aMin = mpFloatMin a1 a2 in
															let aMax = mpFloatMax a1 a2 in
	 														let bMin = mpFloatMin b1 b2 in
															let bMax = mpFloatMax b1 b2 in
															let p = max p1 p2 in
															let h = mpFloatMax (mpFloatMinus bMax bMin p GMP_RNDU) (mpFloatMinus aMax aMin p GMP_RNDU) in
															let a = FVal(FE((aMin,aMin),mpfiJoin e1 e2,p)) in
															let b = FVal(FE((bMax,bMax),mpfiJoin e1 e2,p)) in
															let h' = FVal(FE((h,h),!eZero,p)) in
																(a,b,h')
|	(XVal(FX((a1,b1),ip1,fp1,e1)),XVal(FX((a2,b2),ip2,fp2,e2))) -> 	
															let aMin = fixedMin a1 a2 in
															let aMax = fixedMax a1 a2 in
 															let bMin = fixedMin b1 b2 in
															let bMax = fixedMax b1 b2 in
															let fp = max fp1 fp2 in
															let ip = max ip1 ip2 in
															let h = fixedMax (fst (fixedMinus bMax bMin ip fp)) (fst (fixedMinus aMax aMin ip fp)) in
															let a = XVal(FX((aMin,aMin),ip,fp,mpfiJoin e1 e2)) in
															let b = XVal(FX((bMax,bMax),ip,fp,mpfiJoin e1 e2)) in
															let h' = XVal(FX((h,h),ip,fp,!eZero)) in
																(a,b,h')
|   (IVal(a1,b1),IVal(a2,b2)) -> 							let aMin = integerMin a1 a2 in
															let aMax = integerMax a1 a2 in
 															let bMin = integerMin b1 b2 in
															let bMax = integerMax b1 b2 in
															let h = integerMax (integerMinus bMax bMin) (integerMinus aMax aMin) in
															let a = IVal(aMin,aMin) in
															let b = IVal(bMax,bMax) in
															let h' = IVal(h,h) in
																(a,b,h')	
|   _ -> 						raise (Error ("computeDeltaParams","??? should not happen in computeDeltaParams")) ;;
