(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file fixed.ml contains the domain and operations on fixed-point values
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Float ;;


(*****************************************************************************)
(*                           Auxiliary functions 1                           *)
(*****************************************************************************)

let floatOfFixed x r = match x with
	X(s,i,f) -> M(FR.from_string_prec_base float64_mantissa r 2 ((if s then "" else "-")^i^"."^f))
|   XPInfty -> MPInfty
|   XMInfty -> MMInfty
|   XBottom -> MBottom ;;


let mpfiOfFixedInt (x,y) = (floatOfFixed x GMP_RNDD,floatOfFixed y GMP_RNDU) ;;


let isXNonNegative x = match x with
	X(s,_,_) -> s
|	XPInfty -> true
| 	_ -> false ;;


let isXNonPositive x = match x with
	X(s,_,_) -> not s
|	XMInfty -> true
| 	_ -> false ;;


let isXZero x = match x with
	X(s,i,f) -> ((int_of_string i)=0)&&((int_of_string f)=0)
| 	_ -> false ;;


let mpFloatOfFixed x prec = match x with
X(s,i,f) -> if s then 
				M(FR.from_string_prec_base prec GMP_RNDN 2 (i^"."^f)) 
			else
				M(FR.from_string_prec_base prec GMP_RNDN 2 ("-"^i^"."^f)) 
|  	XPInfty	-> MPInfty
|   XMInfty -> MMInfty
|   XBottom -> MBottom ;;


let removeLeftmostZeros s =
	let i = try 
				String.index s '1'
			with 
				Not_found -> (-1) 
	in if (i>=0) then 
			String.sub s i ((String.length s)-i) 
	   else "0" ;;	
			

let removeRightmostZeros s =
	let i = try 
				String.rindex s '1' 
			with 
				Not_found -> (-1) 
	in if (i>=0) then
			String.sub s 0 (i+1) 
	   else "0" ;;	

(*****************************************************************************)
(*                           Compute Error Terms                             *)
(*****************************************************************************)

let rec computeResizeErrorU n f res = 
	if ((String.length f)=0) then
		mpFloatPlus res (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-n) !errPrec) !errPrec GMP_RNDU
	else
		if ((String.get f 0)='0') then
			computeResizeErrorU (n+1) (String.sub f 1 ((String.length f)-1)) res
		else
			let res' = mpFloatPlus res (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-n) !errPrec) !errPrec GMP_RNDU
			in computeResizeErrorU (n+1) (String.sub f 1 ((String.length f)-1)) res' ;;


let rec computeResizeErrorU n f res = 
	if ((String.length f)=0) then
		mpFloatPlus res (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-n) !errPrec) !errPrec GMP_RNDU
	else
		if ((String.get f 0)='0') then
			computeResizeErrorU (n+1) (String.sub f 1 ((String.length f)-1)) res
		else
			let res' = mpFloatPlus res (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-n) !errPrec) !errPrec GMP_RNDU
			in computeResizeErrorU (n+1) (String.sub f 1 ((String.length f)-1)) res' ;;


let rec computeResizeErrorD n f res = 
	if ((String.length f)=0) then
		mpFloatMult (mpFloatPlus res (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-n) !errPrec) !errPrec GMP_RNDD)
		            (M(FR.from_float_prec !errPrec GMP_RNDN (-1.0))) !errPrec GMP_RNDD
	else
		if ((String.get f 0)='0') then
			computeResizeErrorD (n+1) (String.sub f 1 ((String.length f)-1)) res
		else
			let res' = mpFloatPlus res (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-n) !errPrec) !errPrec GMP_RNDD
			in computeResizeErrorD (n+1) (String.sub f 1 ((String.length f)-1)) res' ;;
			

let resizeX x ip fp ip' fp' = match x with
	X(s,i,f) -> if (ip>ip') then 
	 				if s then (XPInfty,!eZero) else (XMInfty,!eZero)
		     	else
			  		let i' = (String.make (ip'-(String.length i)) '0')^i  
					in 	if (fp=fp') then (X(s,i',f),!eZero)
						else
						if (fp<fp') then
							let f' = f^(String.make (fp'-(String.length f)) '0') 
							in (X(s,i',f'),!eZero)
				   	   else
							let f' = String.sub f 0 fp' in
							let e1 = computeResizeErrorD (fp'+1) (String.sub f (fp'+1) (fp-fp'-1)) (M(FR.from_float_prec !errPrec GMP_RNDN 0.0)) in
							let e2 = computeResizeErrorU (fp'+1) (String.sub f (fp'+1) (fp-fp'-1)) (M(FR.from_float_prec !errPrec GMP_RNDN 0.0)) 
							in  if s then
						 	 		(X(s,i',f'),(e2,e2)) 
					  			else (X(s,i',f'),(e1,e1))
|   _ -> (x,!eZero) 
;;


let resizeFX (FX((x,y),ip,fp,e)) ip' fp' =
	let (x',ex) = resizeX x ip fp ip' fp' in
	let (y',ey) = resizeX y ip fp ip' fp' in
	let e' = mpfiPlus (mpfiPlus ex ey !errPrec OUTSIDE) e !errPrec OUTSIDE 
	in FX((x',y'),ip',fp',e') ;;


(*****************************************************************************)
(*                           Auxiliary functions 2                           *)
(*****************************************************************************)


let xZero ip fp = X(true,String.make ip '0',String.make fp '0') ;;
let xOne ip fp = X(true,(String.make (ip-1) '0')^"1",String.make fp '0') ;;


let fxIntegerLength x = match x with
	X(_,i,_) -> String.length i
|   _ -> 0 ;;
 

let rec int64OfBinStringAux s r = 
  if ((String.length s)=0) then r 
  else int64OfBinStringAux (String.sub s 1 ((String.length s)-1)) 
                      (Int64.add (Int64.mul r (Int64.of_int 2)) (Int64.of_string (String.sub s 0 1))) ;;


let rec int64OfBinString s = int64OfBinStringAux s Int64.zero ;; 


let rec binStringOfInt64Aux n = 
  if ((Int64.compare n Int64.zero)=0) then "" else
  (binStringOfInt64Aux (Int64.div n (Int64.of_int 2)))^
  (if ((Int64.compare (Int64.rem n (Int64.of_int 2)) Int64.zero)=0) then "0" else "1") ;;


let binStringOfInt64 n  =
	try
		let s = "0"^binStringOfInt64Aux n in
		let i = String.index s '1' in
		let s' = String.sub s i ((String.length s)-i) in
			(s',String.length s') 
	with
		Not_found -> ("0",1) ;;
	

let fracAlign f1 f2 =
	let l1 = String.length f1 in
	let l2 = String.length f2 in
		if (l1>l2) then
			(f1,f2^(String.make (l1-l2) '0'))
		else
			if (l1<l2) then
				(f1^(String.make (l2-l1) '0'),f2) 
			else
				(f1,f2) ;;
			

let intAlign i1 i2 =
	let l1 = String.length i1 in
	let l2 = String.length i2 in
		if (l1>l2) then
			(i1,(String.make (l1-l2) '0')^i2)
		else
			if (l1<l2) then
				((String.make (l2-l1) '0')^i1,i2) 
			else
				(i1,i2) ;;
			
			
let fixedCut f l = 
	let fLength = String.length f 
	in if (fLength<l) then
		       ("0",f^(String.make (l-fLength) '0'))
	   else 
			let is = String.sub f 0 (fLength-l) in
			let fs = String.sub f (fLength-l) l in
				(is,fs) ;;


let simplifyZero f = if (String.contains f '1') then f else "0" ;;


(*****************************************************************************)
(*                    operations on Fixed point numbers                      *)
(*****************************************************************************)

		
let rec fixedEq a b = match (a,b) with
	(X(s1,i1,f1),X(s2,i2,f2)) -> 
		let (i1',i2') = intAlign i1 i2 in
		let (f1',f2') = fracAlign f1 f2 in
			(s1=s2)&&((String.compare i1' i2')=0)&&((String.compare f1' f2')=0)
|   (X(_),XPInfty) -> false
|   (XPInfty,X(_)) -> false
|   (X(_),XMInfty) -> false
|   (XMInfty,X(_)) -> false
|   (XPInfty,XPInfty) -> true
|   (XMInfty,XMInfty) -> true
|   (XMInfty,XPInfty) -> false
|   (XPInfty,XMInfty) -> false
|   (XBottom,XBottom) -> true
|   (XBottom,_) -> false
|   (_,XBottom) -> false ;;


let rec fixedGte a b = match (a,b) with
	(X(s1,i1,f1),X(s2,i2,f2)) -> 
		(match (s1,s2) with	
			(true,true) -> let (i1',i2') = intAlign i1 i2 in
							if ((String.compare i1' i2')=0) then 
								let (f1',f2') = fracAlign f1 f2 in
									if ((String.compare f1' f2')>=0) then true else false
							else
								if ((String.compare i1' i2')>=0) then true else false
		|	(false,false) -> fixedGte (X(true,i2,f2)) (X(true,i1,f1))
		|   (true,false) -> true
		|   (false,true) -> false
	)
|   (X(_),XPInfty) -> false
|   (XPInfty,X(_)) -> true
|   (X(_),XMInfty) -> true
|   (XMInfty,X(_)) -> false
|   (XPInfty,XPInfty) -> true
|   (XMInfty,XMInfty) -> true
|   (XMInfty,XPInfty) -> false
|   (XPInfty,XMInfty) -> true
|   (XBottom,XBottom) -> true
|   (XBottom,_) -> false
|   (_,XBottom) -> true ;;


let rec fixedGt a b = match (a,b) with
	(X(s1,i1,f1),X(s2,i2,f2)) -> 
		(match (s1,s2) with	
			(true,true) -> let (i1',i2') = intAlign i1 i2 in
							if ((String.compare i1' i2')=0) then 
								let (f1',f2') = fracAlign f1 f2 in
									if ((String.compare f1' f2')>0) then true else false
							else
								if ((String.compare i1' i2')>0) then true else false
		|	(false,false) -> fixedGt (X(true,i2,f2)) (X(true,i1,f1))
		|   (true,false) -> true
		|   (false,true) -> false
	)
|   (X(_),XPInfty) -> false
|   (XPInfty,X(_)) -> true
|   (X(_),XMInfty) -> true
|   (XMInfty,X(_)) -> false
|   (XPInfty,XPInfty) -> false
|   (XMInfty,XMInfty) -> false
|   (XMInfty,XPInfty) -> false
|   (XPInfty,XMInfty) -> true
|   (XBottom,XBottom) -> false
|   (XBottom,_) -> false
|   (_,XBottom) -> true ;;


let rec fixedLte a b = fixedGte b a ;;
let rec fixedLt a b = fixedGt b a ;;


let rec fracLt f1 f2 i =
	if (i>=(String.length f1)) then
		if (i>=(String.length f2)) then false
		else true
	else
	if ((String.get f1 i)=(String.get f2 i)) then
		fracLt f1 f2 (i+1)
	else
		if ((String.get f1 i)='0') then true else false ;; 


let rec addFrac f1 f2 ret i res =
	if (i<0) then (res,ret) 
	else
		let x1 = String.get f1 i in
		let x2 = String.get f2 i in
		 (match(x1,x2,ret) with
				('0','0','0') -> addFrac f1 f2 '0' (i-1) ("0"^res)
			|	('1','0','0') -> addFrac f1 f2 '0' (i-1) ("1"^res)
			|	('0','1','0') -> addFrac f1 f2 '0' (i-1) ("1"^res)
			|	('1','1','0') -> addFrac f1 f2 '1' (i-1) ("0"^res)
			|	('0','0','1') -> addFrac f1 f2 '0' (i-1) ("1"^res)
			|	('1','0','1') -> addFrac f1 f2 '1' (i-1) ("0"^res)
			|	('0','1','1') -> addFrac f1 f2 '1' (i-1) ("0"^res)
			|	('1','1','1') -> addFrac f1 f2 '1' (i-1) ("1"^res)
			|   _ -> raise (Error ("addFrac","impossible case in addFrac"))
		) ;;
			
						
let uFixedPlus i1 f1 i2 f2 = 
	let i1' = simplifyZero i1 in  
	let i2' = simplifyZero i2 in  
	let f1' = simplifyZero f1 in  
	let f2' = simplifyZero f2 in  
	let vi1 = int64OfBinString i1' in
	let vi2 = int64OfBinString i2' in
	let vi = Int64.add vi1 vi2 in
	let (f1'',f2'') = fracAlign f1' f2' in
	let (f,ret) = addFrac f1'' f2'' '0' ((String.length f1'')-1) "" in
	let vi' = if (ret='1') then (Int64.add vi Int64.one) else vi in
	let (i,ip) = binStringOfInt64 vi' in	
	let f' = try 
				let z = String.rindex f '1' 
				in String.sub f 0 (z+1)
			 with
				Not_found -> "0"
	in (i,f',ip,String.length f')  ;;
 

let rec subFrac f1 f2 ret i res =
	if (i<0) then (res,ret) 
	else
		let x1 = String.get f1 i in
		let x2 = String.get f2 i in
		 (match(x1,x2,ret) with
				('0','0','0') -> subFrac f1 f2 '0' (i-1) ("0"^res)
			|	('1','0','0') -> subFrac f1 f2 '0' (i-1) ("1"^res)
			|	('0','1','0') -> subFrac f1 f2 '1' (i-1) ("1"^res)
			|	('1','1','0') -> subFrac f1 f2 '0' (i-1) ("0"^res)
			|	('0','0','1') -> subFrac f1 f2 '1' (i-1) ("1"^res)
			|	('1','0','1') -> subFrac f1 f2 '0' (i-1) ("0"^res)
			|	('0','1','1') -> subFrac f1 f2 '1' (i-1) ("0"^res)
			|	('1','1','1') -> subFrac f1 f2 '1' (i-1) ("1"^res)
			|   _ -> raise (Error ("subFrac","impossible case in addFrac"))
		) ;;


let uFixedMinus i1 f1 i2 f2 = (* assumes i1_f1 >= i2_f2 *)
	let i1' = simplifyZero i1 in  
	let i2' = simplifyZero i2 in  
	let f1' = simplifyZero f1 in  
	let f2' = simplifyZero f2 in  
	let vi1 = int64OfBinString i1' in
	let vi2 = int64OfBinString i2' in
	let vi = Int64.sub vi1 vi2 in
	let (f1'',f2'') = fracAlign f1' f2' in
	let (f,ret) = subFrac f1'' f2'' '0' ((String.length f1'')-1) "" in
	let vi' = if (ret='1') then (Int64.sub vi Int64.one) else vi in
	let (i,ip) = binStringOfInt64 vi' in	
	let f' = (try 
				let z = String.rindex f '1' 
				in String.sub f 0 (z+1)
			  with
				Not_found -> "0"
			)
	in (i,f',ip,String.length f') ;;


let rec binMult x1 x2 res i =
	if ((String.length x1)=0) then res
	else
		if ((String.get x1 ((String.length x1)-1))='1') then
			let (a,b) = intAlign res (x2^(String.make i '0')) in
			let (res',ret) = addFrac a b '0' ((String.length a)-1) ""  in
			let res'' = if (ret='1') then ("1"^res') else res'
			in binMult (String.sub x1 0 ((String.length x1)-1)) x2 res'' (i+1)
		else
		  binMult (String.sub x1 0 ((String.length x1)-1)) x2 res (i+1) ;;
		

let uFixedMult i1 f1 i2 f2 = 
	let (i1',i2') = intAlign i1 i2 in
	let (f1',f2') = fracAlign f1 f2 in
	let x1 = i1'^f1' in
	let x2 = i2'^f2' in	
	let y = binMult x1 x2 "0" 0 in
	let l = String.length f1' in
	let f = try 
				String.sub y ((String.length y)-(2*l)) (2*l) 
			with _ -> "0"
	in
	let i = try
				String.sub y 0 ((String.length y)-(2*l)) 
			with _ -> "0"
	in
	let i' = removeLeftmostZeros i in
	let f' = removeRightmostZeros f in
	  (i',f',String.length i',String.length f') 
	;;
	
	
let rec fixedPlus a b iPrec fPrec = match (a,b) with
	(X(s1,i1,f1),X(s2,i2,f2)) -> 
		(match (s1,s2) with
			(true,true) -> 	let (i,f,ip,fp) = uFixedPlus i1 f1 i2 f2 in
							let i'= if ((String.length i)<iPrec) then ((String.make (iPrec-(String.length i)) '0')^i) else i 
						 	in	(X(true,i',String.sub (f^(String.make fPrec '0')) 0 fPrec),(String.length f)>fPrec)
		 | (false,false) -> let (i,f,ip,fp) = uFixedPlus i1 f1 i2 f2 in
							let i'= if ((String.length i)<iPrec) then ((String.make (iPrec-(String.length i)) '0')^i) else i 
							in	(X(false,i',String.sub (f^(String.make fPrec '0')) 0 fPrec),(String.length f)>fPrec)
		 | (true,false) -> fixedMinus a (X(true,i2,f2)) iPrec fPrec
		 | (false,true) -> fixedMinus b (X(true,i1,f1)) iPrec fPrec
		)
| 	(X(_),XPInfty) -> (XPInfty,false)
| 	(XPInfty,X(_)) -> (XPInfty,false)
| 	(X(_),XMInfty) -> (XMInfty,false)
| 	(XMInfty,X(_)) -> (XMInfty,false)
| 	(XPInfty,XPInfty) -> (XPInfty,false)
| 	(XMInfty,XMInfty) -> (XMInfty,false)
| 	(XMInfty,XPInfty) -> (XBottom,false)
| 	(XPInfty,XMInfty) -> (XBottom,false)
| 	(XBottom,_) -> (XBottom,false)
| 	(_,XBottom) -> (XBottom,false)


and fixedMinus a b iPrec fPrec = match (a,b) with
	(X(s1,i1,f1),X(s2,i2,f2)) -> 
		(match (s1,s2) with
			(true,true) -> 	if (fixedGte a b) then
								let (i,f,ip,fp) = uFixedMinus i1 f1 i2 f2 in
								let i'= if ((String.length i)<iPrec) then ((String.make (iPrec-(String.length i)) '0')^i) else i in
 									(X(true,i',String.sub (f^(String.make fPrec '0')) 0 fPrec),(String.length f)>fPrec)
							else
								let (i,f,ip,fp) = uFixedMinus i2 f2 i1 f1 in
								let i'= if ((String.length i)<iPrec) then ((String.make (iPrec-(String.length i)) '0')^i) else i in
									(X(false,i',String.sub (f^(String.make fPrec '0')) 0 fPrec),(String.length f)>fPrec)
		 | (false,false) -> fixedMinus (X(true,i2,f2)) (X(true,i1,f1)) iPrec fPrec
		 | (true,false) -> fixedPlus a (X(true,i2,f2)) iPrec fPrec
		 | (false,true) -> fixedPlus a (X(false,i2,f2)) iPrec fPrec
		)
| 	(X(_),XPInfty) -> (XMInfty,false)
| 	(XPInfty,X(_)) -> (XPInfty,false)
| 	(X(_),XMInfty) -> (XPInfty,false)
| 	(XMInfty,X(_)) -> (XMInfty,false)
| 	(XPInfty,XPInfty) -> (XBottom,false)
| 	(XMInfty,XMInfty) -> (XBottom,false)
| 	(XMInfty,XPInfty) -> (XMInfty,false)
| 	(XPInfty,XMInfty) -> (XPInfty,false)
| 	(XBottom,_) -> (XBottom,false)
| 	(_,XBottom) -> (XBottom,false)
;;


let rec fixedMult a b iPrec fPrec = match (a,b) with
	(X(s1,i1,f1),X(s2,i2,f2)) -> 
		(match (s1,s2) with
			(true,true) -> 	
							let (i,f,ip,fp) = uFixedMult i1 f1 i2 f2 in
							let i'= if ((String.length i)<iPrec) then ((String.make (iPrec-(String.length i)) '0')^i) else i 
						 	in	(X(true,i',String.sub (f^(String.make fPrec '0')) 0 fPrec),(String.length f)>fPrec)
		 | (false,false) -> fixedMult (X(true,i1,f1)) (X(true,i2,f2)) iPrec fPrec
		 | (true,false) -> 	let (i,f,ip,fp) = uFixedMult i1 f1 i2 f2 in
							let i'= if ((String.length i)<iPrec) then ((String.make (iPrec-(String.length i)) '0')^i) else i 
						 	in	(X(false,i',String.sub (f^(String.make fPrec '0')) 0 fPrec),(String.length f)>fPrec)
		 | (false,true) -> fixedMult (X(true,i2,f2)) (X(false,i1,f1)) iPrec fPrec
		)
| 	(X(s,_,_),XPInfty) -> if s then (XPInfty,false) else (XMInfty,false)
| 	(XPInfty,X(s,_,_)) -> if s then (XPInfty,false) else (XMInfty,false)
| 	(X(s,_,_),XMInfty) -> if s then (XMInfty,false) else (XPInfty,false)
| 	(XMInfty,X(s,_,_)) -> if s then (XMInfty,false) else (XPInfty,false)
| 	(XPInfty,XPInfty) -> (XPInfty,false)
| 	(XMInfty,XMInfty) -> (XPInfty,false)
| 	(XMInfty,XPInfty) -> (XMInfty,false)
| 	(XPInfty,XMInfty) -> (XMInfty,false)
| 	(XBottom,_) -> (XBottom,false)
| 	(_,XBottom) -> (XBottom,false) 
;;


let fixedDiv a b fp = raise (Error ("fixedDiv","??? Fixed-point division not implemented.")) ;;
			
			
let fixedSqrt a fp = raise (Error ("fixedSqrt","??? Fixed-point square root not implemented.")) ;;	 

					
let fixedNeg a = match a with
	X(s,i,f) -> X(not s,i,f)
| 	XPInfty -> XMInfty
| 	XMInfty -> XPInfty
|   XBottom -> XBottom ;;
					 

let fixedAbs a = match a with
  	X(s,i,f) -> X(true,i,f)
| 	XPInfty -> XPInfty
| 	XMInfty -> XPInfty
|   XBottom -> XBottom ;;


let fixedMin a b = match (a,b) with
  (X(_,_,_),X(_,_,_)) -> if (fixedLt a b) then a else b 
| (X(_),XPInfty) -> a
| (XPInfty,X(_)) -> b
| (X(_),XMInfty) -> XMInfty
| (XMInfty,X(_)) -> XMInfty
| (XPInfty,XPInfty) -> XPInfty
| (XMInfty,XMInfty) -> XMInfty
| (XMInfty,XPInfty) -> XMInfty
| (XPInfty,XMInfty) -> XMInfty
| (XBottom,_) -> a
| (_,XBottom) -> b ;;


let fixedMax a b = match (a,b) with
  (X(_,_,_),X(_,_,_)) -> if (fixedGt a b) then a else b 
| (X(_),XPInfty) -> XPInfty
| (XPInfty,X(_)) -> XPInfty
| (X(_),XMInfty) -> a
| (XMInfty,X(_)) -> b
| (XPInfty,XPInfty) -> XPInfty
| (XMInfty,XMInfty) -> XMInfty
| (XMInfty,XPInfty) -> XPInfty
| (XPInfty,XMInfty) -> XPInfty
| (XBottom,_) -> b
| (_,XBottom) -> a ;;


let fixedMin4 a b c d = fixedMin (fixedMin a b) (fixedMin c d) ;;
let fixedMax4 a b c d = fixedMax (fixedMax a b) (fixedMax c d) ;;


let rec fixedPowAux z n ip fp res ovf =
  if (n=0) then  
    (res,ovf)
  else
    match z with
    	 X(s,i,f) -> if (n>0) then 
						let (x,o) = fixedMult z res ip fp
						in fixedPowAux z (n-1) ip fp x (o||ovf)  
				 	 else
						let (x,o) = fixedDiv res z ip 
						in fixedPowAux z (n+1) ip fp x (o||ovf)
		| _ -> raise (Error ("mpFloatPow","unexpected value: infinite or bottom")) ;;

		
let fixedPow z n ip fp = match z with
			X(_) ->	fixedPowAux z n ip fp (xOne 1 1) false 
		|   XPInfty -> (XPInfty,true)
		|   XMInfty -> if ((n mod 2)=0) then (XPInfty,true) else (XMInfty,true)
		|   XBottom -> (XBottom,false) ;;


(*****************************************************************************)
(*                     operations on Fixed point intervals                   *)
(*****************************************************************************)


let fixedMinO (a,oA) (b,oB) = match (a,b) with
  (X(_,_,_),X(_,_,_)) -> if (fixedLt a b) then (a,oA) else (b,oB) 
| (X(_),XPInfty) -> (a,oA)
| (XPInfty,X(_)) -> (b,oB)
| (X(_),XMInfty) -> (XMInfty,false)
| (XMInfty,X(_)) -> (XMInfty,false)
| (XPInfty,XPInfty) -> (XPInfty,false)
| (XMInfty,XMInfty) -> (XMInfty,false)
| (XMInfty,XPInfty) -> (XMInfty,false)
| (XPInfty,XMInfty) -> (XMInfty,false)
| (XBottom,_) -> (b,oB)
| (_,XBottom) -> (a,oA) ;;


let fixedMaxO (a,oA) (b,oB) = match (a,b) with
  (X(_,_,_),X(_,_,_)) -> if (fixedGt a b) then (a,oA) else (b,oB) 
| (X(_),XPInfty) -> (XPInfty,false)
| (XPInfty,X(_)) -> (XPInfty,false)
| (X(_),XMInfty) -> (a,oA)
| (XMInfty,X(_)) -> (b,oB)
| (XPInfty,XPInfty) -> (XPInfty,false)
| (XMInfty,XMInfty) -> (XMInfty,false)
| (XMInfty,XPInfty) -> (XPInfty,false)
| (XPInfty,XMInfty) -> (XPInfty,false)
| (XBottom,_) -> (b,oB)
| (_,XBottom) -> (a,oA) ;;


let fixedMinO4 (a,oA) (b,oB) (c,oC) (d,oD) = fixedMinO (fixedMinO (a,oA) (b,oB)) (fixedMinO (c,oC) (d,oD)) ;;
let fixedMaxO4 (a,oA) (b,oB) (c,oC) (d,oD) = fixedMaxO (fixedMaxO (a,oA) (b,oB)) (fixedMaxO (c,oC) (d,oD)) ;;


let fixedIntMult (v0Min,v0Max) (v1Min,v1Max) iPrec fPrec = 
    let (f1,o1) = fixedMult v0Min v1Min iPrec fPrec in 
    let (f2,o2) = fixedMult v0Min v1Max iPrec fPrec in
    let (f3,o3) = fixedMult v0Max v1Min iPrec fPrec in 
    let (f4,o4) = fixedMult v0Max v1Max iPrec fPrec in
      (fixedMinO4 (f1,o1) (f2,o2) (f3,o3) (f4,o4),fixedMaxO4 (f1,o1) (f2,o2) (f3,o3) (f4,o4)) ;;


let fixedIntPow (a,b) n ip fp =
	if (((n mod 2)=0)  && (fixedLt a (xZero 1 1))
					   && (fixedGte b (xZero 1 1))) then  
					let (a',ovfA) = fixedPow a n ip fp in
					let (b',ovfB) = fixedPow b n ip fp in
					let c = fixedMax a' b' in
						(match c with
							XBottom ->	((XBottom,XBottom),false)
						|	_ ->		((xZero 1 fp,c),ovfA||ovfB)
						)
				else
					let (a',ovfA) = fixedPow a n ip fp in
					let (b',ovfB) = fixedPow b n ip fp in
						((fixedMin a' b',fixedMax a' b'),ovfA||ovfB) ;; 


let fixedIntEq (v0Min,v0Max) (v1Min,v1Max) = 
    if ((fixedEq v0Min v0Max)&&(fixedEq v1Min v1Max)&&(fixedEq v0Min v1Max)) then 
      B(true) 
	else 
		if ((((fixedLte v0Min v1Min)&&(fixedLte v1Min v0Max)) ||
 		    ((fixedLte v0Min v1Max)&&(fixedLte v1Max v0Max))) 
			||
			(((fixedLte v1Min v0Min)&&(fixedLte v0Min v1Max)) ||
 		    ((fixedLte v1Min v0Max)&&(fixedLte v0Max v1Max))) 
			)
			then BTop
		else
			B(false)
    ;;


let fixedIntLt' v0Min v0Max v1Min v1Max = 
	if (fixedLt v0Max v1Min) then B(true) else
	  if (fixedLte v1Max v0Min) then B(false) else BTop ;;


let fixedIntLt (v0Min,v0Max) (v1Min,v1Max) = (fixedIntLt' v0Min v0Max v1Min v1Max) ;;


let fixedIntGt (v0Min,v0Max) (v1Min,v1Max) = fixedIntLt' v1Min v1Max v0Min v0Max ;; 


let fixedIntLte' v0Min v0Max v1Min v1Max = 
	if (fixedLte v0Max v1Min) then B(true) else
	  if (fixedLt v1Max v0Min) then B(false) else BTop ;;


let fixedIntLte (v0Min,v0Max) (v1Min,v1Max) = (fixedIntLte' v0Min v0Max v1Min v1Max) ;;


let fixedIntGte (v0Min,v0Max) (v1Min,v1Max) = (fixedIntLte' v1Min v1Max v0Min v0Max) ;;


let fixedIntNeq (v0Min,v0Max) (v1Min,v1Max) = (
	 if ((fixedLt v0Max v1Min)||(fixedLt v1Max v0Min)) then B(true) else
	 if ((fixedEq v0Min v0Max)&&(fixedEq v1Min v1Max)&&(not (fixedEq v0Min v1Min))) 
	 then B(false) 
	 else BTop 
	) ;;


let fixedIntJoin a b = match (a,b) with 
  ((XBottom,XBottom),_) -> b
| (_,(XBottom,XBottom)) -> a
| ((v0Min,v0Max),(v1Min,v1Max)) -> ((fixedMin v0Min v1Min,fixedMax v0Max v1Max)) ;;


let fixedIntMeet a b = match (a,b) with 
  ((v0Min,v0Max),(v1Min,v1Max)) -> 
    let (x,y) = (fixedMax v0Min v1Min,fixedMin v0Max v1Max) 
    in if (fixedGt x y) then (XBottom,XBottom) else (x,y) ;;


let fixedIntSseq a b = match (a,b) with 
  ((v0Min,v0Max),(v1Min,v1Max)) -> ((fixedGte v0Min v1Min)&&(fixedLte v0Max v1Max)) ;;


(*****************************************************************************)
(*            operations on Fixed point numbers with errors                  *)
(*****************************************************************************)


let fiPlus (FX((x1,y1),ip1,fp1,e1)) (FX((x2,y2),ip2,fp2,e2)) = 
	let fp = max fp1 fp2 in
	let ip = max ip1 ip2 in
	let (z1,trunced1) = fixedPlus x1 x2 ip fp in
	let (z2,trunced2) = fixedPlus y1 y2 ip fp in
	let err1 =	if trunced1 then
		 			mpFloatNeg (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec) !errPrec GMP_RNDN 
				else 
					(M(FR.from_float_prec !errPrec GMP_RNDN 0.0))
 	in
	let err2 = 	if trunced2 then 
					pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec 	
				else
					(M(FR.from_float_prec !errPrec GMP_RNDN 0.0))
	in
	let e = mpfiPlus (mpfiPlus e1 e2 !errPrec OUTSIDE) (err1,err2) !errPrec OUTSIDE in
	let ip' = max (fxIntegerLength z1) (fxIntegerLength z2) in
	let res = FX((z1,z2),ip',fp,e) 
	in resizeFX res ip' fp ;;


let fiMinus (FX((x1,y1),ip1,fp1,e1)) (FX((x2,y2),ip2,fp2,e2)) =
	let fp = max fp1 fp2 in
	let ip = max ip1 ip2 in
	let (z1,trunced1) = fixedMinus x1 y2 ip fp in
	let (z2,trunced2) = fixedMinus y1 x2 ip fp in
	let err1 =	if trunced1 then
		 			mpFloatNeg (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec) !errPrec GMP_RNDN 
				else 
					(M(FR.from_float_prec !errPrec GMP_RNDN 0.0))
 	in
	let err2 = 	if trunced2 then 
					pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec 	
				else
					(M(FR.from_float_prec !errPrec GMP_RNDN 0.0))
	in
	let e = mpfiPlus (mpfiMinus e1 e2 !errPrec OUTSIDE) (err1,err2) !errPrec OUTSIDE in
	let ip' = max (fxIntegerLength z1) (fxIntegerLength z2) in
	let res = FX((z1,z2),ip',fp,e) 
	in resizeFX res ip' fp ;;
	

let fiNeg (FX((x,y),ip,fp,e))  = 
	let x' = fixedNeg x in
	let y' = fixedNeg y in 
  		(FX((y',x'),ip,fp,mpfiNeg e !errPrec OUTSIDE)) ;;
	
	
let fiAbs (FX((x,y),ip,fp,e))  = 
	let sx = getFixedSign x in
	let sy = getFixedSign y in
		if (sx=sy) then
			let x' = fixedAbs x in
			let y' = fixedAbs y in 
			  	if (fixedGt x' y') then 
					(FX((y',x'),ip,fp,e)) 
				else
					(FX((x',y'),ip,fp,e)) 
		else
			let x' = fixedAbs x in
			let y' = fixedAbs y in 
			  	if (fixedGt x' y') then 
					(FX((xZero ip fp,x'),ip,fp,e)) 
				else
					(FX((xZero ip fp,y'),ip,fp,e)) ;;		


let fiMult (FX((x1,y1),ip1,fp1,e1)) (FX((x2,y2),ip2,fp2,e2)) = 
	let ip = max ip1 ip2 in
	let fp = max fp1 fp2 in
	let (x1',_) = resizeX x1 ip1 fp1 ip fp in 
	let (x2',_) = resizeX x2 ip2 fp2 ip fp in
	let (y1',_) = resizeX y1 ip1 fp1 ip fp in
	let (y2',_) = resizeX y2 ip2 fp2 ip fp in
	let ((z1,trunced1),(z2,trunced2)) = fixedIntMult (x1',y1') (x2',y2') ip fp in
	let err1 =	if trunced1 then
		 			mpFloatNeg (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec) !errPrec GMP_RNDN 
				else 
					(M(FR.from_float_prec !errPrec GMP_RNDN 0.0))
 	in
	let err2 = 	if trunced2 then 
					pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec 	
				else
					(M(FR.from_float_prec !errPrec GMP_RNDN 0.0))
	in
	let v1 = (mpFloatOfFixed x1 (ip1+fp1),mpFloatOfFixed y1 (ip1+fp1)) in
	let v2 = (mpFloatOfFixed x2 (ip2+fp2),mpFloatOfFixed y2 (ip2+fp2)) in
	let e = mpfiPlus
					(mpfiPlus (mpfiPlus (mpfiMult v1 e2 !errPrec OUTSIDE) (mpfiMult v2 e1 !errPrec OUTSIDE) !errPrec OUTSIDE)
			        		  (mpfiMult e1 e2 !errPrec OUTSIDE) !errPrec OUTSIDE)
					(err1,err2)
					!errPrec OUTSIDE
	in
	let ip' = max (fxIntegerLength z1) (fxIntegerLength z2) in
	let res = FX((z1,z2),ip',fp,e) 
	in resizeFX res ip' fp ;;
					

let fiJoin (FX(f1,ip1,fp1,e1)) (FX(f2,ip2,fp2,e2)) = 
    let f = fixedIntJoin f1 f2 in
    let e = mpfiJoin e1 e2 in FX(f,max ip1 ip2,max fp1 fp2,e) ;;

	
let fiMeet (FX(f1,ip1,fp1,e1)) (FX(f2,ip2,fp2,e2)) =
    let f = fixedIntMeet f1 f2 in
    let e = mpfiMeet e1 e2 in FX(f,max ip1 ip2,max fp1 fp2,e) ;;
				
				
let fiSseq (FX(f1,_,_,e1)) (FX(f2,_,_,e2)) =
  if ((fixedIntSseq f1 f2) && (mpfiSseq e1 e2)) then true else false ;;


let fiSseqNeglectErrors (FX(f1,_,_,_)) (FX(f2,_,_,_)) =
	 if (fixedIntSseq f1 f2) then true else false ;;
				
				
let fiPow (FX((x,y),ip,fp,e)) n =
	let ((x',y'),trunced) = fixedIntPow (x,y) n ip fp in
	let err = 	if (trunced&&(isXNonNegative x)) then 
					((M(FR.from_float_prec !errPrec GMP_RNDN 0.0)),(pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec))	
				else
					if (trunced&&(isXNonPositive x)&&((n mod 2)=1)) then 
						((mpFloatNeg (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec) !errPrec GMP_RNDN),
					   	 (M(FR.from_float_prec !errPrec GMP_RNDN 0.0)))
					else
						((mpFloatNeg (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec) !errPrec GMP_RNDN),
						 (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 1.0)) (-fp-1) !errPrec))
	in
	let ip' = max (fxIntegerLength x') (fxIntegerLength y') in
	let res = FX((x',y'),ip',fp,err) 
	in resizeFX res ip' fp ;;


let rec fiPowIntervalAux x (m,n) res = 
	if (m>n) then res
	else (fiPowIntervalAux x (m+1,n) (fiJoin (fiPow x m) res)) ;;
	

let fiPowInterval x ip fp (m,n) = fiPowIntervalAux x (m,n) (FX((XBottom,XBottom),ip,fp,(MBottom,MBottom))) ;;

(************************)
		
			

let rec isStringXZero	x = 
	if ((String.length x)=0) then true
	else
		let c = (String.get x 0)
		in if ((c='0')||(c='_')) then
				(isStringXZero (String.sub x 1 ((String.length x)-1)))
	 	   else false ;;


let fixedOfStringAux x prec =
	let (x',s) = if ((String.get x 0)='-') then
					(String.sub x 1 ((String.length x)-1),false)
				 else (x,true)
	in
	let j = try String.index x' '_' with Not_found -> (-1) in
	let x'' = if (j>=0) then ((String.sub x' 0 j)^"."^(String.sub x' (j+1) ((String.length x')-j-1))) else x' in
	let f = FR.from_string_prec_base float64_mantissa GMP_RNDN 10 x'' in
	let f2 = FR.to_string_base_digits GMP_RNDN 2 float64_mantissa f in
	let ei = try String.rindex f2 'E' with Not_found -> (1) in
	let f3 = String.sub f2 0 ei in
	let e = if (ei>=0) then 
				int_of_string (String.sub f2 (ei+1) ((String.length f2)-ei-1)) 
			else 0
	in
	let i = try 
				String.index f3 '.' 
			with 
				Not_found -> 0 
	in		
		if (i>0) then
			let fs = String.sub f3 (i+1) ((String.length f3)-(i+1)) in
				if (e>=0) then
					let is' = "1"^(String.sub fs 0 e) in
					let fs' = String.sub fs e ((String.length fs)-e) in
					let fs'' = removeRightmostZeros fs' in
					let fs''' = if ((String.length fs'')>prec) then
									String.sub fs'' 0 prec
							    else fs''
					in (X(s,is',fs'''),String.length is',String.length fs''') 
				else
					let is' = "0" in
					let fs' = (String.make (-e-1) '0')^"1" in
					let fs'' = removeRightmostZeros (fs'^fs) in
					let fs''' = if ((String.length fs'')>prec) then
									String.sub fs'' 0 prec
							    else fs''
					in (X(s,is',fs'''),String.length is',String.length fs''') 
	    else
			(X(s,f3,"0"),String.length f3,1) ;;


let fixedOfString x = 
	try 
		let i = String.index x '#' in
		let x' = String.sub x 0 i in		
		if (isStringXZero x') then 
			((X(true,"0","0"),1,1),!eZero)
		else	
			let k = int_of_string (String.sub x (i+1) ((String.length x)-i-1)) in
			let (f,ip,fp) = fixedOfStringAux x' 999 in
			let (f',_,_) = fixedOfStringAux x' k in			
			let (_,e) = resizeX f ip fp ip k in
				((f',ip,k),e)
	with
		Not_found -> 
			if (isStringXZero x) then 
				((X(true,"0","0"),1,1),!eZero)
			else
				let (f,ip,fp) = fixedOfStringAux x !fixed_default_frac_size in
				let (_,_,fp') = fixedOfStringAux x (2*(!fixed_default_frac_size)) in
				let e = if (fp'>(!fixed_default_frac_size)) then
							if (isXNonNegative f) then
								(M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0),
								 pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 2.0)) (-fp) !errPrec)				
							else
								((mpFloatMult (M(FR.from_float_prec !errPrec GMP_RNDN (-1.0))) 
											 (pow2M (M(FR.from_float_prec !errPrec GMP_RNDN 2.0)) (-fp) !errPrec)) !errPrec GMP_RNDN,
								 M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0))				
					  	else !eZero
				in	((f,ip,fp),e) ;;

