(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file float.ml contains the domain and operations on floating-point values
 *)

open Gmp ;;
open Types ;;
open Prelude ;;

(*****************************************************************************)
(*                      operations on MPFR Numbers                           *)
(*****************************************************************************)


let mpFloatPlus a b prec r = match (a,b) with
  (M(x),M(y)) ->  M(FR.add_prec prec r x y)
| (M(x),MPInfty) -> MPInfty
| (MPInfty,M(y)) -> MPInfty
| (M(x),MMInfty) -> MMInfty
| (MMInfty,M(y)) -> MMInfty
| (MPInfty,MPInfty) -> MPInfty
| (MMInfty,MMInfty) -> MMInfty
| (MMInfty,MPInfty) -> MBottom
| (MPInfty,MMInfty) -> MBottom
| (MBottom,_) -> MBottom
| (_,MBottom) -> MBottom ;;


let mpFloatMinus a b prec r = match (a,b) with
  (M(x),M(y)) ->  M(FR.sub_prec prec r x y)
| (M(x),MPInfty) -> MMInfty
| (MPInfty,M(y)) -> MPInfty
| (M(x),MMInfty) -> MPInfty
| (MMInfty,M(y)) -> MMInfty
| (MPInfty,MPInfty) -> MBottom
| (MMInfty,MMInfty) -> MBottom
| (MMInfty,MPInfty) -> MMInfty
| (MPInfty,MMInfty) -> MPInfty
| (MBottom,_) -> MBottom
| (_,MBottom) -> MBottom ;;


let mpFloatMult a b prec r = match (a,b) with
  (M(x),M(y)) ->  M(FR.mul_prec prec r x y)
| (M(x),MPInfty) -> if ((FR.cmp x (FR.from_float 0.0))>0) then MPInfty (* if (x>0) ... *)
                    else if ((FR.cmp x (FR.from_float 0.0))=0) then M(FR.from_float 0.0) (* if (x=0) ... *)
                         else MMInfty 
| (MPInfty,M(y)) -> if ((FR.cmp y (FR.from_float 0.0))>0) then MPInfty (* if (y>0) ... *)
                    else if ((FR.cmp y (FR.from_float 0.0))=0) then M(FR.from_float 0.0) (* if (y=0) ... *)
                         else  MMInfty 
| (M(x),MMInfty) -> if ((FR.cmp x (FR.from_float 0.0))>0) then MMInfty (* if (x>0) ... *)
                    else if ((FR.cmp x (FR.from_float 0.0))=0) then M(FR.from_float 0.0) (* if (x=0) ... *)
                         else MPInfty 
| (MMInfty,M(y)) -> if ((FR.cmp y (FR.from_float 0.0))>0) then MMInfty (* if (y>0) ... *)
                    else if ((FR.cmp y (FR.from_float 0.0))=0) then M(FR.from_float 0.0) (* if (y=0) ... *)
                         else  MPInfty 
| (MPInfty,MPInfty) -> MPInfty
| (MMInfty,MMInfty) -> MPInfty
| (MMInfty,MPInfty) -> MMInfty
| (MPInfty,MMInfty) -> MMInfty
| (MBottom,_) -> MBottom
| (_,MBottom) -> MBottom ;;


let mpFloatDiv a b prec r = match (a,b) with
  (M(x),M(y)) ->  M(FR.div_prec prec r x y)
| (M(x),MPInfty) -> M(FR.from_float 0.0) 
| (MPInfty,M(y)) -> if ((FR.cmp y (FR.from_float 0.0))>0) then MPInfty (* if (y>0) ... *)
                    else if ((FR.cmp y (FR.from_float 0.0))=0) then MBottom (* if (y=0) ... *)
                         else  MMInfty 
| (M(x),MMInfty) -> M(FR.from_float 0.0) 
| (MMInfty,M(y)) -> if ((FR.cmp y (FR.from_float 0.0))>0) then MMInfty (* if (y>0) ... *)
                    else if ((FR.cmp y (FR.from_float 0.0))=0) then MBottom (* if (y=0) ... *)
                         else  MPInfty 
| (MPInfty,MPInfty) -> MBottom
| (MMInfty,MMInfty) -> MBottom
| (MMInfty,MPInfty) -> MBottom
| (MPInfty,MMInfty) -> MBottom
| (MBottom,_) -> MBottom
| (_,MBottom) -> MBottom ;;


let mpFloatAbs x prec r = match x with
  M(x) -> M(FR.abs_prec prec r x)
| MPInfty -> MPInfty
| MMInfty -> MPInfty
| MBottom -> MBottom ;;


let mpFloatNeg x prec r = match x with
  M(x) -> M(FR.neg_prec prec r x)
| MPInfty -> MMInfty
| MMInfty -> MPInfty
| MBottom -> MBottom ;;



let rec pow2M x n prec = (* pow2M computes x*(2^n) *) 
  if (n>0) then mpFloatMult (M((FR.from_float_prec prec GMP_RNDN 2.0))) (pow2M x (n-1) prec) prec GMP_RNDU else
  if (n<0) then mpFloatMult (pow2M x (n+1) prec) (M((FR.from_float_prec prec GMP_RNDN 0.5))) prec GMP_RNDU else x ;;

  
let rec mpFloatPowAux z n prec mode res =
  if (n=0) then  
    res
  else
    match z with
    	 M(x) -> if (n>0) then mpFloatPowAux z (n-1) prec mode (mpFloatMult (M(x)) res prec mode) 
				 else
					mpFloatPowAux z (n+1) prec mode (mpFloatDiv res (M(x)) prec mode)
		| _ -> raise (Error ("mpFloatPow","unexpected value: infinite or bottom")) ;;

		
let mpFloatPow z n prec mode = match z with
	M(_) ->	mpFloatPowAux z n prec mode (M(FR.from_float_prec prec GMP_RNDN 1.0)) 
|   MPInfty -> MPInfty
|   MMInfty -> if ((n mod 2)=0) then MPInfty else MMInfty
|   MBottom -> MBottom ;;
  

let mpFloat2Int32 f = match f with
  M(x) -> I32(Int32.of_float (FR.to_float x))
| MPInfty -> IPInfty
| MMInfty -> IMInfty
| MBottom -> IBottom ;;


let mpFloat2Int64 f = match f with
  M(x) -> I64(Int64.of_float (FR.to_float x))
| MPInfty -> IPInfty
| MMInfty -> IMInfty
| MBottom -> IBottom ;;


let integer2MpFloat i = match i with
  I32(x) -> M(FR.from_float (Int32.to_float x))
| I64(x) -> M(FR.from_float (Int64.to_float x))
| IPInfty -> MPInfty
| IMInfty -> MMInfty
| IBottom -> MBottom ;;


let mpFloatMin a b = match (a,b) with
  (M(x),M(y)) ->  if ((FR.cmp x y)<0) then a else b (* if (x<y) ... *)
| (M(x),MPInfty) -> a
| (MPInfty,M(y)) -> b
| (M(x),MMInfty) -> MMInfty
| (MMInfty,M(y)) -> MMInfty
| (MPInfty,MPInfty) -> MPInfty
| (MMInfty,MMInfty) -> MMInfty
| (MMInfty,MPInfty) -> MMInfty
| (MPInfty,MMInfty) -> MMInfty
| (MBottom,_) -> b
| (_,MBottom) -> a ;;


let mpFloatMax a b = match (a,b) with
  (M(x),M(y)) -> if ((FR.cmp x y)>0) then a else b (* if (x>y) ... *)
| (M(x),MPInfty) -> MPInfty
| (MPInfty,M(y)) -> MPInfty
| (M(x),MMInfty) -> a
| (MMInfty,M(y)) -> b
| (MPInfty,MPInfty) -> MPInfty
| (MMInfty,MMInfty) -> MMInfty
| (MMInfty,MPInfty) -> MPInfty
| (MPInfty,MMInfty) -> MPInfty
| (MBottom,_) -> b
| (_,MBottom) -> a ;;


let mpFloatMin4 a b c d = mpFloatMin (mpFloatMin a b) (mpFloatMin c d) ;;
let mpFloatMax4 a b c d = mpFloatMax (mpFloatMax a b) (mpFloatMax c d) ;;


let mpFloatLt a b = match (a,b) with
  (M(x),M(y)) -> if ((FR.cmp x y)<0) then true else false (* if (x<y) ... *)
| (M(x),MPInfty) -> true
| (MPInfty,M(y)) -> false
| (M(x),MMInfty) -> false
| (MMInfty,M(y)) -> true
| (MPInfty,MPInfty) -> false
| (MMInfty,MMInfty) -> false
| (MMInfty,MPInfty) -> true
| (MPInfty,MMInfty) -> false
| (MBottom,MBottom) -> false
| (MBottom,_) -> true
| (_,MBottom) -> false ;;


let mpFloatGt a b = match (a,b) with
  (M(x),M(y)) -> if ((FR.cmp x y)>0) then true else false (* if (x>y) ... *)
| (M(x),MPInfty) -> false
| (MPInfty,M(y)) -> true
| (M(x),MMInfty) -> true
| (MMInfty,M(y)) -> false
| (MPInfty,MPInfty) -> false
| (MMInfty,MMInfty) -> false
| (MMInfty,MPInfty) -> false
| (MPInfty,MMInfty) -> true
| (MBottom,MBottom) -> false
| (MBottom,_) -> false
| (_,MBottom) -> true ;;


let mpFloatLte a b = match (a,b) with
  (M(x),M(y)) -> if ((FR.cmp x y)<=0) then true else false (* if (x<=y) ... *)
| (M(x),MPInfty) -> true
| (MPInfty,M(y)) -> false
| (M(x),MMInfty) -> false
| (MMInfty,M(y)) -> true
| (MPInfty,MPInfty) -> true
| (MMInfty,MMInfty) -> true
| (MMInfty,MPInfty) -> true
| (MPInfty,MMInfty) -> false
| (MBottom,MBottom) -> true
| (MBottom,_) -> true
| (_,MBottom) -> false ;;


let mpFloatGte a b = match (a,b) with
  (M(x),M(y)) -> if ((FR.cmp x y)>=0) then true else false (* if (x>=y) ... *)
| (M(x),MPInfty) -> false
| (MPInfty,M(y)) -> true
| (M(x),MMInfty) -> true
| (MMInfty,M(y)) -> false
| (MPInfty,MPInfty) -> true
| (MMInfty,MMInfty) -> true
| (MMInfty,MPInfty) -> false
| (MPInfty,MMInfty) -> true
| (MBottom,MBottom) -> true
| (MBottom,_) -> false
| (_,MBottom) -> true ;;


let mpFloatEq a b = match (a,b) with
  (M(x),M(y)) -> if ((FR.cmp x y)=0) then true else false (* if (x=y) ... *)
| (M(x),MPInfty) -> false
| (MPInfty,M(y)) -> false
| (M(x),MMInfty) -> false
| (MMInfty,M(y)) -> false
| (MPInfty,MPInfty) -> true
| (MMInfty,MMInfty) -> true
| (MMInfty,MPInfty) -> false
| (MPInfty,MMInfty) -> false
| (MBottom,MBottom) -> true
| (MBottom,_) -> false
| (_,MBottom) -> false ;;


let initError err = match err with
	(M(a),M(b)) -> 	let ulp = mpFloatMinus (M(b)) (M(a)) !errPrec GMP_RNDU 
					in (mpFloatMult (M(FR.from_float_prec !errPrec GMP_RNDN (-0.5))) ulp !errPrec  GMP_RNDD,
						mpFloatMult (M(FR.from_float_prec !errPrec GMP_RNDN (0.5))) ulp !errPrec GMP_RNDU) 
|	_ -> raise (Error ("initError","unexpected case in initerror")) ;;		


let mpFloatSuccessor   f p = mpFloatPlus  f (pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (-p-1) p) p GMP_RNDU ;;
let mpFloatPredecessor f p = mpFloatMinus f (pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (-p-1) p) p GMP_RNDD ;;


(*****************************************************************************)
(*                      operations on MPFR Intervals                         *)
(*****************************************************************************)


let mpfiPlus (v0Min,v0Max) (v1Min,v1Max) prec mode = match mode with
  NEAREST -> ((mpFloatPlus v0Min v1Min prec GMP_RNDN,mpFloatPlus v0Max v1Max prec GMP_RNDN)) 
| OUTSIDE -> ((mpFloatPlus v0Min v1Min prec GMP_RNDD,mpFloatPlus v0Max v1Max prec GMP_RNDU)) ;;


let mpfiMinus (v0Min,v0Max) (v1Min,v1Max) prec mode = match mode with
  NEAREST -> ((mpFloatMinus v0Min v1Max prec GMP_RNDN,mpFloatMinus v0Max v1Min prec GMP_RNDN)) 
| OUTSIDE -> ((mpFloatMinus v0Min v1Max prec GMP_RNDD,mpFloatMinus v0Max v1Min prec GMP_RNDU)) ;;


let mpfiMult (v0Min,v0Max) (v1Min,v1Max) prec mode = match mode with
  NEAREST -> 
    let f1 = (mpFloatMult v0Min v1Min prec GMP_RNDN) in 
    let f2 = (mpFloatMult v0Min v1Max prec GMP_RNDN) in
    let f3 = (mpFloatMult v0Max v1Min prec GMP_RNDN) in 
    let f4 = (mpFloatMult v0Max v1Max prec GMP_RNDN) in
      ((mpFloatMin4 f1 f2 f3 f4,mpFloatMax4 f1 f2 f3 f4)) 
| OUTSIDE -> 
    let f1 = (mpFloatMult v0Min v1Min prec GMP_RNDD) in 
    let f2 = (mpFloatMult v0Min v1Max prec GMP_RNDD) in
    let f3 = (mpFloatMult v0Max v1Min prec GMP_RNDD) in 
    let f4 = (mpFloatMult v0Max v1Max prec GMP_RNDD) in
    let g1 = (mpFloatMult v0Min v1Min prec GMP_RNDU) in 
    let g2 = (mpFloatMult v0Min v1Max prec GMP_RNDU) in
    let g3 = (mpFloatMult v0Max v1Min prec GMP_RNDU) in 
    let g4 = (mpFloatMult v0Max v1Max prec GMP_RNDU) in
      ((mpFloatMin4 f1 f2 f3 f4,mpFloatMax4 g1 g2 g3 g4)) ;;


let mpfiDiv (v0Min,v0Max) (v1Min,v1Max) prec mode = match mode with
  NEAREST -> 
    let f1 = (mpFloatDiv v0Min v1Min prec GMP_RNDN) in 
    let f2 = (mpFloatDiv v0Min v1Max prec GMP_RNDN) in
    let f3 = (mpFloatDiv v0Max v1Min prec GMP_RNDN) in 
    let f4 = (mpFloatDiv v0Max v1Max prec GMP_RNDN) in
    let (f5,f6) = ((mpFloatMin4 f1 f2 f3 f4,mpFloatMax4 f1 f2 f3 f4)) in
	(* matlab returns +oo or -oo for divisions by zero *)
	let z = M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0) in
	let _ = if (((mpFloatLte v1Min z)&&(mpFloatGte v1Max z)) && (not !warnMsgDivZero)) then
					(warnMsgDivZero := true ;
					 print_string ("\n!!! Warning: result includes division by zero.")
					)
			else () 
	in
		if ((mpFloatLte v1Min z)&&(mpFloatGte v1Max z)) then
				if (mpFloatGte v0Min z) then
					(f5,MPInfty)
				else
					if (mpFloatLte v1Min z) then 
						(MMInfty,f6)
					else (MMInfty,MPInfty)
		else (f5,f6)
| OUTSIDE -> 
    let f1 = (mpFloatDiv v0Min v1Min prec GMP_RNDD) in 
    let f2 = (mpFloatDiv v0Min v1Max prec GMP_RNDD) in
    let f3 = (mpFloatDiv v0Max v1Min prec GMP_RNDD) in 
    let f4 = (mpFloatDiv v0Max v1Max prec GMP_RNDD) in
    let g1 = (mpFloatDiv v0Min v1Min prec GMP_RNDU) in 
    let g2 = (mpFloatDiv v0Min v1Max prec GMP_RNDU) in
    let g3 = (mpFloatDiv v0Max v1Min prec GMP_RNDU) in 
    let g4 = (mpFloatDiv v0Max v1Max prec GMP_RNDU) in
    let (f5,f6) = ((mpFloatMin4 f1 f2 f3 f4,mpFloatMax4 g1 g2 g3 g4)) in
	(* matlab returns +oo or -oo for divisions by zero *)	
	let z = M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0) 
	in
		if ((mpFloatLte v1Min z)&&(mpFloatGte v1Max z)) then
				if (mpFloatGte v0Min z) then
					(f5,MPInfty)
				else
					if (mpFloatLte v1Min z) then 
						(MMInfty,f6)
					else (MMInfty,MPInfty)
		else (f5,f6) ;;


	  
let mpfiAbs (a,b) prec mode = match mode with
  NEAREST -> 
    if (mpFloatLt b (M(FR.from_float_prec prec GMP_RNDN 0.0))) then
      (mpFloatAbs b prec GMP_RNDN,mpFloatAbs a prec GMP_RNDN)
    else
    if ((mpFloatLt a (M(FR.from_float_prec prec GMP_RNDN 0.0)))&&
       (mpFloatGte b (M(FR.from_float_prec prec GMP_RNDN 0.0)))) 
    then
      (M(FR.from_float_prec prec GMP_RNDN 0.0),mpFloatMax (mpFloatAbs a prec GMP_RNDN) b)
    else
      (a,b) 
| OUTSIDE -> 
    if (mpFloatLt b (M(FR.from_float_prec prec GMP_RNDN 0.0))) then
      (mpFloatAbs b prec GMP_RNDD,mpFloatAbs a prec GMP_RNDU)
    else
    if ((mpFloatLt a (M(FR.from_float_prec prec GMP_RNDN 0.0)))&&
       (mpFloatGte b (M(FR.from_float_prec prec GMP_RNDN 0.0)))) 
    then
      (M(FR.from_float_prec prec GMP_RNDN 0.0),mpFloatMax (mpFloatAbs a prec GMP_RNDU) b)
    else
      (a,b) ;;


let mpfiNeg (a,b) prec mode =  match mode with
  NEAREST ->
    let a' = mpFloatNeg a prec GMP_RNDN in
    let a'' = mpFloatNeg a prec GMP_RNDN in
    let b' = mpFloatNeg b prec GMP_RNDN in
    let b'' = mpFloatNeg b prec GMP_RNDN in
      (mpFloatMin4 a' a'' b' b'',mpFloatMax4 a' a'' b' b'') 
| OUTSIDE ->
    let a' = mpFloatNeg a prec GMP_RNDD in
    let a'' = mpFloatNeg a prec GMP_RNDU in
    let b' = mpFloatNeg b prec GMP_RNDD in
    let b'' = mpFloatNeg b prec GMP_RNDU in
      (mpFloatMin4 a' a'' b' b'',mpFloatMax4 a' a'' b' b'') ;;


let mpfiPow (a,b) n prec mode = match mode with
	NEAREST ->	if (((n mod 2)=0)  && (mpFloatLt a (M(FR.from_float_prec prec GMP_RNDN 0.0)))
								   && (mpFloatGte b (M(FR.from_float_prec prec GMP_RNDN 0.0)))) then  
					let a' = mpFloatPow a n prec GMP_RNDN in
					let b' = mpFloatPow b n prec GMP_RNDN in
					let c = mpFloatMax a' b' in
						(match c with
							MBottom ->	(MBottom,MBottom)
						|	_ ->		(M(FR.from_float_prec prec GMP_RNDN 0.0),c)
						)
				else
					let a' = mpFloatPow a n prec GMP_RNDN in
					let b' = mpFloatPow b n prec GMP_RNDN in
						(mpFloatMin a' b',mpFloatMax a' b') 
|	OUTSIDE ->	if (((n mod 2)=0)  && (mpFloatLt a (M(FR.from_float_prec prec GMP_RNDN 0.0)))
								   && (mpFloatGte b (M(FR.from_float_prec prec GMP_RNDN 0.0)))) then  
					let a' = mpFloatPow a n prec GMP_RNDD in
					let b' = mpFloatPow b n prec GMP_RNDD in
					let a'' = mpFloatPow a n prec GMP_RNDU in
					let b'' = mpFloatPow b n prec GMP_RNDU in
					let c = mpFloatMax4 a' b' a'' b'' in
						(match c with
							MBottom ->	(MBottom,MBottom)
						|	_ ->		(M(FR.from_float_prec prec GMP_RNDN 0.0),c)
						)
				else
					let a' = mpFloatPow a n prec GMP_RNDD in
					let b' = mpFloatPow b n prec GMP_RNDD in
					let a'' = mpFloatPow a n prec GMP_RNDU in
					let b'' = mpFloatPow b n prec GMP_RNDU in
						(mpFloatMin4 a' b' a'' b'',mpFloatMax4 a' b' a'' b'') ;;


 let mpfiLt' v0Min v0Max v1Min v1Max = 
  if (mpFloatLt v0Max v1Min) then B(true) else
    if (mpFloatLte v1Max v0Min) then B(false) else 
	BTop ;;


let mpfiLt (v0Min,v0Max) (v1Min,v1Max) = (mpfiLt' v0Min v0Max v1Min v1Max) ;;


let mpfiGt (v0Min,v0Max) (v1Min,v1Max) = mpfiLt' v1Min v1Max v0Min v0Max ;; 


let mpfiLte' v0Min v0Max v1Min v1Max = 
  if (mpFloatLte v0Max v1Min) then B(true) else
    if (mpFloatLt v1Max v0Min) then B(false) else BTop ;;


let mpfiLte (v0Min,v0Max) (v1Min,v1Max) = (mpfiLte' v0Min v0Max v1Min v1Max) ;;


let mpfiGte (v0Min,v0Max) (v1Min,v1Max) = (mpfiLte' v1Min v1Max v0Min v0Max) ;;


let mpfiEq (v0Min,v0Max) (v1Min,v1Max) = 
    if ((mpFloatEq v0Min v0Max)&&(mpFloatEq v1Min v1Max)&&(mpFloatEq v0Min v1Max)) then 
      B(true) 
	else 
		if ((((mpFloatLte v0Min v1Min)&&(mpFloatLte v1Min v0Max)) ||
 		    ((mpFloatLte v0Min v1Max)&&(mpFloatLte v1Max v0Max))) 
			||
			(((mpFloatLte v1Min v0Min)&&(mpFloatLte v0Min v1Max)) ||
 		    ((mpFloatLte v1Min v0Max)&&(mpFloatLte v0Max v1Max))) 
			)
			then BTop
		else
			B(false)
    ;;
   
   
let mpfiNeq (v0Min,v0Max) (v1Min,v1Max) = (
    if ((mpFloatLt v0Max v1Min)||(mpFloatLt v1Max v0Min)) then B(true) else
    if ((mpFloatEq v0Min v0Max)&&(mpFloatEq v1Min v1Max)&&(not (mpFloatEq v0Min v1Min))) 
    then B(false) 
    else BTop 
   ) ;;


let mpfiJoin a b = match (a,b) with 
  ((MBottom,MBottom),_) -> b
| (_,(MBottom,MBottom)) -> a
| ((v0Min,v0Max),(v1Min,v1Max)) -> 
 ((mpFloatMin v0Min v1Min,mpFloatMax v0Max v1Max)) ;;


let mpfiMeet' a b = match (a,b) with 
  ((v0Min,v0Max),(v1Min,v1Max)) -> 
    let (x,y) = (mpFloatMax v0Min v1Min,mpFloatMin v0Max v1Max) 
    in if (mpFloatGt x y) then (MBottom,MBottom) else (x,y) ;;

    
let mpfiMeet a b = match (a,b) with 
  ((v0Min,v0Max),(v1Min,v1Max)) -> (mpfiMeet' (v0Min,v0Max) (v1Min,v1Max)) ;;


let mpfiSseq a b = match (a,b) with 
  ((v0Min,v0Max),(v1Min,v1Max)) -> ((mpFloatGte v0Min v1Min)&&(mpFloatLte v0Max v1Max)) ;;


let mpfi2I32 v = match v with
  (f1,f2) -> (mpFloat2Int32 f1,mpFloat2Int32 f2) ;;

  
let mpfi2I64 v = match v with
  (f1,f2) -> (mpFloat2Int64 f1,mpFloat2Int64 f2) ;;


(*****************************************************************************)
(*                      Roundoff errors on operations                        *)
(*****************************************************************************)

let ulp (mx,my) prec = match (mx,my) with 
  (M(x),M(y)) -> 
            let r = let x' = (FR.abs_prec prec GMP_RNDU x) in
					let y' = (FR.abs_prec prec GMP_RNDU y) in
						if ((FR.cmp x' y')<0) then y' else x' 
            in
			if (FR.cmp r (FR.from_float_prec prec GMP_RNDN 0.0)=0) then
				(M((FR.from_float_prec prec GMP_RNDN 0.0)),M((FR.from_float_prec prec GMP_RNDN 0.0)))
			else
			let eInt = int_of_float (floor ((log (FR.to_float_mode GMP_RNDD r)) /. 0.69314719)) in 
			let a = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) eInt (!errPrec) in	
			let b = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt+1) (!errPrec) 
			in
				if ((mpFloatLte a (M(r)))&&(mpFloatLte (M(r)) b)) then (* just in case caml logarithm fails *)
					let u = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt-prec) (!errPrec)
					in (mpFloatMult u (M((FR.from_float_prec (!errPrec) GMP_RNDN (-1.0)))) (!errPrec) GMP_RNDD,u) 
				else
					let a' = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt-1) (!errPrec) in
					let b' = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt) (!errPrec) in
						if ((mpFloatLte a' (M(r)))&&(mpFloatLte (M(r)) b')) then (* just in case caml logarithm fails *)
							let u' = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt-prec-1) (!errPrec)
							in (mpFloatMult u' (M((FR.from_float_prec (!errPrec) GMP_RNDN (-1.0)))) (!errPrec) GMP_RNDD,u') 
						else				
							let a'' = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt+1) (!errPrec) in
							let b'' = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt+2) (!errPrec) in
								if ((mpFloatLt a'' (M(r)))&&(mpFloatLt (M(r)) b'')) then (* just in case caml logarithm fails *)
									let u'' = pow2M (M((FR.from_float_prec (!errPrec) GMP_RNDN 1.0))) (eInt-prec+1) (!errPrec)
									in (mpFloatMult u'' (M((FR.from_float_prec (!errPrec) GMP_RNDN (-1.0)))) (!errPrec) GMP_RNDD,u'') 
								else
									raise (Error ("ulp","problem in extraction of exponent"^(FR.to_string_base_digits GMP_RNDN 10 prec r)))
| (M(r),MPInfty) -> (MMInfty,MPInfty)
| (MMInfty,M(r)) -> (MMInfty,MPInfty)
| (MMInfty,MPInfty) -> (MMInfty,MPInfty)
| (MPInfty,MPInfty) -> (MMInfty,MPInfty)
| (MMInfty,MMInfty) -> (MMInfty,MPInfty)
| _ -> (MBottom,MBottom) 
;;


let ulpAddConcrete x y prec = match (x,y) with 
  (M(x'),M(y')) -> 
    let rFloat = FR.add_prec prec GMP_RNDN x' y' in
    let rRealD  = FR.add_prec (!errPrec) GMP_RNDD x' y' in
    let rRealU  = FR.add_prec (!errPrec) GMP_RNDU x' y' in
    let uD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealD rFloat) in
    let vD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealU rFloat) in
    let uU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealD rFloat) in
    let vU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealU rFloat) 
    in (mpFloatMin4 uD vD uU vU,mpFloatMax4 uD vD uU vU)
| (M(_),MPInfty) -> (MMInfty,MPInfty)
| (M(_),MMInfty) -> (MMInfty,MPInfty)
| (M(_),MBottom) -> (MBottom,MBottom)
| (MPInfty,M(_)) -> (MMInfty,MPInfty)
| (MMInfty,M(_)) -> (MMInfty,MPInfty)
| (MPInfty,MPInfty) -> (MMInfty,MPInfty)
| (MMInfty,MMInfty) -> (MMInfty,MPInfty)
| (MMInfty,MPInfty) -> (MMInfty,MPInfty)
| (MPInfty,MMInfty) -> (MMInfty,MPInfty)
| (MBottom,M(_)) -> (MBottom,MBottom)
| (_,MBottom) -> (MBottom,MBottom)
| (MBottom,_) -> (MBottom,MBottom) ;;
    

let ulpAdd (f1l,f1r) (f2l,f2r) prec = 
  if ((mpFloatEq f1l f1r) & (mpFloatEq f2l f2r)) then
    ulpAddConcrete f1l f2l prec
  else 
    let rl = mpFloatPlus f1l f2l prec GMP_RNDD in
    let rr = mpFloatPlus f1r f2r prec GMP_RNDU 
    in (match (rl,rr) with 
          (M(_),M(_)) -> ulp (rl,rr) prec
	    | (M(r),MPInfty) -> (MMInfty,MPInfty)
	    | (MMInfty,M(r)) -> (MMInfty,MPInfty)
        | (MMInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MPInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MMInfty,MMInfty) -> (MMInfty,MPInfty)
	    | _ -> (MBottom,MBottom) 
       ) ;;


let ulpSubConcrete x y prec = match (x,y) with (* exact error *)
  (M(x'),M(y')) -> 
    let rFloat = FR.sub_prec prec GMP_RNDN x' y' in
    let rRealD  = FR.sub_prec (!errPrec) GMP_RNDD x' y' in
    let rRealU  = FR.sub_prec (!errPrec) GMP_RNDU x' y' in
    let uD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealD rFloat) in
    let vD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealU rFloat) in
    let uU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealD rFloat) in
    let vU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealU rFloat) 
    in (mpFloatMin4 uD vD uU vU,mpFloatMax4 uD vD uU vU)
| (M(_),MPInfty) -> (MMInfty,MPInfty)
| (M(_),MMInfty) -> (MMInfty,MPInfty)
| (M(_),MBottom) -> (MBottom,MBottom)
| (MPInfty,M(_)) -> (MMInfty,MPInfty)
| (MMInfty,M(_)) -> (MMInfty,MPInfty)
| (MPInfty,MPInfty) -> (MMInfty,MPInfty)
| (MMInfty,MMInfty) -> (MMInfty,MPInfty)
| (MMInfty,MPInfty) -> (MMInfty,MPInfty)
| (MPInfty,MMInfty) -> (MMInfty,MPInfty)
| (MBottom,M(_)) -> (MBottom,MBottom)
| (_,MBottom) -> (MBottom,MBottom)
| (MBottom,_) -> (MBottom,MBottom) ;;


let ulpSub (f1l,f1r) (f2l,f2r) prec = 
  if ((mpFloatEq f1l f1r) & (mpFloatEq f2l f2r)) then
      ulpSubConcrete f1l f2l prec
  else 
    let rl = mpFloatMinus f1l f2l prec GMP_RNDN in
    let rr = mpFloatMinus f1r f2r prec GMP_RNDU
    in (match (rl,rr) with 
          (M(_),M(_)) -> ulp (rl,rr) prec
	    | (M(r),MPInfty) -> (MMInfty,MPInfty)
	    | (MMInfty,M(r)) -> (MMInfty,MPInfty)
        | (MMInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MPInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MMInfty,MMInfty) -> (MMInfty,MPInfty)
	    | _ -> (MBottom,MBottom) 
       ) ;;


let ulpMultConcrete x y prec = match (x,y) with (* exact error *)
  (M(x'),M(y')) -> 
    let rFloat = FR.mul_prec prec GMP_RNDN x' y' in
    let rRealD  = FR.mul_prec (!errPrec) GMP_RNDD x' y' in
    let rRealU  = FR.mul_prec (!errPrec) GMP_RNDU x' y' in
    let uD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealD rFloat) in
    let vD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealU rFloat) in
    let uU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealD rFloat) in
    let vU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealU rFloat) 
    in (mpFloatMin4 uD vD uU vU,mpFloatMax4 uD vD uU vU)
| (M(_),MPInfty) -> (MMInfty,MPInfty)
| (M(x),MMInfty) -> (MMInfty,MPInfty)
| (M(_),MBottom) -> (MBottom,MBottom)
| (MPInfty,M(x)) -> (MMInfty,MPInfty)
| (MMInfty,M(x)) -> (MMInfty,MPInfty)
| (MPInfty,MPInfty) -> (MMInfty,MPInfty)
| (MMInfty,MMInfty) -> (MMInfty,MPInfty)
| (MMInfty,MPInfty) -> (MMInfty,MPInfty)
| (MPInfty,MMInfty) -> (MMInfty,MPInfty)
| (MBottom,M(_)) -> (MBottom,MBottom)
| (_,MBottom) -> (MBottom,MBottom)
| (MBottom,_) -> (MBottom,MBottom) ;;


let ulpMult (f1l,f1r) (f2l,f2r) prec =  
  if ((mpFloatEq f1l f1r) & (mpFloatEq f2l f2r)) then
    ulpMultConcrete f1l f2l prec
  else 
    let r1 = mpFloatMult f1l f2l prec GMP_RNDN in
    let r2 = mpFloatMult f1l f2r prec GMP_RNDN in
    let r3 = mpFloatMult f1r f2l prec GMP_RNDN in
    let r4 = mpFloatMult f1r f2r prec GMP_RNDN in
    let rl = mpFloatMin4 r1 r2 r3 r4 in
    let rr = mpFloatMax4 r1 r2 r3 r4 
    in (match (rl,rr) with 
          (M(_),M(_)) -> ulp (rl,rr) prec
	    | (M(r),MPInfty) -> (MMInfty,MPInfty)
	    | (MMInfty,M(r)) -> (MMInfty,MPInfty)
        | (MMInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MPInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MMInfty,MMInfty) -> (MMInfty,MPInfty)
	    | _ -> (MBottom,MBottom) 
       ) ;;


let ulpDivConcrete x y prec = match (x,y) with (* exact error *)
  (M(x'),M(y')) -> 
    let rFloat = FR.div_prec prec GMP_RNDN x' y' in
    let rRealD  = FR.div_prec (!errPrec) GMP_RNDD x' y' in
    let rRealU  = FR.div_prec (!errPrec) GMP_RNDU x' y' in
    let uD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealD rFloat) in
    let vD = M(FR.sub_prec (!errPrec) GMP_RNDD rRealU rFloat) in
    let uU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealD rFloat) in
    let vU = M(FR.sub_prec (!errPrec) GMP_RNDU rRealU rFloat) 
    in (mpFloatMin4 uD vD uU vU,mpFloatMax4 uD vD uU vU)
| (M(_),MPInfty) -> (MMInfty,MPInfty)
| (M(x),MMInfty) -> (MMInfty,MPInfty)
| (M(_),MBottom) -> (MBottom,MBottom)
| (MPInfty,M(x)) -> (MMInfty,MPInfty)
| (MMInfty,M(x)) -> (MMInfty,MPInfty)
| (MPInfty,MPInfty) -> (MMInfty,MPInfty)
| (MMInfty,MMInfty) -> (MMInfty,MPInfty)
| (MMInfty,MPInfty) -> (MMInfty,MPInfty)
| (MPInfty,MMInfty) -> (MMInfty,MPInfty)
| (MBottom,M(_)) -> (MBottom,MBottom)
| (_,MBottom) -> (MBottom,MBottom)
| (MBottom,_) -> (MBottom,MBottom) ;;

	   
let ulpDiv (f1l,f1r) (f2l,f2r) prec =  
  if ((mpFloatEq f1l f1r) & (mpFloatEq f2l f2r)) then
    ulpDivConcrete f1l f2l prec
  else 
    let r1 = mpFloatDiv f1l f2l prec GMP_RNDN in
    let r2 = mpFloatDiv f1l f2r prec GMP_RNDN in
    let r3 = mpFloatDiv f1r f2l prec GMP_RNDN in
    let r4 = mpFloatDiv f1r f2r prec GMP_RNDN in
    let rl = mpFloatMin4 r1 r2 r3 r4 in
    let rr = mpFloatMax4 r1 r2 r3 r4 
    in (match (rl,rr) with 
          (M(_),M(_)) -> ulp (rl,rr) prec
	    | (M(r),MPInfty) -> (MMInfty,MPInfty)
	    | (MMInfty,M(r)) -> (MMInfty,MPInfty)
        | (MMInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MPInfty,MPInfty) -> (MMInfty,MPInfty)
        | (MMInfty,MMInfty) -> (MMInfty,MPInfty)
	    | _ -> (MBottom,MBottom) 
       ) ;;
	   

(*****************************************************************************)
(*                    operations on Floats with errors                       *)
(*****************************************************************************)


let fePlus (FE(f1,e1,_)) (FE(f2,e2,_)) prec = 
    let f = mpfiPlus f1 f2 prec NEAREST in 
    let ulp = ulpAdd f1 f2 prec
    in FE(f,mpfiPlus (mpfiPlus ulp e1 (!errPrec) OUTSIDE) e2 (!errPrec) OUTSIDE,prec) ;;

	
let feMinus (FE(f1,e1,_)) (FE(f2,e2,_)) prec = 
    let f = mpfiMinus f1 f2 prec NEAREST in 
    let ulp = ulpSub f1 f2 prec
    in FE(f,mpfiPlus (mpfiMinus e1 e2 (!errPrec) OUTSIDE) ulp (!errPrec) OUTSIDE,prec) ;;


let feMult (FE(f1,e1,_)) (FE(f2,e2,_)) prec = 
    let f = mpfiMult f1 f2 prec NEAREST in 
    let ulp = ulpMult f1 f2 prec
    in FE(f,mpfiPlus (mpfiPlus (mpfiMult f1 e2 (!errPrec) OUTSIDE) (mpfiMult f2 e1 (!errPrec) OUTSIDE) (!errPrec) OUTSIDE) 
	                 (mpfiPlus ulp (mpfiMult e1 e2 (!errPrec) OUTSIDE) (!errPrec) OUTSIDE) (!errPrec) OUTSIDE,prec) ;;


let feDiv (FE(f1,e1,_)) (FE(f2,e2,_)) prec = 
    let f = mpfiDiv f1 f2 prec NEAREST in 
    let ulp = ulpDiv f1 f2 prec 
    in FE(f,mpfiPlus (mpfiPlus ulp (mpfiDiv e1 f2 (!errPrec) OUTSIDE) (!errPrec) OUTSIDE)
			  		 		(mpfiDiv (mpfiMult e1 e2 (!errPrec) OUTSIDE) 
							   		 (mpfiMult f2 (mpfiPlus f2 e2 (!errPrec) OUTSIDE) (!errPrec) OUTSIDE) 
							  		 (!errPrec) OUTSIDE)
					 		(!errPrec) OUTSIDE,prec) ;;
										 
				
let feNeg s prec = match s with
	FE(f,e,_) -> FE(mpfiNeg f prec NEAREST,mpfiNeg e (!errPrec) OUTSIDE,prec) ;;
					 
					 
let feAbs s prec = match s with
  FE(f,e,_) -> FE(mpfiAbs f prec NEAREST,mpfiAbs e (!errPrec) OUTSIDE,prec) ;;


let rec fact n res = if (n=0) then res else fact (n-1) (n*res) ;; 		
		
		
let comb n k = 
	let c = float_of_int ((fact n 1)/((fact k 1)*(fact (n-k) 1)))
	in (M(FR.from_float_prec float64_mantissa GMP_RNDD c),M(FR.from_float_prec float64_mantissa GMP_RNDU c)) ;;
		
		
let rec fePowAux f e p n k res =
	if (n=k) then
		mpfiPlus res (mpfiPow e n p OUTSIDE) p OUTSIDE
	else 
		let c = comb n k in
		let xf = mpfiPow f (n-k) p OUTSIDE in
		let xe = mpfiPow e k p OUTSIDE in
		let r = mpfiMult c (mpfiMult xf xe p OUTSIDE) p  OUTSIDE in
			fePowAux f e p n (k+1) (mpfiPlus r res p OUTSIDE) ;;
			
	
let fePow fe n = 
	if (n=1) then fe else
	 	(match fe with
			FE(f,e,p) -> let x = mpfiPow f n p NEAREST in
							FE(x,mpfiPlus (fePowAux f e p n 1 !f64Zero) (ulp x p) p OUTSIDE,p) 
		);;
	
	
let feSseq (FE(f1,e1,_)) (FE(f2,e2,_)) =
  if ((mpfiSseq f1 f2) && (mpfiSseq e1 e2)) then true else false ;;
  

let feSseqNeglectErrors (FE(f1,_,_)) (FE(f2,_,_)) =
  if (mpfiSseq f1 f2) then true else false ;;
    

let feEq fe1 fe2 = ((feSseq fe1 fe2) && (feSseq fe2 fe1)) ;;
  
  
let feJoin (FE(f1,e1,p1)) (FE(f2,e2,p2)) = 
    let f = mpfiJoin f1 f2 in
    let e = mpfiJoin e1 e2 in FE(f,e,max p1 p2) ;;

	
let feMeet (FE(f1,e1,p1)) (FE(f2,e2,p2)) =
    let f = mpfiMeet f1 f2 in
    let e = mpfiMeet e1 e2 in FE(f,e,max p1 p2) ;;
	

let rec fePowIntervalAux x (m,n) res = 
	if (m>n) then res
	else (fePowIntervalAux x (m+1,n) (feJoin (fePow x m) res)) ;;
	

let fePowInterval x (m,n) = fePowIntervalAux x (m,n) (FE((MBottom,MBottom),(MBottom,MBottom),0)) ;;

				
let cst1over2 = (M(FR.from_float_prec !errPrec GMP_RNDD 0.5),M(FR.from_float_prec !errPrec GMP_RNDU 0.5)) ;;
let cst1over8 = (M(FR.from_float_prec !errPrec GMP_RNDD 0.125),M(FR.from_float_prec !errPrec GMP_RNDU 0.125)) ;;
let cst1over16 = (M(FR.from_float_prec !errPrec GMP_RNDD 0.0625),M(FR.from_float_prec !errPrec GMP_RNDU 0.0625)) ;;
let cst5over128 = (M(FR.from_float_prec !errPrec GMP_RNDD 0.0390625),M(FR.from_float_prec !errPrec GMP_RNDU 0.0390625)) ;;

	
let feSqrt v p = match v with
	FE((M(a),M(b)),e,_) ->	let a' = sqrt (FR.to_float a) in
							let a'' = FR.from_float_prec p GMP_RNDD a' in
							let b' = sqrt (FR.to_float b) in
							let b'' = FR.from_float_prec p GMP_RNDU b' in
							let x = (M(a''),M(b'')) in
							let ef1 = mpfiDiv e (M(a),M(b)) p OUTSIDE in
							let ef2 = mpfiPow ef1 2 p OUTSIDE in
							let ef3 = mpfiPow ef1 3 p OUTSIDE in
							let ef4 = mpfiPow ef1 4 p OUTSIDE in
							let o1 = mpfiMult ef1 cst1over2 !errPrec OUTSIDE in
							let o2 = mpfiMult ef2 cst1over8 !errPrec OUTSIDE in
							let o3 = mpfiMult ef3 cst1over16 !errPrec OUTSIDE in
							let o4 = mpfiMult ef4 cst5over128 !errPrec OUTSIDE in
							let e' = mpfiPlus (mpfiMinus o1 o2 !errPrec OUTSIDE) 
											  (mpfiMinus o3 o4 !errPrec OUTSIDE) !errPrec OUTSIDE  in
							let e'' = mpfiMult e' x !errPrec OUTSIDE in
							let e''' = mpfiPlus e'' (ulp x p) !errPrec OUTSIDE 
							in FE(x,e''',p) 
| 	_ ->					raise (Error ("feSqrt","??? Type mismatch: floatingpoint number expected.")) ;;			
								


								
								
								