(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file integer.ml contains the domain and operations on integers and integer intervals
 *)

open Gmp ;;
open Types ;;
open Prelude ;;

(*****************************************************************************)
(*                      operations on intNumbers                             *)
(*****************************************************************************)

let integerPlus a b = match (a,b) with
  (I32(x),I32(y)) ->  I32(Int32.add x y) (* pas de gestion des infinis, caml retourne le resultat modulo 2^31 *)
| (I64(x),I64(y)) ->  I64(Int64.add x y)
| (I32(x),I64(y)) ->  I64(Int64.add (Int64.of_int32 x) y)
| (I64(x),I32(y)) ->  I64(Int64.add x (Int64.of_int32 y))
| (I32(x),IPInfty) -> IPInfty
| (IPInfty,I32(y)) -> IPInfty
| (I32(x),IMInfty) -> IMInfty
| (IMInfty,I32(y)) -> IMInfty
| (I64(x),IPInfty) -> IPInfty
| (IPInfty,I64(y)) -> IPInfty
| (I64(x),IMInfty) -> IMInfty
| (IMInfty,I64(y)) -> IMInfty
| (IPInfty,IPInfty) -> IPInfty
| (IMInfty,IMInfty) -> IMInfty
| (IMInfty,IPInfty) -> IBottom
| (IPInfty,IMInfty) -> IBottom
| (IBottom,_) -> IBottom
| (_,IBottom) -> IBottom ;;


let integerMinus a b = match (a,b) with
  (I32(x),I32(y)) ->  I32(Int32.sub x y) (* pas de gestion des infinis, caml retourne le resultat modulo 2^31 *)
| (I64(x),I64(y)) ->  I64(Int64.sub x y)
| (I32(x),I64(y)) ->  I64(Int64.sub (Int64.of_int32 x) y)
| (I64(x),I32(y)) ->  I64(Int64.sub x (Int64.of_int32 y))
| (I32(x),IPInfty) -> IMInfty
| (IPInfty,I32(y)) -> IPInfty
| (I32(x),IMInfty) -> IPInfty
| (IMInfty,I32(y)) -> IMInfty
| (I64(x),IPInfty) -> IMInfty
| (IPInfty,I64(y)) -> IPInfty
| (I64(x),IMInfty) -> IPInfty
| (IMInfty,I64(y)) -> IMInfty
| (IPInfty,IPInfty) -> IBottom
| (IMInfty,IMInfty) -> IBottom
| (IMInfty,IPInfty) -> IMInfty
| (IPInfty,IMInfty) -> IPInfty
| (IBottom,_) -> IBottom
| (_,IBottom) -> IBottom ;;


let integerMult a b = match (a,b) with
  (I32(x),I32(y)) ->  I32(Int32.mul x y) (* pas de gestion des infinis, caml retourne le resultat modulo 2^31 *)
| (I64(x),I64(y)) ->  I64(Int64.mul x y)
| (I32(x),I64(y)) ->  I64(Int64.mul (Int64.of_int32 x) y)
| (I64(x),I32(y)) ->  I64(Int64.mul x (Int64.of_int32 y))
| (I32(x),IPInfty) -> 
    if ((Int32.compare x Int32.zero)=0) then
      IBottom
    else if ((Int32.compare x Int32.zero)>0) then IPInfty else IMInfty
| (IPInfty,I32(y)) -> 
    if ((Int32.compare y Int32.zero)=0) then
      IBottom
    else if ((Int32.compare y Int32.zero)>0) then IPInfty else IMInfty
| (I32(x),IMInfty) -> 
    if ((Int32.compare x Int32.zero)=0) then
      IBottom
    else if ((Int32.compare x Int32.zero)>0) then IMInfty else IPInfty
| (IMInfty,I32(y)) -> 
    if ((Int32.compare y Int32.zero)=0) then
      IBottom
    else if ((Int32.compare y Int32.zero)>0) then IMInfty else IPInfty
| (I64(x),IPInfty) -> 
    if ((Int64.compare x Int64.zero)=0) then
      IBottom
    else if ((Int64.compare x Int64.zero)>0) then IPInfty else IMInfty
| (IPInfty,I64(y)) -> 
    if ((Int64.compare y Int64.zero)=0) then
      IBottom
    else if ((Int64.compare y Int64.zero)>0) then IPInfty else IMInfty
| (I64(x),IMInfty) -> 
    if ((Int64.compare x Int64.zero)=0) then
      IBottom
    else if ((Int64.compare x Int64.zero)>0) then IMInfty else IPInfty
| (IMInfty,I64(y)) -> 
    if ((Int64.compare y Int64.zero)=0) then
      IBottom
    else if ((Int64.compare y Int64.zero)>0) then IMInfty else IPInfty
| (IPInfty,IPInfty) -> IPInfty
| (IMInfty,IPInfty) -> IMInfty
| (IPInfty,IMInfty) -> IMInfty
| (IMInfty,IMInfty) -> IPInfty	
| (IBottom,_) -> IBottom
| (_,IBottom) -> IBottom ;;


let integerDiv a b = match (a,b) with
  (I32(x),I32(y)) -> if ((Int32.compare y Int32.zero)=0) then IBottom
                 else I32(Int32.div x y) 
| (I64(x),I64(y)) -> if ((Int64.compare y Int64.zero)=0) then IBottom
                 else I64(Int64.div x y) 
| (I32(x),I64(y)) -> if ((Int64.compare y Int64.zero)=0) then IBottom
                 else I64(Int64.div (Int64.of_int32 x) y) 
| (I64(x),I32(y)) -> if ((Int32.compare y Int32.zero)=0) then IBottom
                 else I64(Int64.div x (Int64.of_int32 y)) 
| (I32(x),IPInfty) -> I32(Int32.zero)
| (IPInfty,I32(y)) -> if ((Int32.compare y Int32.zero)=0) then 
                      IBottom 
                    else 
                      if ((Int32.compare y Int32.zero)>0) then IPInfty else IMInfty
| (I32(x),IMInfty) -> I32(Int32.zero)
| (IMInfty,I32(y)) -> if ((Int32.compare y Int32.zero)=0) then 
                      IBottom 
                    else 
                      if ((Int32.compare y Int32.zero)>0) then IMInfty else IPInfty
| (I64(x),IPInfty) -> I64(Int64.zero)
| (IPInfty,I64(y)) -> if ((Int64.compare y Int64.zero)=0) then 
                      IBottom 
                    else 
                      if ((Int64.compare y Int64.zero)>0) then IPInfty else IMInfty
| (I64(x),IMInfty) -> I64(Int64.zero)
| (IMInfty,I64(y)) -> if ((Int64.compare y Int64.zero)=0) then 
                      IBottom 
                    else 
                      if ((Int64.compare y Int64.zero)>0) then IMInfty else IPInfty	  
| (IPInfty,IPInfty) -> IBottom
| (IPInfty,IMInfty) -> IBottom
| (IMInfty,IMInfty) -> IBottom
| (IMInfty,IPInfty) -> IBottom
| (IBottom,_) -> IBottom
| (_,IBottom) -> IBottom ;;


let integerAbs a = match a with
   I32(x) -> if ((Int32.compare x Int32.zero)<0) then 
             I32(Int32.mul x Int32.minus_one)
           else a
|  I64(x) -> if ((Int64.compare x Int64.zero)<0) then 
             I64(Int64.mul x Int64.minus_one)
           else a
| IPInfty -> IPInfty
| IMInfty -> IPInfty
| IBottom -> IBottom ;;


let integerNeg a = match a with
  I32(x) -> I32(Int32.mul x Int32.minus_one)
| I64(x) -> I64(Int64.mul x Int64.minus_one)
| IPInfty -> IMInfty
| IMInfty -> IPInfty
| IBottom -> IBottom ;;


let rec int32Pow x n = 
	if (n<0) then
		raise (Error ("int32Pow","Cannot compute the nth power of a negative integer number"))
	else
		if (n=0) then Int32.one
		else (Int32.mul x (int32Pow x (n-1))) ;;

		
let rec int64Pow x n = 
	if (n<0) then
		raise (Error ("int64Pow","Cannot compute the nth power of a negative integer number"))
	else
		if (n=0) then Int64.one
		else (Int64.mul x (int64Pow x (n-1))) ;;
		
		
let integerPow a n = match a with 		
  I32(x) -> I32(int32Pow x n)
| I64(x) -> I64(int64Pow x n)
| IPInfty -> IPInfty
| IMInfty -> if ((n mod 2)=0) then IPInfty else IMInfty 
| IBottom -> IBottom ;;
		

let integerMin a b = match (a,b) with
  (I32(x),I32(y)) ->  if ((Int32.compare x y)<0) then a else b
| (I64(x),I64(y)) ->  if ((Int64.compare x y)<0) then a else b
| (I32(x),I64(y)) ->  if ((Int64.compare (Int64.of_int32 x) y)<0) then a else b
| (I64(x),I32(y)) ->  if ((Int64.compare x (Int64.of_int32 y))<0) then a else b
| (I32(x),IPInfty) -> a
| (IPInfty,I32(y)) -> b
| (I32(x),IMInfty) -> IMInfty
| (IMInfty,I32(y)) -> IMInfty
| (I64(x),IPInfty) -> a
| (IPInfty,I64(y)) -> b
| (I64(x),IMInfty) -> IMInfty
| (IMInfty,I64(y)) -> IMInfty
| (IPInfty,IPInfty) -> IPInfty
| (IMInfty,IMInfty) -> IMInfty
| (IMInfty,IPInfty) -> IMInfty
| (IPInfty,IMInfty) -> IMInfty
| (IBottom,_) -> IBottom
| (_,IBottom) -> IBottom ;;


let integerMax a b = match (a,b) with
  (I32(x),I32(y)) ->  if ((Int32.compare x y)>0) then a else b
| (I64(x),I64(y)) ->  if ((Int64.compare x y)>0) then a else b
| (I32(x),I64(y)) ->  if ((Int64.compare (Int64.of_int32 x) y)>0) then a else b
| (I64(x),I32(y)) ->  if ((Int64.compare x (Int64.of_int32 y))>0) then a else b
| (I32(x),IPInfty) -> IPInfty
| (IPInfty,I32(y)) -> IPInfty
| (I32(x),IMInfty) -> a
| (IMInfty,I32(y)) -> b
| (I64(x),IPInfty) -> IPInfty
| (IPInfty,I64(y)) -> IPInfty
| (I64(x),IMInfty) -> a
| (IMInfty,I64(y)) -> b
| (IPInfty,IPInfty) -> IPInfty
| (IMInfty,IMInfty) -> IMInfty
| (IMInfty,IPInfty) -> IPInfty
| (IPInfty,IMInfty) -> IPInfty
| (IBottom,_) -> b
| (_,IBottom) -> a ;;


let integerMin4 a b c d = integerMin (integerMin a b) (integerMin c d) ;;
let integerMax4 a b c d = integerMax (integerMax a b) (integerMax c d) ;;


let rec pow2I x n = 
  if (n>0) then integerMult (I32(Int32.of_int 2)) (pow2I x (n-1)) else
  if (n<0) then integerDiv (pow2I x (n+1)) (I32(Int32.of_int 2)) else x ;;

  
let integerLt a b = match (a,b) with
  (I32(x),I32(y)) ->  if ((Int32.compare x y)<0) then true else false
| (I64(x),I64(y)) ->  if ((Int64.compare x y)<0) then true else false  
| (I32(x),I64(y)) ->  if ((Int64.compare (Int64.of_int32 x) y)<0) then true else false  
| (I64(x),I32(y)) ->  if ((Int64.compare x (Int64.of_int32 y))<0) then true else false  
| (I32(x),IPInfty) -> true
| (IPInfty,I32(y)) -> false
| (I32(x),IMInfty) -> false
| (IMInfty,I32(y)) -> true
| (I64(x),IPInfty) -> true
| (IPInfty,I64(y)) -> false
| (I64(x),IMInfty) -> false
| (IMInfty,I64(y)) -> true
| (IPInfty,IPInfty) -> false
| (IMInfty,IMInfty) -> false
| (IMInfty,IPInfty) -> true
| (IPInfty,IMInfty) -> false
| (IBottom,IBottom) -> false
| (IBottom,_) -> true
| (_,IBottom) -> false ;;


let integerGt a b = match (a,b) with
  (I32(x),I32(y)) -> if ((Int32.compare x y)>0) then true else false 
| (I64(x),I64(y)) ->  if ((Int64.compare x y)>0) then true else false  
| (I32(x),I64(y)) ->  if ((Int64.compare (Int64.of_int32 x) y)>0) then true else false  
| (I64(x),I32(y)) ->  if ((Int64.compare x (Int64.of_int32 y))>0) then true else false  
| (I32(x),IPInfty) -> false
| (IPInfty,I32(y)) -> true
| (I32(x),IMInfty) -> true
| (IMInfty,I32(y)) -> false
| (I64(x),IPInfty) -> false
| (IPInfty,I64(y)) -> true
| (I64(x),IMInfty) -> true
| (IMInfty,I64(y)) -> false
| (IPInfty,IPInfty) -> false
| (IMInfty,IMInfty) -> false
| (IMInfty,IPInfty) -> false
| (IPInfty,IMInfty) -> true
| (IBottom,IBottom) -> false
| (IBottom,_) -> false
| (_,IBottom) -> true ;;


let integerLte a b = match (a,b) with
  (I32(x),I32(y)) -> if ((Int32.compare x y)<=0) then true else false
| (I64(x),I64(y)) ->  if ((Int64.compare x y)<=0) then true else false  
| (I32(x),I64(y)) ->  if ((Int64.compare (Int64.of_int32 x) y)<=0) then true else false  
| (I64(x),I32(y)) ->  if ((Int64.compare x (Int64.of_int32 y))<=0) then true else false  
| (I32(x),IPInfty) -> true
| (IPInfty,I32(y)) -> false
| (I32(x),IMInfty) -> false
| (IMInfty,I32(y)) -> true
| (I64(x),IPInfty) -> true
| (IPInfty,I64(y)) -> false
| (I64(x),IMInfty) -> false
| (IMInfty,I64(y)) -> true
| (IPInfty,IPInfty) -> true
| (IMInfty,IMInfty) -> true
| (IMInfty,IPInfty) -> true
| (IPInfty,IMInfty) -> false
| (IBottom,IBottom) -> true
| (IBottom,_) -> true
| (_,IBottom) -> false ;;


let integerGte a b = match (a,b) with
  (I32(x),I32(y)) -> if ((Int32.compare x y)>=0) then true else false
| (I64(x),I64(y)) ->  if ((Int64.compare x y)>=0) then true else false  
| (I32(x),I64(y)) ->  if ((Int64.compare (Int64.of_int32 x) y)>=0) then true else false  
| (I64(x),I32(y)) ->  if ((Int64.compare x (Int64.of_int32 y))>=0) then true else false  
| (I32(x),IPInfty) -> false
| (IPInfty,I32(y)) -> true
| (I32(x),IMInfty) -> true
| (IMInfty,I32(y)) -> false
| (I64(x),IPInfty) -> false
| (IPInfty,I64(y)) -> true
| (I64(x),IMInfty) -> true
| (IMInfty,I64(y)) -> false
| (IPInfty,IPInfty) -> true
| (IMInfty,IMInfty) -> true
| (IMInfty,IPInfty) -> false
| (IPInfty,IMInfty) -> true
| (IBottom,IBottom) -> true
| (IBottom,_) -> false
| (_,IBottom) -> true ;;


let integerEq a b = match (a,b) with
  (I32(x),I32(y)) -> ((Int32.compare x y)=0)
| (I64(x),I64(y)) -> ((Int64.compare x y)=0)
| (I64(x),I32(y)) -> ((Int64.compare x (Int64.of_int32 y))=0)
| (I32(x),I64(y)) -> ((Int64.compare (Int64.of_int32 x) y)=0)
| (I32(x),IPInfty) -> false
| (IPInfty,I32(y)) -> false
| (I32(x),IMInfty) -> false
| (IMInfty,I32(y)) -> false
| (I64(x),IPInfty) -> false
| (IPInfty,I64(y)) -> false
| (I64(x),IMInfty) -> false
| (IMInfty,I64(y)) -> false
| (IPInfty,IPInfty) -> true
| (IMInfty,IMInfty) -> true
| (IMInfty,IPInfty) -> false
| (IPInfty,IMInfty) -> false
| (IBottom,IBottom) -> true
| (IBottom,_) -> false
| (_,IBottom) -> false ;;


(*****************************************************************************)
(*                      operations on intIntervals                           *)
(*****************************************************************************)

let iPlus (v0Min,v0Max) (v1Min,v1Max) = ((integerPlus v0Min v1Min,integerPlus v0Max v1Max)) ;;


let iMinus (v0Min,v0Max) (v1Min,v1Max) = ((integerMinus v0Min v1Max,integerMinus v0Max v1Min)) ;;


let iMult (v0Min,v0Max) (v1Min,v1Max) =
      let i1 = (integerMult v0Min v1Min) in 
      let i2 = (integerMult v0Min v1Max) in
      let i3 = (integerMult v0Max v1Min) in 
      let i4 = (integerMult v0Max v1Max) in
       ((integerMin4 i1 i2 i3 i4,integerMax4 i1 i2 i3 i4)) ;;


let iDiv (v0Min,v0Max) (v1Min,v1Max) =
    let i1 = (integerDiv v0Min v1Min) in 
    let i2 = (integerDiv v0Min v1Max) in
    let i3 = (integerDiv v0Max v1Min) in 
    let i4 = (integerDiv v0Max v1Max) in
	let z = I32(Int32.zero) in
    let (i5,i6) = ((integerMin4 i1 i2 i3 i4,integerMax4 i1 i2 i3 i4)) in
  	(* matlab returns +oo or -oo for divisions by zero *)
	let _ = if (((integerLte v1Min z)&&(integerGte v1Max z)) && (not !warnMsgDivZero)) then
				(warnMsgDivZero := true ;
				print_string ("\n!!! Warning: result includes division by zero."))
	in
		if ((integerLte v1Min z)&&(integerGte v1Max z)) then
				if (integerGte v0Min z) then
					(i5,IPInfty)
				else
					if (integerLte v1Min z) then 
						(IMInfty,i6)
					else (IMInfty,IPInfty)
		else (i5,i6) ;;


let iAbs' a b = 
  if (integerLt (integerMult a b) (I32(Int32.zero))) then
    let a' = integerAbs a in
    let b' = integerAbs b in
      (I32(Int32.zero),integerMax a' b')
  else 
    let a' = integerAbs a in
    let b' = integerAbs b in
      (integerMin a' b',integerMax a' b') ;;    


let iAbs a b = match (a,b) with
  (((v0Min,v0Max)),((v1Min,v1Max))) ->
     if ((integerEq v0Min v1Min)&&(integerEq v0Max v1Max)) then
       (iAbs' v0Min v0Max)
     else 
       raise (Error ("iAbs ","arguments should be the same")) ;;


let iNeg (a,b) = 
  let a' = integerNeg a in
  let b' = integerNeg b in
    (integerMin a' b',integerMax a' b') ;;    
	

let iModulo' a b c d = match (a,b,c,d) with
  (I32(v0Min),I32(v0Max),I32(v1Min),I32(v1Max)) -> 
    if ((v0Min=v0Max) && (v1Min=v1Max)) then 
      (I32(Int32.rem v0Min v1Min),I32(Int32.rem v0Min v1Min))
    else 
      raise (Error ("iModulo'","not yet implemented")) 
| _ -> (IMInfty,IPInfty) ;;


let iModulo ((x,y),(c,d)) = (iModulo' x y c d) ;;


let iLeftShift' a b c d = match (a,b,c,d) with
  (I32(v0Min),I32(v0Max),I32(v1Min),I32(v1Max)) ->
    if (((Int32.compare v0Min v0Max)=0) && ((Int32.compare v1Min v1Max)=0)) then 
      (I32(Int32.shift_left v0Min (Int32.to_int v1Min)),I32(Int32.shift_left v0Min (Int32.to_int v1Min)))
    else 
      raise (Error ("iShiftLeft'","not yet implemented")) 
| _ -> (IMInfty,IPInfty) ;;


let iLeftShift ((a,b),(c,d)) = ((iLeftShift' a b c d)) ;;


let iRightShift' a b c d = match (a,b,c,d) with
  (I32(v0Min),I32(v0Max),I32(v1Min),I32(v1Max)) -> 
    if (((Int32.compare v0Min v0Max)=0) && ((Int32.compare v1Min v1Max)=0)) then 
      (I32(Int32.shift_right_logical v0Min (Int32.to_int v1Min)),
       I32(Int32.shift_right_logical v0Min (Int32.to_int v1Min)))
    else 
      raise (Error ("iRightShift'","not yet implemented")) 
| _ -> (IMInfty,IPInfty) ;;


let iRightShift ((a,b),(c,d)) = (iRightShift' a b c d) ;;


let iBitOr' a b c d = match (a,b,c,d) with
  (I32(v0Min),I32(v0Max),I32(v1Min),I32(v1Max)) ->
    if (((Int32.compare v0Min v0Max)=0) && ((Int32.compare v1Min v1Max)=0)) then 
      (I32(Int32.logor v0Min v1Min),I32(Int32.logor v0Min v1Min))
    else 
      raise (Error ("iBitOr'","not yet implemented")) 
| _ -> raise (Error ("iBitOr'","type mismach in iBitOr'")) ;;


let iBitOr ((a,b),(c,d)) = (iBitOr' a b c d) ;;


let iBitXor' a b c d = match (a,b,c,d) with
  (I32(v0Min),I32(v0Max),I32(v1Min),I32(v1Max)) ->
    if (((Int32.compare v0Min v0Max)=0) && ((Int32.compare v1Min v1Max)=0)) then 
      (I32(Int32.logxor v0Min v1Min),I32(Int32.logxor v0Min v1Min))
    else 
      raise (Error ("iBitXor'","not yet implemented"))  
| _ -> raise (Error ("iBitXor'","type mismach in iBitXor'")) ;;


let iBitXor (a,b) (c,d) = (iBitXor' a b c d) ;;


let iBitAnd' a b c d = match (a,b,c,d) with
  (I32(v0Min),I32(v0Max),I32(v1Min),I32(v1Max)) ->
    if (((Int32.compare v0Min v0Max)=0) && ((Int32.compare v1Min v1Max)=0)) then 
      (I32(Int32.logand v0Min v1Min),I32(Int32.logand v0Min v1Min))
    else 
      raise (Error ("iBitAnd'","not yet implemented"))  
| _ -> raise (Error ("iBitAnd","type mismach in iBitAnd'")) ;;


let iBitAnd (a,b) (c,d) = (iBitAnd' a b c d) ;;


let iLt' v0Min v0Max v1Min v1Max = 
  if (integerLt v0Max v1Min) then B(true) else
    if (integerLte v1Max v0Min) then B(false) else BTop ;;


let iLt (v0Min,v0Max) (v1Min,v1Max) = (iLt' v0Min v0Max v1Min v1Max) ;;


let iGt (v0Min,v0Max) (v1Min,v1Max) = iLt' v1Min v1Max v0Min v0Max ;; 


let iLte' v0Min v0Max v1Min v1Max = 
  if (integerLte v0Max v1Min) then B(true) else
    if (integerLt v1Max v0Min) then B(false) else BTop ;;


let iLte (v0Min,v0Max) (v1Min,v1Max) = (iLte' v0Min v0Max v1Min v1Max) ;;


let iGte (v0Min,v0Max) (v1Min,v1Max) = (iLte' v1Min v1Max v0Min v0Max) ;;


let iEq (v0Min,v0Max) (v1Min,v1Max) = 
    if ((integerEq v0Min v0Max)&&(integerEq v1Min v1Max)&&(integerEq v0Min v1Max)) then 
      B(true) 
	else 
		if ((((integerLte v0Min v1Min)&&(integerLte v1Min v0Max)) ||
 		    ((integerLte v0Min v1Max)&&(integerLte v1Max v0Max))) 
			||
			(((integerLte v1Min v0Min)&&(integerLte v0Min v1Max)) ||
 		    ((integerLte v1Min v0Max)&&(integerLte v0Max v1Max))) 
			)
			then BTop
		else
			B(false)
    ;;
   
   
let iNeq (v0Min,v0Max) (v1Min,v1Max) = (
    if ((integerLt v0Max v1Min)||(integerLt v1Max v0Min)) then B(true) else
    if ((integerEq v0Min v0Max)&&(integerEq v1Min v1Max)&&(not (integerEq v0Min v1Min))) 
    then B(false) 
    else BTop 
   ) ;;


let integer2Float i prec = match i with
  I32(x) -> M(FR.from_float_prec prec GMP_RNDN (Int32.to_float x))
| I64(x) -> M(FR.from_float_prec prec GMP_RNDN (Int64.to_float x))
| IPInfty -> MPInfty
| IMInfty -> MMInfty
| IBottom -> MBottom ;;


let i2f (a,b) prec = (integer2Float a prec,integer2Float b prec) ;;

let i2PairOfInt (a,b) = match (a,b) with
	(I32(x),I32(y)) -> (Int32.to_int x,Int32.to_int y)
| 	(I64(x),I64(y)) -> (Int64.to_int x,Int64.to_int y)
| 	(I32(x),I64(y)) -> (Int32.to_int x,Int64.to_int y)
| 	(I64(x),I32(y)) -> (Int64.to_int x,Int32.to_int y)
| _ -> raise (Error ("i2PairOfInt","finite values expected")) ;;


let iJoin a b = match (a,b) with 
  ((IBottom,IBottom),_) -> b
| (_,(IBottom,IBottom)) -> a
| ((v0Min,v0Max),(v1Min,v1Max)) -> (integerMin v0Min v1Min,integerMax v0Max v1Max) ;;


let iMeet a b = match (a,b) with 
  ((v0Min,v0Max),(v1Min,v1Max)) -> 
    let (x,y) = (integerMax v0Min v1Min,integerMin v0Max v1Max) 
    in if (integerGt x y) then (IBottom,IBottom) else (x,y) ;;
    

let iSseq (v0Min,v0Max) (v1Min,v1Max) = ((integerGte v0Min v1Min)&&(integerLte v0Max v1Max)) ;;


let iPow (a,b) n = 
	if (((n mod 2)=0) && (integerLt a (I32(Int32.zero))) && (integerGte b (I32(Int32.zero)))) then  
		let a' = integerPow a n in
		let b' = integerPow b n in
		let c =  integerMax a' b' in
			match c with
				I32(_) -> 	(I32(Int32.zero),c)
			|	I64(_) -> 	(I64(Int64.zero),c)			
			|	IPInfty ->	(I32(Int32.zero),c)
			|	IMInfty ->	raise (Error ("iPow","impossible"))
			|	IBottom ->	(IBottom,IBottom)
	else
		let a' = integerPow a n in
		let b' = integerPow b n in
			(integerMin a' b',integerMax a' b') ;;


let rec iPowIntervalAux x (m,n) res = 
	if (m>n) then res
	else (iPowIntervalAux x (m+1,n) (iJoin (iPow x m) res)) ;;
	

let iPowInterval x (m,n) = iPowIntervalAux x (m,n) (IBottom,IBottom) ;;
			
			
let iSqrt (u,v) = match(u,v) with
	(I32(a),I32(b)) -> 		let a' = floor (sqrt (Int32.to_float a)) in
							let a'' = Int32.of_float a' in
							let b' = ceil (sqrt (Int32.to_float b)) in
							let b'' = Int32.of_float b' in
								(I32(a''),I32(b''))
|	(I64(a),I64(b)) ->	 	let a' = floor (sqrt (Int64.to_float a)) in
							let a'' = Int64.of_float a' in
							let b' = ceil (sqrt (Int64.to_float b)) in
							let b'' = Int64.of_float b' in
								(I64(a''),I64(b''))								
|	_ ->					raise (Error ("iSqrt","??? Type mismatch: integer expected.")) ;; 
								
			
let iAbs (u,v) = match(u,v) with
	(I32(a),I32(b)) -> 		let a' = Int32.abs a in
							let b' = Int32.abs b in
								(I32(a'),I32(b'))
|	(I64(a),I64(b)) ->		let a' = Int64.abs a in
							let b' = Int64.abs b in
								(I64(a'),I64(b'))			
|	_ ->					raise (Error ("iAbs	","??? Type mismatch: integer expected.")) ;; 
		
			