(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file matrix.ml contains all the matrix operations 
 *)

open Types ;;
open Prelude ;;
open Gmp ;;
open Float ;;
open Coercion ;;			
			
			
let subMatrix m y0 x0 y1 x1 =
	let n = Array.sub m y0 (y1-y0+1) in
	let _ = for i=0 to ((Array.length n)-1) do
				n.(i) <- Array.sub n.(i) x0 (x1-x0+1)
			done
	in n ;;

	
let matrixOnes i j = AVal(Array.make_matrix i j (FVal(!fe64One))) ;;
let matrixZeros i j = AVal(Array.make_matrix i j (FVal(!fe64Zero))) ;;


let matrixEye i = 
	let m = Array.make_matrix i i (FVal(!fe64Zero)) in
	let _ = for j=0 to (i-1) do
				m.(j).(j) <- (FVal(!fe64One))
			done
	in AVal(m) ;;
	

let matrixDiag m k =
	let y = Array.length m in
	let x = (Array.length m.(0))-k+1 in
	let d = min x y in
	let n = Array.make_matrix 1 d ValBottom in 
	let _ =
		for i=0 to (d-1) do
			n.(0).(i) <- m.(i).(i+k-1)
		done
	in AVal(n) 
			

let matrixTranspose m =
	let x = (Array.length m.(0)) in
	let y = (Array.length m) in
	let n = Array.make_matrix x y ValBottom in
	let _ =	for i=0 to (y-1) do
				for j=0 to (x-1) do
					n.(j).(i) <- m.(i).(j)
				done
			done
	in AVal(n) ;;
	
	
let matrixRandom y x =
	let m = Array.make_matrix y x ValBottom in
	let _ = for i=0 to (y-1) do
				for j=0 to (x-1) do
					let r = M(FR.random float64_mantissa)
					in m.(i).(j) <- FVal(FE((r,r),!eZero,float64_mantissa))  
				done
			done
	in AVal(m) ;;
	

let matrixRandomErrors y x =
	let m = Array.make_matrix y x ValBottom in
	let _ = for i=0 to (y-1) do
				for j=0 to (x-1) do
					let r = M(FR.random float64_mantissa)
					in m.(i).(j) <- FVal(FE(!f64Zero,(r,r),float64_mantissa))  
				done
			done
	in AVal(m) ;;

	
let matrixLinspace a b n = if (n<2) then b else match (a,b) with
	(IVal(I32(x),_),IVal(_,I32(y))) ->	let h = ((Int32.to_float (Int32.sub y x))+.1.0) /. (Int32.to_float (Int32.of_int (n-1))) in
										let m = Array.make_matrix 1 n (IVal(!i32Zero)) in
										let _ = for i=0 to (n-2) do
													let r = Int32.add x (Int32.of_float (h *. (float_of_int i)))
													in m.(0).(i) <- IVal(I32(r),I32(r)) 
												done
										in
										let _ = m.(0).(n-1) <- b
										in AVal(m) 
|	(IVal(I64(x),_),IVal(_,I64(y))) ->	let h = ((Int64.to_float (Int64.sub y x))+.1.0) /. (Int64.to_float (Int64.of_int (n-1))) in
										let m = Array.make_matrix 1 n (IVal(!i64Zero)) in
										let _ = for i=0 to (n-2) do
													let r = Int64.add x (Int64.of_float (h *. (float_of_int i)))
													in m.(0).(i) <- IVal(I64(r),I64(r)) 
												done
										in
										let _ = m.(0).(n-1) <- b 
										in AVal(m) 	
|	(IVal(IMInfty,_),_) ->				raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))	
|	(IVal(IPInfty,_),_) ->				raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))	
|	(IVal(IBottom,_),_) ->				raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))	
|	(_,IVal(_,IMInfty)) ->				raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))	
|	(_,IVal(_,IPInfty)) ->				raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))	
|	(_,IVal(_,IBottom)) ->				raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))	
|	(FVal(FE((M(x),_),_,px)),FVal(FE((_,M(y)),_,py))) ->
										let p = max px py in
										let h = mpFloatDiv (mpFloatMinus (M(y)) (M(x)) p GMP_RNDN) (M(FR.from_float (float_of_int (n-1)))) p GMP_RNDN in
										let m = Array.make_matrix 1 n (FVal(!fe64Zero)) in
										let _ = for i=0 to (n-2) do
										let r = mpFloatPlus (M(x)) (mpFloatMult h (M(FR.from_float (float_of_int i))) p GMP_RNDN) p GMP_RNDN 
													in m.(0).(i) <- FVal(FE((r,r),!eZero,p)) 
												done
										in
										let _ = m.(0).(n-1) <- b 
										in AVal(m)
|	(FVal(FE((MMInfty,_),_,_)),_) ->	raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))
|	(FVal(FE((MPInfty,_),_,_)),_) ->	raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))
|	(FVal(FE((MBottom,_),_,_)),_) ->	raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))
|	(FVal(FE((_,MMInfty),_,_)),_) ->	raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))
|	(FVal(FE((_,MPInfty),_,_)),_) ->	raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))
|	(FVal(FE((_,MBottom),_,_)),_) ->	raise (Error ("matrixLinspace","??? Cannot apply linspace to an unbound domain."))
|	_ ->								raise (Error ("matrixLinspace","??? Type mismatch in linspace."))
;;


let matrixLogspace a b n = if (n<2) then b else match (a,b) with
	(IVal(I32(x),_),IVal(_,I32(y))) ->	let p = float64_mantissa in
										let x' = mpFloatPow (M(FR.from_float (float_of_int 10))) (Int32.to_int x) p GMP_RNDN in
										let y' = mpFloatPow (M(FR.from_float (float_of_int 10))) (Int32.to_int y) p GMP_RNDN in
										let h = mpFloatDiv (mpFloatMinus y' x' p GMP_RNDN) (M(FR.from_float (float_of_int (n-1)))) p GMP_RNDN in
										let m = Array.make_matrix 1 n (FVal(!fe64Zero)) in
										let _ = for i=0 to (n-2) do
										let r = mpFloatPlus x' (mpFloatMult h (M(FR.from_float (float_of_int i))) p GMP_RNDN) p GMP_RNDN 
													in m.(0).(i) <- FVal(FE((r,r),!eZero,p)) 
												done
										in
										let _ = m.(0).(n-1) <- FVal(FE((y',y'),!eZero,p))
										in AVal(m)
|	(IVal(I64(x),_),IVal(_,I64(y))) ->	let p = float64_mantissa in
										let x' = mpFloatPow (M(FR.from_float (float_of_int 10))) (Int64.to_int x) p GMP_RNDN in
										let y' = mpFloatPow (M(FR.from_float (float_of_int 10))) (Int64.to_int y) p GMP_RNDN in
										let h = mpFloatDiv (mpFloatMinus y' x' p GMP_RNDN) (M(FR.from_float (float_of_int (n-1)))) p GMP_RNDN in
										let m = Array.make_matrix 1 n (FVal(!fe64Zero)) in
										let _ = for i=0 to (n-2) do
										let r = mpFloatPlus x' (mpFloatMult h (M(FR.from_float (float_of_int i))) p GMP_RNDN) p GMP_RNDN 
													in m.(0).(i) <- FVal(FE((r,r),!eZero,p)) 
												done
										in
										let _ = m.(0).(n-1) <- FVal(FE((y',y'),!eZero,p))
										in AVal(m) 	
|	(IVal(IMInfty,_),_) ->				raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))	
|	(IVal(IPInfty,_),_) ->				raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))	
|	(IVal(IBottom,_),_) ->				raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))	
|	(_,IVal(_,IMInfty)) ->				raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))	
|	(_,IVal(_,IPInfty)) ->				raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))	
|	(_,IVal(_,IBottom)) ->				raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))	
|	(FVal(FE((M(x),_),_,px)),FVal(FE((_,M(y)),_,py))) ->
										let p = max px py in
										let x' = mpFloatPow (M(FR.from_float (float_of_int 10))) (int_of_float (FR.to_float x)) p GMP_RNDN in
										let y' = mpFloatPow (M(FR.from_float (float_of_int 10))) (int_of_float (FR.to_float y)) p GMP_RNDN in
										let h = mpFloatDiv (mpFloatMinus y' x' p GMP_RNDN) (M(FR.from_float (float_of_int (n-1)))) p GMP_RNDN in
										let m = Array.make_matrix 1 n (FVal(!fe64Zero)) in
										let _ = for i=0 to (n-2) do
										let r = mpFloatPlus x' (mpFloatMult h (M(FR.from_float (float_of_int i))) p GMP_RNDN) p GMP_RNDN 
													in m.(0).(i) <- FVal(FE((r,r),!eZero,p)) 
												done
										in
										let _ = m.(0).(n-1) <- FVal(FE((y',y'),!eZero,p))
										in AVal(m)
|	(FVal(FE((MMInfty,_),_,_)),_) ->	raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))
|	(FVal(FE((MPInfty,_),_,_)),_) ->	raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))
|	(FVal(FE((MBottom,_),_,_)),_) ->	raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))
|	(FVal(FE((_,MMInfty),_,_)),_) ->	raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))
|	(FVal(FE((_,MPInfty),_,_)),_) ->	raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))
|	(FVal(FE((_,MBottom),_,_)),_) ->	raise (Error ("matrixLogspace","??? Cannot apply Logspace to an unbound domain."))
|	_ ->								raise (Error ("matrixLogspace","??? Type mismatch in Logspace."))
;;


(***** j:i:k  --> [j,j+i,j+2i, ...,j+m*i], where m = fix((k-j)/i *)
let rec matrixColonStep j k i = match (j,k,i) with
	(IVal(I32(j'),_),IVal(_,I32(k')),IVal(_)) ->	let i' = (match i with	
																IVal((I32(a),I32(b))) ->	Int32.to_int b
												 	 		|	IVal((I64(a),I64(b))) ->	Int64.to_int b
													 		|	_ -> raise (Error ("matrixColonStep","??? Integer expected."))
													 		)
													in
														let m = (Int32.to_int (Int32.sub k' j')) / i' in
														let a = Array.make_matrix 1 (m+1) (IVal(!i32Zero)) in
														let _ = for u=0 to m do
																	let r = Int32.add j' (Int32.of_int (u*i'))
																	in a.(0).(u) <- IVal(I32(r),I32(r)) 
																done
														in AVal(a) 
| 	(IVal(I64(j'),_),IVal(_,I64(k')),IVal(_)) ->	let i' = (match i with	
																IVal((I32(a),I32(b))) ->	Int32.to_int b
												 	 		|	IVal((I64(a),I64(b))) ->	Int64.to_int b
													 		|	_ -> raise (Error ("matrxColonStep","??? Integer expected."))
													 		)
													in
														let m = (Int64.to_int (Int64.sub k' j')) / i' in
														let a = Array.make_matrix 1 (m+1) (IVal(!i64Zero)) in
														let _ = for u=0 to m do
															let r = Int64.add j' (Int64.of_int (u*i'))
															in a.(0).(u) <- IVal(I64(r),I64(r)) 
														done
													in AVal(a) 
|	(IVal(IMInfty,_),_,_) ->				raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))	
|	(IVal(IPInfty,_),_,_) ->				raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))	
|	(IVal(IBottom,_),_,_) ->				raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))	
|	(_,IVal(_,IMInfty),_) ->				raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))	
|	(_,IVal(_,IPInfty),_) ->				raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))	
|	(_,IVal(_,IBottom),_) ->				raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))	
|	(FVal(FE((M(j'),_),_,pj)),FVal(FE((_,M(k')),_,pk)),_) ->
											let i' = (match i with	
														IVal(_,b) ->	castInt2Float64 b
													 |	FVal(FE((_,b),_,_)) -> b
													 |	_ -> raise (Error ("matrxColonStep","??? Integer expected."))
													 )
											in

											let p = max pj pk in
											let m = int_of_float (ceil (getFloat (mpFloatDiv (mpFloatMinus (M(k')) (M(j')) p GMP_RNDN) i' p GMP_RNDN))) in
											let a = Array.make_matrix 1 (m+1) (FVal(!fe64Zero)) in
											let _ = for u=0 to m do
											let rr = mpFloatMult i' (M(FR.from_float (float_of_int u))) p GMP_RNDN in
											let r = mpFloatPlus (M(j')) rr p GMP_RNDN 
														in a.(0).(u) <- FVal(FE((r,r),!eZero,p)) 
													done
											in AVal(a)
|	(FVal(FE((MMInfty,_),_,_)),_,_) ->	raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))
|	(FVal(FE((MPInfty,_),_,_)),_,_) ->	raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))
|	(FVal(FE((MBottom,_),_,_)),_,_) ->	raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))
|	(FVal(FE((_,MMInfty),_,_)),_,_) ->	raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))
|	(FVal(FE((_,MPInfty),_,_)),_,_) ->	raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))
|	(FVal(FE((_,MBottom),_,_)),_,_) ->	raise (Error ("matrixColonstep","??? Cannot apply Colonstep to an unbound domain."))
|   (IVal(j'),IVal(k'),FVal(_)) ->	 	let j'' = castIntInterval2FE64 j' in 
  								 		let k'' = castIntInterval2FE64 k' in 
											matrixColonStep (FVal(j'')) (FVal(k'')) i
|   (IVal(j'),FVal(_),IVal(i')) ->	 	let j'' = castIntInterval2FE64 j' in 
							 			let i'' = castIntInterval2FE64 i' in 
											matrixColonStep (FVal(j'')) k (FVal(i'')) 
|   (FVal(_),IVal(k'),IVal(i')) ->	 	let i'' = castIntInterval2FE64 i' in 
								 		let k'' = castIntInterval2FE64 k' in 
											matrixColonStep j (FVal(k'')) (FVal(i'')) 
|   (IVal(j'),FVal(_),FVal(_)) ->	 	let j'' = castIntInterval2FE64 j' in 
											matrixColonStep (FVal(j'')) k i
|   (FVal(_),IVal(k'),FVal(_)) ->	 	let k'' = castIntInterval2FE64 k' in 
											matrixColonStep j (FVal(k'')) i
|   (FVal(_),FVal(_),IVal(i')) ->	 	let i'' = castIntInterval2FE64 i' in 
											matrixColonStep j k (FVal(i''))

|	_ ->								raise (Error ("matrixColonstep","??? Type mismatch in Colonstep."))
;;


let rec matrixPow m n res = if (n=1) then res else 
												let r = getArray (matrixMult m res) 
												in matrixPow m (n-1) r ;; 
	

