(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file prelude.ml contains declarations of constants and basic functions used by most of the other files
 *)

open Gmp ;;
open List ;;
open Types ;;

exception Error of string * string ;;


(*****************************************************************************)
(*                                  Constants                                *)
(*****************************************************************************)

let debug = ref false ;; (* verbose error messages *)
let theEnd = ref false ;; (* to terminate the interactive loop *)

let codeLine = ref 1 ;;   (* at parsing, indicates the current line being processed *)
let currentLine = ref 1 ;; (* at runtime, indicates the code line where error arises *)


let float128_mantissa = 113 ;;			(* size of the mantissa of a quadruple precision number  with hidden bit = 112+1 *)
let float64_mantissa = 53 ;;			(* size of the mantissa of a double precision number with hidden bit = 52+1 *)
let float32_mantissa = 24 ;;			(* size of the mantissa of a single precision number with hidden bit = 23+1 *)
let fixed_default_frac_size = ref 32 ;;     (* size by default of the fractional part of a fixed-point number *)

let errPrec = ref 256 ;;                (* size of the mantissa of multiple precision numbers used for error terms *)
let errPrecPrint = ref 53 ;;            (* used to convert error terms to strings, for display *)
let slicePrec = ref 8 ;;                (* number of slices for interval slicing *) 
let formalExprMaxHeight = ref 4 ;;		(* maximal height of formal expressions before evaluation *)
let veListLimit = ref 100 ;;            (* maximal number of equivalent formal expressions in semi-formal evaluation *)

let printBinary = ref false ;;			(* to print the numbers either in decimal or binary in the interactive loop *)
let printAbsolute = ref true ;;         (* to print error terms either as absolute or relative errors *)
let forLoopWiden = ref false ;;         (* does the system perform widenings in for loops?  *)

let pi = FE((M(FR.from_float_prec float64_mantissa GMP_RNDD 3.141592653589793),M(FR.from_float_prec float64_mantissa GMP_RNDU 3.141592653589793)),
            (M(FR.from_float_prec !errPrec GMP_RNDD         0.000000000000000),M(FR.from_float_prec !errPrec GMP_RNDU         0.000000000000001)),float64_mantissa) ;;

let typeEnv = ref (("ans",UndefinedType)::("pi",Type_Float_64)::[]) ;; 		(* global environment mapping identifiers to their types *)
let valEnv = ref (("ans",ValTop)::("pi",FVal(pi))::[]) ;;         			(* global environment mapping identifiers to their values *)
let funEnv = ref (("typeGeneralization",(Fun([],[],Nop,Lab(0))))::[]) ;; 	(* global environment mapping identifiers to functions *)

let currentTypeEnv = ref (("ans",UndefinedType)::[]) ;; (* current environment mapping identifiers to their types *)
let currentValEnv = ref (("ans",ValTop)::[]) ;;         (* current environment mapping identifiers to their values *)

let verbose = ref true ;;           				    (* do we print the result of the command? *)
let unrollLevel = ref 5 ;;								(* number of times loops are unrolled before joining environments *)
let widenLevel = ref 10;;								(* number of iterations before widening *)

let stopWiden = ref 10 ;;                               (* after stopWiden staged widening, next widening goes directly to infty *)

let noFunCallYet = ref false ;;					        (* to avoid slicing in nested fun calls: only the first call is sliced *)

let folder = ref (Sys.getcwd ()) ;;						(* current working directory - used to load .m files *)

let warnMsgDivZero = ref false ;;
let warnMsgScalarProd = ref false ;;
let warnMsgMatProd = ref false ;;

let printEnvVars () = 
	print_string ("\n") ;
	print_string ("Mantissa size of error terms (error) ............. "^(string_of_int !errPrec)^"\n") ;
	print_string ("Mantissa size of error display (error_display) ... "^(string_of_int !errPrecPrint)^"\n") ;
	print_string ("Size of the default fractional part (frac) ....... "^(string_of_int !fixed_default_frac_size)^"\n") ;
	print_string ("Base for data display (bin or dec)................ "^(if (!printBinary) then "Binary" else "Decimal")^"\n") ;
	print_string ("Error terms display (abs or rel).................. "^(if (!printAbsolute) then "Absolute" else "Relative")^"\n") ;
	print_string ("Loop unrolling level (unroll) .................... "^(string_of_int !unrollLevel)^"\n") ;
	print_string ("Widening level (widen) ........................... "^(string_of_int !widenLevel)^"\n") ;
	print_string ("For loops widening (widen_for) ................... "^(if (!forLoopWiden) then "Enabled" else "Disabled")^"\n") ;
	print_string ("Slice function calls (slice_fun) ................. "^(if (!noFunCallYet) then "Enabled" else "Disabled")^"\n") ;
	print_string ("Interval slice number (slice) .................... "^(string_of_int !slicePrec)^"\n") ;
	print_string ("Maximal formal expression height (formal) ........ "^(string_of_int !formalExprMaxHeight)^"\n") ;	
	print_string ("Current Folder ................................... '"^(!folder)^"'\n") ;
	print_string ("\n") ;;

	
let	setDefault () = 
	errPrec := 256 ;
	errPrecPrint := 53 ;
	fixed_default_frac_size := 32  ;
	unrollLevel := 5 ;
	widenLevel := 10 ;
	slicePrec := 8 ;
	noFunCallYet := false ;
	formalExprMaxHeight := 5 ;
	printBinary := false ;
	printAbsolute := true ;
	forLoopWiden := false ;;

	
let changeDir d = 
	try
		Sys.chdir d ; 
		folder := Sys.getcwd () 
	with
		e -> raise (Error ("changeDir","?? Incorrect directory.")) ;;
		
		
(*****************************************************************************)
(*                          Operations on Types                              *)
(*****************************************************************************)


(* precision i.e. size of the mantissa following the type *)
let precOfType t = match t with
	Type_Float_32 -> float32_mantissa 
|	Type_Float_64 -> float64_mantissa 
|	Type_Float_128 -> float128_mantissa 
|   Type_MP(p) -> p
|	_ -> raise (Error ("precOfType","it does not make sense to ask the precision of a non floating-point type")) ;;


let typeOfPrec m = 
  if (m=float32_mantissa) then Type_Float_32 else
  if (m=float64_mantissa) then Type_Float_64 else
  if (m=float128_mantissa) then Type_Float_128 else Type_MP(m) ;;
 
 
let rec maxType t1 t2 = match (t1,t2) with
	(Type_Int_32,Type_Int_32) -> 		Type_Int_32
|	(Type_Int_32,Type_Int_64) -> 		Type_Int_64
|	(Type_Int_64,Type_Int_32) -> 		Type_Int_64
|	(Type_Float_32,Type_Float_32) -> 	Type_Float_32
|	(Type_Float_64,Type_Float_64) -> 	Type_Float_64
|	(Type_Float_32,Type_Float_64) -> 	Type_Float_64
|	(Type_Float_64,Type_Float_32) -> 	Type_Float_64
|	(Type_Float_128,Type_Float_128) -> 	Type_Float_128
|	(Type_Float_32,Type_Float_128) -> 	Type_Float_128
|	(Type_Float_64,Type_Float_128) -> 	Type_Float_128
|	(Type_Float_128,Type_Float_32) -> 	Type_Float_128
|	(Type_Float_128,Type_Float_64) -> 	Type_Float_128
| 	(Type_Int_32,Type_Float_32) ->		Type_Float_32
| 	(Type_Int_32,Type_Float_64) ->		Type_Float_64
| 	(Type_Int_32,Type_Float_128) ->		Type_Float_128
| 	(Type_Float_32,Type_Int_32) ->		Type_Float_32
| 	(Type_Float_64,Type_Int_32) ->		Type_Float_64
| 	(Type_Float_128,Type_Int_32) ->		Type_Float_128
| 	(Type_Int_64,Type_Float_32) ->		Type_Float_32
| 	(Type_Int_64,Type_Float_64) ->		Type_Float_64
| 	(Type_Int_64,Type_Float_128) ->		Type_Float_128
| 	(Type_Float_32,Type_Int_64) ->		Type_Float_32
| 	(Type_Float_64,Type_Int_64) ->		Type_Float_64
| 	(Type_Float_128,Type_Int_64) ->		Type_Float_128
|   (Type_MP(p),Type_MP(q)) ->          Type_MP(max p q)
|   (Type_MP(p),Type_Float_32) ->       if (p<=float32_mantissa) then Type_Float_32 else Type_MP(p)  
|   (Type_Float_32,Type_MP(p)) ->       if (p<=float32_mantissa) then Type_Float_32 else Type_MP(p)  
|   (Type_MP(p),Type_Float_64) ->       if (p<=float64_mantissa) then Type_Float_64 else Type_MP(p)  
|   (Type_Float_64,Type_MP(p)) ->       if (p<=float64_mantissa) then Type_Float_64 else Type_MP(p)  
|   (Type_MP(p),Type_Float_128) ->      if (p<=float128_mantissa) then Type_Float_128 else Type_MP(p)  
|   (Type_Float_128,Type_MP(p)) ->      if (p<=float128_mantissa) then Type_Float_128 else Type_MP(p)  
|   (Type_Fixed(i1,f1),Type_Fixed(i2,f2)) -> Type_Fixed(max i1 i2,max f1 f2) 
|   (Type_Fixed(_,_),Type_Int_32) ->    t1
|   (Type_Int_32,Type_Fixed(_,_)) ->    t2
|   (Type_Fixed(_,_),Type_Int_64) ->    t1
|   (Type_Int_64,Type_Fixed(_,_)) ->    t2
|   (Type_Fixed(_,_),Type_Float_32) ->   t1
|   (Type_Float_32,Type_Fixed(_,_)) ->   t2
|   (Type_Fixed(_,_),Type_Float_64) ->   t1
|   (Type_Float_64,Type_Fixed(_,_)) ->   t2
|   (Type_Fixed(_,_),Type_Float_128) ->   t1
|   (Type_Float_128,Type_Fixed(_,_)) ->   t2
|   (Type_Array(I32(n1),I32(n2),t1'),Type_Array(I32(n3),I32(n4),t2')) -> 
		if (((Int32.compare n1 n3)=0)&&((Int32.compare n2 n4)=0)) then Type_Array(I32(n1),I32(n3),maxType t1' t2')
		else raise (Error ("maxType","??? Matrix dimensions must agree."))
|   (Type_Array(I64(n1),I64(n2),t1'),Type_Array(I64(n3),I64(n4),t2')) -> 
		if (((Int64.compare n1 n3)=0)&&((Int64.compare n2 n4)=0)) then Type_Array(I64(n1),I64(n3),maxType t1' t2')
		else raise (Error ("maxType","??? Matrix dimensions must agree."))
| 	(Type_Array(IBottom,IBottom,t1),Type_Array(IBottom,IBottom,t2)) -> 	Type_Array(IBottom,IBottom,maxType t1 t2)
| 	(Type_Array(_),Type_Array(_)) -> 	raise (Error ("maxType","??? Matrix dimensions must be integers."))
|	(Type_Array(a,b,t1'),_) -> Type_Array(a,b,maxType t1' t2)
|	(_,Type_Array(a,b,t2')) -> Type_Array(a,b,maxType t1 t2')
|	(UndefinedType,_) -> 				t2
|	(_,UndefinedType) -> 				t1 
|	_ -> 								raise (Error ("maxType","??? Types do not agree.")) ;;


let maxTypeMatrixMult t1 t2 = match (t1,t2) with
    (Type_Array(I32(n1),I32(n2),t1'),Type_Array(I32(n3),I32(n4),t2')) -> 
	if ((Int32.compare n2 n3)=0) then Type_Array(I32(n1),I32(n4),maxType t1' t2')
		else raise (Error ("maxTypeMM","??? Matrix dimensions must agree."))
|   (Type_Array(I64(n1),I64(n2),t1'),Type_Array(I64(n3),I64(n4),t2')) -> 
		if ((Int64.compare n2 n3)=0) then Type_Array(I64(n1),I64(n4),maxType t1' t2')
		else raise (Error ("maxTypeMM","??? Matrix dimensions must agree."))
|   (Type_Array(I32(n1),I32(n2),t1'),Type_Array(I64(n3),I64(n4),t2')) ->
	 	let n1' = Int64.of_int32 n1 in
		let n2' = Int64.of_int32 n2 in
		if ((Int64.compare n2' n3)=0) then Type_Array(I64(n1'),I64(n4),maxType t1' t2')
		else raise (Error ("maxTypeMM","??? Matrix dimensions must agree."))
|   (Type_Array(I64(n1),I64(n2),t1'),Type_Array(I32(n3),I32(n4),t2')) ->
		let n3' = Int64.of_int32 n3 in
		let n4' = Int64.of_int32 n4 in
		if ((Int64.compare n2 n3')=0) then Type_Array(I64(n1),I64(n4'),maxType t1' t2')
		else raise (Error ("maxTypeMM","??? Matrix dimensions must agree."))
|   (Type_Array(I32(n1),I32(n2),t1'),Type_Array(IBottom,IBottom,t2')) -> 
		Type_Array(I32(n1),I32(n2),maxType t1' t2')
|   (Type_Array(IBottom,IBottom,t1'),Type_Array(I32(n1),I32(n2),t2')) -> 
		Type_Array(I32(n1),I32(n2),maxType t1' t2')
|   (Type_Array(I64(n1),I64(n2),t1'),Type_Array(IBottom,IBottom,t2')) -> 
		Type_Array(I64(n1),I64(n2),maxType t1' t2')
|   (Type_Array(IBottom,IBottom,t1'),Type_Array(I64(n1),I64(n2),t2')) -> 
		Type_Array(I64(n1),I64(n2),maxType t1' t2')
|   _ -> maxType t1 t2 ;;


let rec isArray id tEnv = match tEnv with
	[] -> false
| 	(id',Type_Array(_))::tEnv' -> if ((String.compare id id')=0) then true else isArray id tEnv'
| 	(id',_)::tEnv' -> if ((String.compare id id')=0) then false else isArray id tEnv' ;;


let rec isVArray id vEnv = match vEnv with
	[] -> false
| 	(id',AVal(_))::vEnv' -> if ((String.compare id id')=0) then true else isVArray id vEnv'
| 	(id',_)::vEnv' -> if ((String.compare id id')=0) then false else isVArray id vEnv' ;;


let rec isFun id tEnv = match tEnv with
	[] -> false
| 	(id',Type_Fun(_))::tEnv' -> if ((String.compare id id')=0) then true else false
| 	(id',_)::tEnv' -> if ((String.compare id id')=0) then false else isFun id tEnv' ;;


let funExists f = Sys.file_exists (f^".m") ;;


let rec getReturnType id tEnv = match tEnv with
	[] -> 						raise (Error("getReturnType","??? Undefined identifier '"^id^"'.")) 
|	(id',Type_Fun(t))::tEnv' -> if ((String.compare id id')=0) then t
								else raise (Error("getReturnType","??? '"^"id"^"' is not a function."))
|	(id',_)::tEnv' -> 			if ((String.compare id id')=0) then
									raise (Error("getReturnType","??? '"^"id"^"' is not a function."))
								else getReturnType id tEnv' ;;
								
								
(*****************************************************************************)
(*                                 Environments                              *)
(*****************************************************************************)

let copyArray a =		
	let a' = Array.copy a in
	let _ =	for i=0 to ((Array.length a)-1) do
				a'.(i) <- Array.copy a.(i)
			done
	in a' ;;

	
let getEnv id env = 
  try 
    assoc id env
  with
    Not_found -> raise (Error ("getEnv","??? Undefined identifier '"^id^"'.")) ;;


let getEnvVal id env = 
  try 
    let res = assoc id env 
	in (match res with
			AVal(a) -> AVal(copyArray a)
		| _ -> res
		)
  with
    Not_found -> raise (Error ("getEnv","??? Undefined identifier '"^id^"'.")) ;;
	
	
let getEnvComplete id env e = 
  try 
    assoc id env
  with
    Not_found -> e ;;


let rec setEnvAux env id v aux = match env with
    [] -> (id,v)::aux
  | (id',v')::env' -> if id=id' then ((id,v)::(env'@aux))
                           else setEnvAux env' id v ((id',v')::aux) ;;                      

						   
let setEnv env id v = setEnvAux env id v [] ;;


let rec copyEnv env = match env with
	[] -> []
| 	(x,AVal(a))::xs -> (x,AVal(copyArray a))::(copyEnv xs)
|	x::xs -> x::(copyEnv xs) ;;



(*****************************************************************************)
(*                       Extraction of fields in types                       *)
(*****************************************************************************)

let getInt i = match i with
	I32(x) -> 	Int32.to_int x
|	I64(x) -> 	Int64.to_int x
|	_ ->		raise (Error ("getInt","??? Integer expression expected.")) ;;

				
let getArray a = match a with
	AVal(v) -> v
|	 _ -> raise (Error ("getArray","??? Matrix expected.")) ;;


let getFE f = match f with
	FVal(fe) -> fe
|	_ -> raise (Error ("getFE","??? Floating-point expression expected.")) ;;


let getFX f = match f with
	XVal(fe) -> fe
|	_ -> raise (Error ("getFX","??? Fixed-point expression expected.")) ;;


let getFEPrec f = match f with
	FVal(FE(_,_,p)) -> p
|	_ -> raise (Error ("getFEPrec","??? Floating-point expression expected.")) ;;


let rec getArrayType t = match t with
	Type_Array(_,_,t') -> t'
|	_ -> raise (Error ("getArrayType","??? Matrix expected.")) ;;


let rec getArrayLineLength t = match t with
	Type_Array(_,i,_) -> i
	|	_ -> raise (Error ("getArrayLineLength","??? Matrix expected.")) ;;


let rec getArrayLength t = match t with
	Type_Array(i,_,_) -> i
|	_ -> raise (Error ("getArrayLineLength","??? Matrix expected.")) ;;


let getId e = match e with
	Id(s,_) -> s
|	_ -> raise (Error ("getId","??? Identifier expected.")) ;;


let getFloat f = match f with
	M(x) -> FR.to_float x
|	_ ->	raise (Error ("getFloat","??? Floating-point number expected.")) ;;


let getFixedSign f = match f with
	X(s,_,_) -> s
|   XPInfty -> true
| 	XMInfty -> false 
| 	XBottom -> raise (Error ("getFixedSign","??? Bottom not expected.")) ;;
 

let getIPrec (FX(_,i,_,_)) = i ;;
let getFPrec (FX(_,_,f,_)) = f ;;


let rec typeOf v = match v with
	IVal(I32(_),I32(_)) 	-> Type_Int_32
|	IVal(I64(_),I64(_)) 	-> Type_Int_64
|	IVal(I32(_),I64(_)) 	-> raise (Error ("typeOf","??? Inconsistant bounds in interval: Int32/Int64."))
|	IVal(I64(_),I32(_)) 	-> raise (Error ("typeOf","??? Inconsistant bounds in interval: Int64/Int32."))
|	IVal(I32(_),_) 			-> Type_Int_32
|	IVal(_,I32(_)) 			-> Type_Int_32
|	IVal(I64(_),_) 			-> Type_Int_64
|	IVal(_,I64(_)) 			-> Type_Int_64
|	FVal(FE(_,_,p))			-> 	if (p=float32_mantissa) then Type_Float_32
								else
									if (p=float64_mantissa) then Type_Float_64
									else
										if (p=float128_mantissa) then Type_Float_128
										else
											Type_MP(p)
|   XVal(FX(_,ip,fp,_))		-> Type_Fixed(ip,fp)									
|	BVal(_)					-> Type_Bool
|	AVal(a)					-> 	let t = typeOf a.(0).(0) in
								let y = Array.length a in
								let x = Array.length a.(0) in
									if ((x=1)&&(y=1)) then t 
									else
										Type_Array(I32(Int32.of_int y),I32(Int32.of_int x),t) 
|	_						-> UndefinedType 
;;
	
	
let rec maxTypeOfValList l res = match l with
	[] ->  	 res
|  	v::vs -> maxTypeOfValList vs (maxType (typeOf v) res) ;;
 

let fValFst x = match x with
  FVal(FE((a,_),_,_)) -> a 
| _ -> raise (Error ("fValFst","impossible case.")) ;;


let fValSnd x = match x with
  FVal(FE((_,a),_,_)) -> a 
| _ -> raise (Error ("fValFst","impossible case.")) ;;



(*****************************************************************************)
(*                                  Misc                                     *)
(*****************************************************************************)


let extractMantissaAndExp s = 
	let b = String.index s 'b' in
	let m = String.sub s 0 b in
	let m' = M(FR.from_string_prec_base float64_mantissa GMP_RNDD 10 m) in
	let m'' = M(FR.from_string_prec_base float64_mantissa GMP_RNDU 10 m) in
	let e = String.sub s (b+1) ((String.length s) -b-1) in
		(m',m'',int_of_string e) ;;
		
		
let computeMP s = 		
	let i = String.index s '#' in 
	let s' = String.sub s 0 i in 
	let prec = int_of_string (String.sub s (i+1) ((String.length s)-i-1)) in
		((M(FR.from_string_prec_base prec GMP_RNDD 10 s'),
		 M(FR.from_string_prec_base prec GMP_RNDU 10 s')),Type_MP(prec)) ;;


let rec list2Seq cl = match cl with
	[] -> Nop
| 	c::[] -> c
|	c::cs -> Seq(c,list2Seq cs,Lab(0)) ;; 


let rec setNth x n l =
 try
	if (n=0) then 
	 	(x::(List.tl l))
	else
		((List.hd l)::(setNth x (n-1) (List.tl l))) 
 with _ -> raise (Error ("setNth","??? cannot insert in nth position of list.")) ;;
		

let showC () =		
	print_string "\n  This program is free software: you can redistribute it and/or modify";
	print_string "\n  it under the terms of the GNU General Public License as published by";
	print_string "\n  the Free Software Foundation, either version 3 of the License, or";
	print_string "\n  (at your option) any later version.\n\n" ;;


let showW () =
	print_string "\n  This program is distributed in the hope that it will be useful,";
	print_string "\n  but WITHOUT ANY WARRANTY; without even the implied warranty of";
	print_string "\n  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the";
	print_string "\n  GNU General Public License for more details.\n\n" ;;


let show s = 
	if ((String.compare s "c")=0) then showC () else
	if ((String.compare s "w")=0) then showW () else 
		raise (Error ("show","??? Illegal argument '"^s^"'")) ;;

(*****************************************************************************)
(*         Constants for the zeros etc of the interval domains               *)
(*****************************************************************************)

let f32Zero = ref (M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0),M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0)) ;;
let f64Zero = ref (M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0),M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0)) ;;
let f128Zero = ref (M(FR.from_float_prec float128_mantissa GMP_RNDN 0.0),M(FR.from_float_prec float128_mantissa GMP_RNDN 0.0)) ;;

let fErrZero = ref (M(FR.from_float_prec !errPrec GMP_RNDN 0.0),M(FR.from_float_prec !errPrec GMP_RNDN 0.0)) ;;

let f32One = ref (M(FR.from_float_prec float32_mantissa GMP_RNDN 1.0),M(FR.from_float_prec float32_mantissa GMP_RNDN 1.0)) ;;
let f64One = ref (M(FR.from_float_prec float64_mantissa GMP_RNDN 1.0),M(FR.from_float_prec float64_mantissa GMP_RNDN 1.0)) ;;
let f128One = ref (M(FR.from_float_prec float128_mantissa GMP_RNDN 1.0),M(FR.from_float_prec float128_mantissa GMP_RNDN 1.0)) ;;

let f32MinusOne = ref (M(FR.from_float_prec float32_mantissa GMP_RNDN (-1.0)),M(FR.from_float_prec float32_mantissa GMP_RNDN (-1.0))) ;;
let f64MinusOne = ref (M(FR.from_float_prec float64_mantissa GMP_RNDN (-1.0)),M(FR.from_float_prec float64_mantissa GMP_RNDN (-1.0))) ;;
let f128MinusOne = ref (M(FR.from_float_prec float128_mantissa GMP_RNDN (-1.0)),M(FR.from_float_prec float128_mantissa GMP_RNDN (-1.0))) ;;

let eZero = ref (M(FR.from_float_prec !errPrec GMP_RNDN 0.0),M(FR.from_float_prec !errPrec GMP_RNDN 0.0)) ;;

let i32Zero = ref (I32(Int32.zero),I32(Int32.zero)) ;;
let i64Zero = ref (I64(Int64.zero),I64(Int64.zero)) ;;

let i32One = ref (I32(Int32.one),I32(Int32.one)) ;;
let i64One = ref (I64(Int64.one),I64(Int64.one)) ;;

let feBottom = ref (FE((MBottom,MBottom),(MBottom,MBottom),-1)) ;;
let feTop = ref (FE((MMInfty,MPInfty),(MMInfty,MPInfty),-1)) ;;

let fe32Zero = ref (FE(!f32Zero,!eZero,float32_mantissa)) ;;
let fe64Zero = ref (FE(!f64Zero,!eZero,float64_mantissa)) ;;
let fe128Zero = ref (FE(!f128Zero,!eZero,float128_mantissa)) ;;


let fe32One = ref (FE(!f32One,!eZero,float32_mantissa)) ;;
let fe64One = ref (FE(!f64One,!eZero,float64_mantissa)) ;;
let fe128One = ref (FE(!f128One,!eZero,float128_mantissa)) ;;

let fe32MinusOne = ref (FE(!f32MinusOne,!eZero,float32_mantissa)) ;;
let fe64MinusOne = ref (FE(!f64MinusOne,!eZero,float64_mantissa)) ;;
let fe128MinusOne = ref (FE(!f128MinusOne,!eZero,float128_mantissa)) ;;

let fe128ZeroInfty = ref (FE((M(FR.from_float_prec float128_mantissa GMP_RNDN 0.0),MPInfty),(MMInfty,MPInfty),float128_mantissa)) ;;
let fe128MInftyZero = ref (FE((MMInfty,M(FR.from_float_prec float128_mantissa GMP_RNDN 0.0)),(MMInfty,MPInfty),float128_mantissa)) ;;

let zeroPrec p = 
	if (p=float32_mantissa) then FVal(!fe32Zero) else
	if (p=float64_mantissa) then FVal(!fe64Zero) else
	if (p=float128_mantissa) then FVal(!fe128Zero) else
		raise (Error ("zeroPrec","??? Unexpected precision.")) ;;
		
