(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file print.ml contains all the printing functions 
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Matrix ;;


(*****************************************************************************)
(*                                    misc                                   *)
(*****************************************************************************)

let rec printType t = match t with
	Type_Int_32 -> "int32"
| 	Type_Int_64 -> "int64"
|	Type_Float_32 -> "float32"
|	Type_Float_64 -> "float64"
|	Type_Float_128 -> "float128"
|   Type_MP(p) -> "mp("^(string_of_int p)^")"
|   Type_Fixed(ip,fp) -> "fixed("^(string_of_int ip)^","^(string_of_int fp)^")"
|	Type_Bool -> "Boolean"
| 	Type_Array(_,_,t') -> ((printType t')^"[]")
|	Type_Fun(t') ->	((printType t')^" function")
| 	UndefinedType -> "" ;;


let printWhos id = match (getEnv id !typeEnv) with
	Type_Int_32 -> id^"      "^"int32\n"
| 	Type_Int_64 -> id^"      "^"int64\n"
|	Type_Float_32 -> id^"      "^"float32\n"
|	Type_Float_64 -> id^"      "^"float64\n"
|	Type_Float_128 -> id^"      "^"float128\n"
|   Type_MP(p) -> id^"      "^"mp("^(string_of_int p)^")\n"
|   Type_Fixed(ip,fp) -> id^"      "^"fixed("^(string_of_int ip)^","^(string_of_int fp)^")\n"
|	Type_Bool -> id^"      "^"Boolean\n"
|	Type_Array(I32(i),I32(j),t') -> id^"      "^(printType t')^"["^(Int32.to_string i)^","^(Int32.to_string j)^"]\n"
|	Type_Array(I64(i),I64(j),t') -> id^"      "^(printType t')^"["^(Int64.to_string i)^","^(Int64.to_string j)^"]\n"
|	Type_Array(I32(i),I64(j),t') -> id^"      "^(printType t')^"["^(Int32.to_string i)^","^(Int64.to_string j)^"]\n"
|	Type_Array(I64(i),I32(j),t') -> id^"      "^(printType t')^"["^(Int64.to_string i)^","^(Int32.to_string j)^"]\n"	
|	Type_Fun(t') ->	id^"      "^((printType t')^" function\n")
| 	Type_Array(_,_,t') -> (id^"      "^(printType t')^"[]\n")
| 	UndefinedType -> ""  ;;


(*****************************************************************************)
(*                                  Conversions                              *)
(*****************************************************************************)

let rec binStringOfInt32 i aux = 
  if ((Int32.compare i Int32.zero)=0) then 
		aux 
  else
  		binStringOfInt32 (Int32.div i (Int32.of_int 2))
  						 ((if ((Int32.compare (Int32.rem i (Int32.of_int 2)) Int32.zero)=0) then "0" else "1")^aux) ;;


let rec binStringOfInt64 i aux = 
  if ((Int64.compare i Int64.zero)=0) then 
		aux 
  else
  		binStringOfInt64 (Int64.div i (Int64.of_int 2))
  						 ((if ((Int64.compare (Int64.rem i (Int64.of_int 2)) Int64.zero)=0) then "0" else "1")^aux) ;;


let rec pow2 n aux = if (n<=0) then aux else pow2 (n-1) (2*aux)


let rec binaryExponentAux n m e str = 
	if (m=e) then 
		if (n>0) then (str^(String.make n '0')) else str
	else
		if ((m+(pow2 (n-1) 1))>e) then
			binaryExponentAux (n-1) m e (str^"0")
		else
			binaryExponentAux (n-1) (m+(pow2 (n-1) 1)) e (str^"1") ;; 


let binaryExponent e prec = 
	if (prec=float32_mantissa) then binaryExponentAux 8 0 (127+e) "" else
	if (prec=float64_mantissa) then binaryExponentAux 11 0 (1023+e) "" else
	if (prec=float128_mantissa) then binaryExponentAux 15 0 (16383+e) ""
	else 
		let r = binaryExponentAux 30 0 (abs e) "" in
		try
			let i = String.index r '1'  
			in ((if (e<0) then "-" else "+")^(String.sub r i ((String.length r)-i)))
		with
			Not_found -> "0" ;;
			

let binaryZero prec =                           
	if (prec=float32_mantissa) then "0 00000000 00000000000000000000000" else
	if (prec=float64_mantissa) then "0 00000000000 0000000000000000000000000000000000000000000000000000" else
	if (prec=float128_mantissa) then "0 000000000000000 0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
	else ("0 [0] "^(String.make (prec-1) '0')) ;;
		

let binaryFloat f r prec =
	if ((FR.cmp f (FR.from_float_prec prec GMP_RNDN 0.0))=0) then
		binaryZero prec
	else
		let fString = FR.to_string_base_digits r 2 prec f in
		let s = if ((String.get fString 0)='-') then "1" else "0" in
		let fString' = if ((String.get fString 0)='-') then (String.sub fString 1 ((String.length fString)-1)) else fString in
		let i = String.index fString' 'E' in
		let e = String.sub fString' (i+1) ((String.length fString')-i-1) in
		let e' = int_of_string e in
		let e'' = binaryExponent e' prec in
		let m = String.sub fString' 2 (i-2) 
		in (s^" "^e''^" "^m) ;;


(*****************************************************************************)
(*                                  Print values                             *)
(*****************************************************************************)


let printI i = match i with
	I32(a) -> if !printBinary then
					let s = (String.make 32 '0')^(binStringOfInt32 a "")
				  	in ((if ((Int32.compare a Int32.zero)>=0) then "0" else "1")^(String.sub s ((String.length s)-31) 31))
			  else Int32.to_string a
| 	I64(a) -> if !printBinary then
					let s = (String.make 64 '0')^(binStringOfInt64 a "")
				  	in ((if ((Int64.compare a Int64.zero)>=0) then "0" else "1")^(String.sub s ((String.length s)-63) 63))
			  else Int64.to_string a
|	IPInfty -> "+oo"
|	IMInfty -> "-oo"
|	IBottom -> "Bottom" ;;


let printIVal (a,b) = if (a=b) then (printI a) else ("["^(printI a)^","^(printI b)^"]") ;;


let printF f r prec = match f with
 	M(a) -> if !printBinary then
				binaryFloat a r prec
			else
				FR.to_string_base_digits r 10 prec a
|	MPInfty -> "+oo"
|	MMInfty -> "-oo"
|	MBottom -> "Bottom" ;;


let printF10 f r prec = match f with
 	M(a) -> FR.to_string_base_digits r 10 prec a
|	MPInfty -> "+oo"
|	MMInfty -> "-oo"
|	MBottom -> "Bottom" ;;


let rec integerPartOfBinString s r = 
	  if ((String.length s)=0) then 
		(string_of_int r) 
	  else integerPartOfBinString (String.sub s 1 ((String.length s)-1)) 
	                      		  ((r*2)+(int_of_string (String.sub s 0 1)));;
	

let rec fracPartOfBinString s r = 
	  if ((String.length s)=0) then 
		let rs = string_of_float r in
		let i = String.index rs '.' in
			String.sub rs (i+1) ((String.length rs)-i-1)
	  else fracPartOfBinString (String.sub s 0 ((String.length s)-1)) 
	                    	   (0.5*.(r+.(float_of_string (String.sub s ((String.length s)-1) 1)))) ;;


let printX x ip fp = match x with
	X(s,i,f) -> if !printBinary then
					((if s then "" else "-")^i^"_"^f)
				else
 					let is = integerPartOfBinString i 0 in
					let fs = fracPartOfBinString f 0.0 in
							((if s then "" else "-")^is^"_"^fs) 
|	XPInfty -> "+oo"
|	XMInfty -> "-oo"
|	XBottom -> "Bottom" ;;


let stringOfFloatInterval (a,b) prec =
	let prec' = if !printBinary then prec else (3*prec/10+1) 
    in if (mpFloatEq a b) then
			printF a GMP_RNDN prec'
	   else
			let sa = printF a GMP_RNDD prec' in
			let sb = printF b GMP_RNDU prec' in 
				if ((String.compare sa sb)=0) then sa else ("["^sa^","^sb^"]") ;;


let stringOfFloatIntervalRelError f e prec = 
	let _ = warnMsgDivZero := true in
	let (a,b) = if ((mpfiEq e !eZero)=(B(true))) then !eZero else (mpfiAbs (mpfiDiv e f !errPrec OUTSIDE)  !errPrec OUTSIDE)
	in if (a=b) then ((printF a GMP_RNDN prec)^"%") else 
		("["^(printF a GMP_RNDD prec)^","^(printF b GMP_RNDU prec)^"]%") ;;


let stringOfFloatIntervalErr (a,b) prec = 
	if (a=b) then printF10 a GMP_RNDN prec else 
		("["^(printF10 a GMP_RNDD prec)^","^(printF10 b GMP_RNDU prec)^"]") ;;


let stringOfFixedInterval (a,b) ip fp = 
	if (fixedEq a b) then printX a ip fp else 
		("["^(printX a ip fp)^","^(printX b ip fp)^"]") ;;


let printBVal b = match b with
  B(b) -> if b then "true" else "false"
| BTop -> "Top"
| BBottom -> "Bottom" ;;  

   
let printFVal (FE(f,e,prec)) = ("{"^(stringOfFloatInterval f prec)^","^(stringOfFloatIntervalErr e (3*(!errPrecPrint)/10+1))^"}") ;;

let printXVal (FX(f,ip,fp,e)) = ("{"^(stringOfFixedInterval f ip fp)^","^(stringOfFloatIntervalErr e (3*(!errPrecPrint)/10+1))^"}") ;;


let rec printVal v = match v with
 	IVal(v) -> printIVal v
|	FVal(v) -> printFVal v
| 	XVal(v) -> printXVal v
| 	BVal(v) -> printBVal v
|	AVal(v) -> printAVal v 
|   ValBottom -> "Bottom"
|   ValTop  -> "Top"
|   ArrayLines(_) -> raise (Error ("printVal","No arrayLines expected"))
|   ArrayColumns(_) -> raise (Error ("printVal","No arrayColumns expected"))


and printAVal a = Array.fold_left (fun s' al -> ((Array.fold_left (fun s v -> ((printVal v)^" "^s)) "\n" al))^s') "\n" a ;;

	  
let printFValInteractive (FE(f,e,prec)) n = 
	if !printAbsolute then
		((stringOfFloatInterval f prec)^"\n"^(String.make (n-7) ' ')^"error: "^(stringOfFloatIntervalErr e (3*(!errPrecPrint)/10+1))) 
	else
		((stringOfFloatInterval f prec)^"\n"^(String.make (n-7) ' ')^"error: "^(stringOfFloatIntervalRelError f e (3*(!errPrecPrint)/10+1))) 

let printXValInteractive (FX(f,ip,fp,e)) n = 
	if !printAbsolute then
		((stringOfFixedInterval f ip fp)^"\n"^(String.make (n-7) ' ')^"error: "^(stringOfFloatIntervalErr e (3*(!errPrecPrint)/10+1))) 
	else
		((stringOfFixedInterval f ip fp)^"\n"^(String.make (n-7) ' ')^"error: "^(stringOfFloatIntervalRelError (mpfiOfFixedInt f) e (3*(!errPrecPrint)/10+1))) ;;

let printFValFloat v = match v with
		FVal(FE(f,_,p)) -> stringOfFloatInterval f p 
	|	_ -> raise (Error ("printFValFloat","??? Inconsitent types in array.")) ;;


let printXValFixed v = match v with
		XVal(FX((x,y),ip,fp,_)) -> printX x ip fp
	|	_ -> raise (Error ("printXValFixed","??? Inconsitent types in array.")) ;;


let printXValD v = match v with
		XVal(FX((x,y),ip,fp,_)) -> printX x ip fp
	|	_ -> raise (Error ("printXValD","??? Inconsitent types in array.")) ;;

let printXValU v = match v with
		XVal(FX((x,y),ip,fp,_)) -> printX y ip fp
	|	_ -> raise (Error ("printXValU","??? Inconsitent types in array.")) ;;

	
let printFValFloatD v = match v with
		FVal(FE((d,_),_,p)) -> printF d GMP_RNDD (3*p/10+1) 
	|	_ -> raise (Error ("printFValFloatD","??? Inconsitent types in array.")) ;;


let printFValFloatU v = match v with
		FVal(FE((_,u),_,p)) -> printF u GMP_RNDU (3*p/10+1) 
	|	_ -> raise (Error ("printFValFloatU","??? Inconsitent types in array.")) ;;
	
	
let printFValError v = match v with
		FVal(FE(_,e,_)) -> stringOfFloatIntervalErr e (3*(!errPrecPrint)/10+1)
	| 	XVal(FX((_,_),_,_,e)) -> stringOfFloatIntervalErr e (3*(!errPrecPrint)/10+1)
	|	_ -> raise (Error ("printFValError","??? Inconsitent types in array.")) ;;


let printFValErrorRel v = match v with
		FVal(FE(f,e,_)) -> stringOfFloatIntervalRelError f e (3*(!errPrecPrint)/10+1)
	| 	XVal(FX((x,y),_,_,e)) -> stringOfFloatIntervalRelError ((mpFloatOfFixed x !errPrec),(mpFloatOfFixed y !errPrec)) e (3*(!errPrecPrint)/10+1)
	|	_ -> raise (Error ("printFValRelError","??? Inconsitent types in array.")) ;;

	
let printFValErrorD v = match v with
		FVal(FE(_,(d,_),_)) -> printF d GMP_RNDD (3*(!errPrecPrint)/10+1) 
	| 	XVal(FX((_,_),_,_,(d,_))) -> printF d GMP_RNDD (3*(!errPrecPrint)/10+1)
	|	_ -> raise (Error ("printFValErrorD","??? Inconsitent types in array.")) ;;


let printFValErrorU v = match v with
		FVal(FE(_,(_,u),_)) -> printF u	GMP_RNDU (3*(!errPrecPrint)/10+1) 
	| 	XVal(FX((_,_),_,_,(_,u))) -> printF u GMP_RNDU (3*(!errPrecPrint)/10+1)
	|	_ -> raise (Error ("printFValErrorU","??? Inconsitent types in array.")) ;;
	
	
(* Only is elements are singletons *)	
let printFValArrayFloatSingletons a n =
			(Array.fold_right (fun al s' -> ((Array.fold_right (fun v s -> ((printFValFloat v)^" "^s)) 
															   al ("\n"^(String.make n ' ')))
											)^s') 
							a "\n") ;;


let printXValArrayFixedSingletons a n =
			(Array.fold_right (fun al s' -> ((Array.fold_right (fun v s -> ((printXValFixed v)^" "^s)) 
															   al ("\n"^(String.make n ' ')))
											)^s') 
							a "") ;;
						
						
let printFValArrayErrorSingletons a n =				
			(Array.fold_right (fun al s' -> ((Array.fold_right (fun v s -> ((printFValError v)^" "^s)) 
															   al ("\n"^(String.make n ' ')))
											)^s') 
							a (if !printAbsolute then "" else "%")) ;; 
			
			
let rec isEveryValSingleton a i j =
	if (i=(Array.length a)) then true
	else
		if (j=(Array.length a.(i))) then
			isEveryValSingleton a (i+1) 0
		else
			(match a.(i).(j) with
				FVal(FE((x,y),_,p)) -> if (mpFloatEq x y) then 
											isEveryValSingleton a i (j+1)
										else false
			|   XVal(FX((x,y),_,_,_)) -> if (fixedEq x y) then
											isEveryValSingleton a i (j+1)
									else false
			| 	IVal(x,y) -> if (integerEq x y) then 
											isEveryValSingleton a i (j+1)
										else false
			|	_ -> raise (Error ("isEveryFloatSingleon","??? Inconsistent types in matrix."))
			) ;;


let rec isEveryErrorSingleton a i j =
	if (i=(Array.length a)) then true
	else
		if (j=(Array.length a.(i))) then
			isEveryErrorSingleton a (i+1) 0
		else
			(match a.(i).(j) with
				FVal(FE(_,(x,y),p)) -> if (mpFloatEq x y) then 
											isEveryErrorSingleton a i (j+1)
										else false
			| 	IVal(_) -> isEveryErrorSingleton a i (j+1)
			|   XVal(FX(_,_,_,(x,y))) -> if (mpFloatEq x y) then 
											isEveryErrorSingleton a i (j+1)
										 else false
			|	_ -> raise (Error ("isEveryErrorSingleon","??? Inconsistent types in matrix."))
			) ;;

			
let printFValArrayLine al printFunction =		
	Array.fold_right (fun v s -> ((printFunction v)^" "^s)) al "\n" ;;
			

let rec printFValArrayLines a n i printFunction =
	if (i=(Array.length a)-1) then 
		((printFValArrayLine a.(i) printFunction))	
	else
		((printFValArrayLine a.(i) printFunction)^(String.make (n+1) ' ')^(printFValArrayLines a n (i+1) printFunction)) ;;


let printFValArray a n = 
			(if (isEveryValSingleton a 0 0) then
					(printFValArrayFloatSingletons a n)
			 else
					"["^(printFValArrayLines a n 0 printFValFloatD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValFloatU)^(String.make (n) ' ')^"]\n"
			)^(String.make (n-8) ' ')^"Errors: "^	
			(if (isEveryErrorSingleton a 0 0) then
					(printFValArrayErrorSingletons a n)
			 else		
					"["^(printFValArrayLines a n 0 printFValErrorD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValErrorU)^(String.make (n) ' ')^"]" 
			);;


let printXValArray a n = 
			(if (isEveryValSingleton a 0 0) then
					(printXValArrayFixedSingletons a n)
			 else 
					"["^(printFValArrayLines a n 0 printXValD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printXValU)^(String.make (n) ' ')^"]\n"
			)^(String.make (n-8) ' ')^"Errors: "^	 
			(if (isEveryErrorSingleton a 0 0) then
					(printFValArrayErrorSingletons a n)
			 else		
					"["^(printFValArrayLines a n 0 printFValErrorD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValErrorU)^(String.make (n) ' ')^"]" 
			);;


let abs2RelArray a =
	let m = Array.make_matrix (Array.length a) (Array.length a.(0)) ValBottom in
	let _ = for i=0 to ((Array.length a)-1) do
				for j=0 to ((Array.length a.(0))-1) do
					  (match a.(i).(j) with
					 			FVal(FE(f,e,p)) ->  let r = (mpfiAbs (mpfiDiv e f !errPrec OUTSIDE)  !errPrec OUTSIDE)
													in m.(i).(j) <- FVal(FE(f,r,p))
							|   XVal(FX((x,y),ip,fp,e)) -> let r = (mpfiAbs (mpfiDiv e ((mpFloatOfFixed x !errPrec),(mpFloatOfFixed y !errPrec)) !errPrec OUTSIDE)  !errPrec OUTSIDE)
										 				 in m.(i).(j) <- XVal(FX((x,y),ip,fp,r))
							| _ -> raise (Error ("abs2Rel","??? unexpected type for array element."))
					  )
				done
			done
	in m ;;


let printFValArrayRel aa n = 
			let a = abs2RelArray aa in
			(if (isEveryValSingleton a 0 0) then
					(printFValArrayFloatSingletons a n)
			 else
					"["^(printFValArrayLines a n 0 printFValFloatD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValFloatU)^(String.make (n) ' ')^"]\n"
			)^(String.make (n-8) ' ')^"Errors: "^	
			(if (isEveryErrorSingleton a 0 0) then
					(printFValArrayErrorSingletons a n)
			 else		
					"["^(printFValArrayLines a n 0 printFValErrorD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValErrorU)^(String.make (n) ' ')^"]%" 
			);;


let printXValArrayRel aa n = 
			let a = abs2RelArray aa in
			(if (isEveryValSingleton a 0 0) then
					(printXValArrayFixedSingletons a n)
			 else
					"["^(printFValArrayLines a n 0 printFValFloatD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValFloatU)^(String.make (n) ' ')^"]\n"
			)^(String.make (n-8) ' ')^"Errors: "^	
			(if (isEveryErrorSingleton a 0 0) then
					(printFValArrayErrorSingletons a n)
			 else		
					"["^(printFValArrayLines a n 0 printFValErrorD)^(String.make (n) ' ')^",\n"^(String.make n ' ')^
					" "^(printFValArrayLines a n 0 printFValErrorU)^(String.make (n) ' ')^"]%" 
			);;
		
		
(* different print in the interactive loop *)
let rec printValInteractive vi n = match vi with
 	IVal(v) -> printIVal v
|	FVal(v) -> printFValInteractive v n
|	XVal(v) -> printXValInteractive v n
| 	BVal(v) -> printBVal v 
|	AVal(v) -> let x = try v.(0).(0) 
					   with Invalid_argument(_) -> raise (Error ("printValInteractive","??? Array expected.")) 
				in 
				(match x with
					FVal(_) -> if (!printAbsolute) then (printFValArray v n) else (printFValArrayRel v n)
				 |  XVal(_) -> if (!printAbsolute) then (printXValArray v n) else (printXValArrayRel v n)
				 |	_ -> printAValInteractive v n
				)
|   ValBottom -> "Bottom"
|   ValTop  -> "Top"
|   ArrayLines(_) -> raise (Error ("printValInter","No arrayLines expected"))
|   ArrayColumns(_) -> raise (Error ("printValInter","No arrayColumns expected"))


and printAValInteractive a n = 
		Array.fold_right (fun al s' -> ((Array.fold_right (fun v s -> ((printValInteractive v n)^" "^s)) 
														al ("\n"^(String.make n ' ')) )
									   )^s') 
						a "" ;; 

	  
let printValInteractive id typeEnv valEnv =  
	print_string ("\n"^id^" = \n") ;
	let s = (String.make ((String.length id)+4) ' ')^(printType (getEnv id typeEnv))^": " 
	in
		print_string s ; flush stdout;
		print_string ((printValInteractive (getEnv id valEnv) (String.length s))^"\n\n") ;
		flush stdout ;;
  
