(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file semiformal.ml contains all the operations on semi-formal expressions
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Coercion ;;
open Matrix ;;
open Slice ;;

					
(*****************************************************************************)
(*                                  FORMAL                                   *)
(*****************************************************************************)					

		
let rec combineTwoOpAux l1 l2 funOp = match l1 with
  [] -> []
| ve::ves -> ((List.map (fun x -> (funOp (ve,x))) l2)@(combineTwoOpAux ves l2 funOp)) ;;


let combineTwoOp l1 l2 op = 
  let funOp = match op with
               OP_PLUS -> 		fun (x,y) -> VEPlus(x,y)
             | OP_MINUS -> 		fun (x,y) -> VEMinus(x,y)
             | OP_MULT -> 		fun (x,y) -> VETimes(x,y)
             | OP_MAT_MULT -> 	fun (x,y) -> VETimes(x,y)
             | OP_DOT_MULT ->	fun (x,y) -> VEDotTimes(x,y)  
             | OP_DOT_DIV ->	fun (x,y) -> VEDotDiv(x,y)  
			 | OP_DIV ->		fun (x,y) -> VEDiv(x,y)
			 | OP_POW ->		fun (x,y) -> VEPower(x,y)
			 | OP_UMINUS -> 	raise (Error ("combineTwoOp","Uminus not expected here"))
  in combineTwoOpAux l1 l2 funOp ;;
  
  
let rec evalVEPlus ve ve' tEnv = 
	let l1 = match (ve,ve') with
			(***** 0 + e -> e *)
			(VECst(v),_) ->	(try
								let v' = cast v Type_Float_64 in
		      					(match v' with
								FVal(FE((M(a),M(b)),(M(c),M(d)),p)) ->
									if (((FR.cmp a (FR.from_float_prec p GMP_RNDN 0.0))=0) &&
										((FR.cmp b (FR.from_float_prec p GMP_RNDN 0.0))=0) &&
										((FR.cmp c (FR.from_float_prec !errPrec GMP_RNDN 0.0))=0) &&
										((FR.cmp d (FR.from_float_prec !errPrec GMP_RNDN 0.0))=0)) then [ve'] 
									else
											[]
								| _ -> raise (Error ("evalVEPlus","??? Cannot happen."))
								)
							with  Error(_) -> []
								| e -> raise e
							)
			| _ -> []
	in
			insertList l1 [(VEPlus(ve,ve'));(VEPlus(ve',ve))]


and evalVEMinus ve ve' tEnv = 
	let l1 =	match (ve,ve') with
					(VECst(c),VECst(c')) -> [VECst(valMinus c c')]
				| _ -> []
	in
	let l2 = if (ve=ve') then
		 		let resType = typeOfVE ve tEnv in
				let res = (match resType with
								UndefinedType -> VECst(IVal(!i32Zero))
					 		|   _ -> VECst(cast (IVal(!i32Zero)) resType)
							)
				in
					[res;VEMinus(ve,ve')] 
			else [VEMinus(ve,ve')] 
	in
	let l3 = match (ve,ve') with
				(VEFunCall(f1,ael1),VEFunCall(f2,ael2)) -> 	if ((String.compare f1 f2)=0) then
																[VEDeltaFun(f1,ael1,ael2)]
															else []
			 | _ -> []	
	in
		insertList l1 (insertList l2 l3)
		

and evalVEMult ve ve' tEnv = 
	let l1 = match (ve,ve') with
			(***** 1 * e -> e *)
			(***** 0 * e -> e *)
			(VECst(v),_) ->	(try
								let v' = cast v Type_Float_64 in
		      					(match v' with
								FVal(FE((M(a),M(b)),(M(c),M(d)),p)) ->
									if (((FR.cmp a (FR.from_float_prec p GMP_RNDN 1.0))=0) &&
										((FR.cmp b (FR.from_float_prec p GMP_RNDN 1.0))=0) &&
										((FR.cmp c (FR.from_float_prec !errPrec GMP_RNDN 0.0))=0) &&
										((FR.cmp d (FR.from_float_prec !errPrec GMP_RNDN 0.0))=0)) then [ve'] 
									else
										if (((FR.cmp a (FR.from_float_prec p GMP_RNDN 0.0))=0) &&
											((FR.cmp b (FR.from_float_prec p GMP_RNDN 0.0))=0) &&
											((FR.cmp c (FR.from_float_prec !errPrec GMP_RNDN 0.0))=0) &&
											((FR.cmp d (FR.from_float_prec !errPrec GMP_RNDN 0.0))=0)) then [VECst(zeroPrec p)] 
										else
											[]
								| _ -> raise (Error ("evalVEMult","??? Cannot happen."))
								)
							with  Error(_) -> []
								| e -> raise e
							)
			| _ -> []
	in
	let l2 = match (ve,ve') with
		(VECst(c),VECst(c')) -> [VECst(valMult c c');VETimes(ve,ve')]
		(***** x^m * x^n = x^(m+n) *)
		| (VEId(v1,n1),VEId(v2,n2)) -> 
				if ((String.compare v1 v2)=0) then 
					[VEId(v1,n1+n2);VETimes(ve,ve');VETimes(ve,ve')] 
				else
					[]
		| _ -> []
	in 
		insertList l1 (insertList l2 [VETimes(ve,ve')])


and evalVEDotMult ve ve' tEnv = [VEDotTimes(ve,ve')]
and evalVEDotDiv ve ve' tEnv = [VEDotDiv(ve,ve')]

	
and evalVEDiv ve ve' tEnv = 
	let l1 = match (ve,ve') with
		(VECst(c1),VECst(c2)) ->	[VECst(valDiv c1 c2)]
	| 	_ -> 						[]
	in
	let l2 = match ve' with
				(***** e / 1 = e *)
				VECst(c) -> (try
								let c' = cast c Type_Float_64 
								in if (c'=(FVal(!fe64One))) then
										[ve]
							   	   else []
							with  Error(_) -> []
								| e -> raise e			
							)
			| _ -> []
	in
	let l3 = if (ve=ve') then 
				(* e / e = 1 *)
				[VECst(IVal(!i32One));VEDiv(ve,ve')]
		 	 else
				[VEDiv(ve,ve')]  
	in insertList l1 (insertList l2 l3)
	


and evalVEPower ve ve' tEnv = 
	let l1 = match (ve,ve') with
				(VECst(c1),VECst(IVal(c2))) -> 	[VECst(valPower c1 (iVal2PairOfInt (IVal(c2)) []))] 
				(* (x^a)^b *)
			|	(VEId(id,n),VECst(m)) -> 		let (m1,m2) = iVal2PairOfInt m []
												in if (m1=m2) then [VEId(id,m1*n)] else []
			|	_ -> []
	in
	let l2 = match ve' with
				VECst(m) ->		let (m1,m2) = iVal2PairOfInt m []
								in if ((m1=m2)&&(m1=0)) then 
										(* e^0 = 1 *)
										[VECst(IVal(!i32One))]
								   else
									if ((m1=m2)&&(m1=1)) then
										(* e^1 = e *)
										[ve]
									else []
				| _ ->			[]
	in insertList l1 (insertList l2 [VEPower(ve,ve')]) 


and evalVEUMinus ve tEnv = match ve with
  VECst(c) -> [VECst(valNeg c)] 
| VEUMinus(ve') -> [ve']  
| _ -> [VEUMinus(ve)] 



and evalVESqrt ve tEnv = match ve with
  VECst(c) -> [VECst(valSqrt c)] 
(* sqrt(x^2n)=x^n *)
| VEId(x,n) -> if ((n mod 2)=0) then [VEId(x,n/2)] else [VESqrt(ve)] 
| _ -> [VESqrt(ve)] 


and evalVE ve1 ve2 op tEnv = match op with
	OP_PLUS -> 		insertList (evalVEPlus ve1 ve2 tEnv) (evalVEPlus ve2 ve1 tEnv) 
| 	OP_MINUS -> 	(evalVEMinus ve1 ve2 tEnv)
| 	OP_MULT ->		insertList (evalVEMult ve1 ve2 tEnv) (evalVEMult ve2 ve1 tEnv)
| 	OP_MAT_MULT ->	evalVEMult ve1 ve2 tEnv
| 	OP_DOT_MULT ->	insertList (evalVEDotMult ve1 ve2 tEnv) (evalVEDotMult ve2 ve1 tEnv)
| 	OP_DOT_DIV ->	(evalVEDotDiv ve1 ve2 tEnv) 
|	OP_DIV -> 		evalVEDiv ve1 ve2 tEnv 
|	OP_POW ->		evalVEPower ve1 ve2 tEnv
|	OP_UMINUS ->	evalVEUMinus ve1 tEnv
;;


	  
let rec feWidthError x = match x with
	FVal(FE(_,e,_)) -> 		widthPrec e !errPrec 
|   XVal(FX(_,_,_,e)) -> 	widthPrec e !errPrec 
|	IVal(_) -> 				M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0)
|	BVal(_)	->				M(FR.from_float_prec float32_mantissa GMP_RNDN 0.0)
|	AVal(a) ->			 						Array.fold_left (fun x l -> (mpFloatMax x 
																			(Array.fold_left (fun y z -> (mpFloatMax (feWidthError z) y)) 
																						 	MBottom l))) 
												MBottom  a
| 	ValTop -> 				MPInfty
|	ValBottom -> 			MBottom 
|   ArrayLines(_) ->		raise (Error ("feWidthError","Cannot happen"))
|   ArrayColumns(_) ->		raise (Error ("feWidthError","Cannot happen"))
;;


let rec feWidthFloat x = match x with
	FVal(FE(f,e,p)) -> 			widthPrec f p 
|  	XVal(FX(f,ip,fp,e)) -> 		fixedWidthPrec f (ip+fp)
|	IVal(I32(i1),I32(i2)) -> 	let r = Int32.to_float (Int32.sub i2 i1) 
								in M(FR.from_float_prec float32_mantissa GMP_RNDN r)
|	IVal(I64(i1),I64(i2)) -> 	let r = Int64.to_float (Int64.sub i2 i1) 
								in M(FR.from_float_prec float32_mantissa GMP_RNDN r)
|	IVal(I64(i1),I32(i2)) -> 	let r = Int64.to_float (Int64.sub (Int64.of_int32 i2) i1) 
								in M(FR.from_float_prec float32_mantissa GMP_RNDN r)
|	IVal(I32(i1),I64(i2)) -> 	let r = Int64.to_float (Int64.sub i2 (Int64.of_int32 i1)) 
								in M(FR.from_float_prec float32_mantissa GMP_RNDN r)
|	IVal(_) ->					MPInfty
|	BVal(BTop) -> 				M(FR.from_float_prec float64_mantissa GMP_RNDN 1.0)
|	BVal(B(_)) -> 				M(FR.from_float_prec float64_mantissa GMP_RNDN 0.0)
|	BVal(BBottom) -> 			MBottom
|	AVal(a)-> 					Array.fold_left (fun x l -> (mpFloatMax x 
																		(Array.fold_left (fun y z -> (mpFloatMax (feWidthFloat z) y)) 
																						 MBottom l))) 
												MBottom  a
|	ValTop -> 					MPInfty
|	ValBottom -> 				MBottom
|   ArrayLines(_) ->			raise (Error ("feWidthError","Cannot happen"))
|   ArrayColumns(_) ->			raise (Error ("feWidthError","Cannot happen"))
;;

  
let rec array2List a i j res = 
  if (i=j) then a.(i)::res
  else array2List a (i+1) j (a.(i)::res) ;;
  

	
let evalDim a =
let m = Array.length a in
let n = Array.length a.(0) in
let b = Array.make_matrix 1 2 (IVal(I32(Int32.of_int n),I32(Int32.of_int n))) in
let _ = b.(0).(0) <- (IVal(I32(Int32.of_int m),I32(Int32.of_int m)))
in AVal(b) ;;
	
	
let rec valArraySize v = match v with 
AVal(a) -> 	evalDim a
|   _ -> raise (Error ("valArraySize","??? 'size' expects a matrix as argument."))
;;


(*****************************************************************************)
(*    Eval of the roundoff error to be added to the transformed expression   *)
(*****************************************************************************)


let rec evalEpsVal v = match v with
    FVal(FE(_,e,_)) -> e
  | XVal(FX(_,_,_,e)) -> e
  | AVal(a) ->  let b = ref (evalEpsVal a.(0).(0)) in
				let _ = 
						for i=0 to ((Array.length a)-1) do
							for j=0 to ((Array.length a.(i))-1) do
								let x = evalEpsVal a.(i).(j) in
								if (errLt (!b) x)  then
									b := x 
							done	
						done
				in (!b) 	
  | _ -> !eZero ;;

