(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file slice.ml contains functions for the numerical evaluation of expressions
 *)

open Gmp ;;
open Types ;;
open Prelude ;;
open Integer ;;
open Float ;;
open Fixed ;;
open Coercion ;;
open Matrix ;;

(*****************************************************************************)
(*                                    MISC                                   *)
(*****************************************************************************)
  
let widthPrec (a,b) prec = match (a,b) with
  (M(x),M(y)) -> M(FR.sub_prec prec GMP_RNDU y x) 
| (M(_),MPInfty) -> MPInfty
| (MPInfty,M(_)) ->  MPInfty
| (M(_),MMInfty) -> MPInfty
| (MMInfty,M(_)) -> MPInfty
| (MPInfty,MPInfty) -> MPInfty
| (MMInfty,MMInfty) -> MPInfty
| (MMInfty,MPInfty) -> MPInfty
| (MPInfty,MMInfty) -> MPInfty
| (MBottom,_) -> MBottom
| (_,MBottom) -> MBottom ;;


let fixedWidthPrec (a,b) prec = match (a,b) with
  (X(_),X(_)) -> 	let (u,v) =  (mpFloatOfFixed a prec,mpFloatOfFixed b prec) 
					in (match (u,v) with
							(M(x),M(y)) -> M(FR.sub_prec prec GMP_RNDU y x)  
						| _ -> raise (Error("fixedWidthPrec","??? impossible result of subtraction."))
						)
| (X(_),XPInfty) -> MPInfty
| (XPInfty,X(_)) ->  MPInfty
| (X(_),XMInfty) -> MPInfty
| (XMInfty,X(_)) -> MPInfty
| (XPInfty,XPInfty) -> MPInfty
| (XMInfty,XMInfty) -> MPInfty
| (XMInfty,XPInfty) -> MPInfty
| (XPInfty,XMInfty) -> MPInfty
| (XBottom,_) -> MBottom
| (_,XBottom) -> MBottom ;;

  
let rec insertList l1 l2 = match l1 with
	[] -> 		l2
|	x::xs ->	if (List.mem x l2) then
					insertList xs l2
				else 
					insertList xs (x::l2) ;;


let rec typeOfVE ve tEnv = match ve with
	VECst(IVal(I64(_),_)) -> Type_Int_64
|	VECst(IVal(_,I64(_))) -> Type_Int_64	
|	VECst(IVal(_)) -> Type_Int_32
|   VECst(FVal(FE(_,_,p))) -> typeOfPrec p 
|   VECst(XVal(FX((_,_),ip,fp,_))) -> Type_Fixed(ip,fp)
|   VECst(BVal(_)) -> Type_Bool
|   VECst(AVal(a)) -> Type_Array(I32(Int32.of_int (Array.length a)),I32(Int32.of_int (Array.length a.(0))),typeOfVE (VECst(a.(0).(0))) tEnv)
|   VECst(_) -> UndefinedType  
| 	VECast(t,ve1) -> t
| 	VEArrayElt(id,_,_) -> getArrayType (getEnv id tEnv)
| 	VEId(id,_) -> getEnv id tEnv
| 	VEPlus(ve1,ve2) -> maxType (typeOfVE ve1 tEnv) (typeOfVE ve2 tEnv)
| 	VEMinus(ve1,ve2) -> maxType (typeOfVE ve1 tEnv) (typeOfVE ve2 tEnv)
| 	VETimes(ve1,ve2) -> maxType (typeOfVE ve1 tEnv) (typeOfVE ve2 tEnv)
| 	VEDotTimes(ve1,ve2) ->  maxType (typeOfVE ve1 tEnv) (typeOfVE ve2 tEnv)
| 	VEDotDiv(ve1,ve2) ->  maxType (typeOfVE ve1 tEnv) (typeOfVE ve2 tEnv)
| 	VEDiv(ve1,ve2) -> maxType (typeOfVE ve1 tEnv) (typeOfVE ve2 tEnv) 
| 	VEPower(ve1,ve2) -> typeOfVE ve1 tEnv
| 	VEUMinus(ve1) -> typeOfVE ve1 tEnv
| 	VESqrt(ve1) -> typeOfVE ve1 tEnv
| 	VEAbs(ve1) -> typeOfVE ve1 tEnv
| 	VEMin(ve1) -> getArrayType (typeOfVE ve1 tEnv)
| 	VEMax(ve1) -> getArrayType (typeOfVE ve1 tEnv)
| 	VESize(ve1) -> Type_Array(I32(Int32.one),I32(Int32.one),Type_Int_32)
|   _ -> UndefinedType 
;;


(*****************************************************************************)
(*                                  SLICE                                    *)
(*****************************************************************************)


let rec iVal2PairOfInt i env = match i with
	IVal((I32(a),I32(b))) ->	(Int32.to_int a,Int32.to_int b)
|	IVal((I64(a),I64(b))) -> 	(Int64.to_int a,Int64.to_int b)
|	IVal((I32(a),I64(b))) -> 	(Int32.to_int a,Int64.to_int b)
|	IVal((I64(a),I32(b))) -> 	(Int64.to_int a,Int32.to_int b)
|	AVal(_) -> raise (Error ("iVal2PairOfInt","??? Subscript must be integer (array).")) 
|	FVal(_) -> raise (Error ("iVal2PairOfInt","??? Subscript must be integer (float).")) 
|	XVal(_) -> raise (Error ("iVal2PairOfInt","??? Subscript must be integer (fixed).")) 
|   ValBottom -> raise (Error ("iVal2PairOfInt","??? Subscript must be integer (bottom).")) 
|   BVal(_) -> raise (Error ("iVal2PairOfInt","??? Subscript must be integer (bool).")) 
|   ValTop -> raise (Error ("iVal2PairOfInt","??? Subscript must be integer (top).")) 
|   ArrayColumns(id) -> let a = getArray (getEnv id env) in
						if (((Array.length a)=1)&((Array.length a.(0))=1)) then 
							iVal2PairOfInt (a.(0).(0)) env
						else
							raise (Error ("iVal2PairOfInt","??? Subscript must be integer (AC).")) 
|   ArrayLines(id) -> 	let a = getArray (getEnv id env) in
						if (((Array.length a)=1)&((Array.length a.(0))=1)) then 
							iVal2PairOfInt (a.(0).(0)) env
						else
							raise (Error ("iVal2PairOfInt","??? Subscript must be integer (AL).")) 
|	_ -> 	raise (Error ("iVal2PairOfInt","??? Subscript must be integer.")) 
 ;;


let rec joinTabElts t i j res = 
	if (i=j) then 
		valJoin t.(i) res 
	else  
		joinTabElts t (i+1) j (valJoin t.(i) res) ;;


let rec joinTabElts2D t a b c d res =
	try
		if (a=b) then
			joinTabElts t.(a) c d res
		else
			joinTabElts2D t (a+1) b c d (joinTabElts t.(a) c d res) 
	with
		Invalid_argument(_) -> raise (Error ("joinTabElts2D","??? Index exceeds matrix dimensions.")) ;;
		

let rec joinTabLinesAux t a b tRes = 
	if (a>b) then
		tRes
	else
		let _ = for i=0 to ((Array.length t.(a))-1) do
					tRes.(0).(i) <- valJoin (tRes.(0).(i)) (t.(a).(i)) 
				done 
		in joinTabLinesAux t (a+1) b tRes ;;
		

let joinTabLines t a b =
	try 
		let m = Array.make_matrix 1 (Array.length t.(a)) ValBottom 
		in joinTabLinesAux t a b m 
	with
		Invalid_argument(_) -> raise (Error ("joinTabLines","??? Index exceeds matrix dimensions.")) ;;

								
(*****************************************************************************)
(*                            EVAL NUMERICAL                                 *)
(*****************************************************************************)

let rec getVEVarList ve res = match ve with (* voir gervarlist ds analyze.ml pour completer fct *)
	VECst(_) ->				res
|	VECast(_,ve1) ->		getVEVarList ve1 res
|	VEId(v,_) ->			if (List.mem v res) then res else v::res
|	VEArrayElt(_) ->		res 										(* for now, no slicing on array elmts *)
|	VECreateMatrix(_) ->	res											(* idem *)
|	VEArrayBlock(_) -> 		res											(* idem *)
|	VEPlus(ve1,ve2) ->		getVEVarList ve2 (getVEVarList ve1 res)
|	VEMinus(ve1,ve2) ->		getVEVarList ve2 (getVEVarList ve1 res)
|	VETimes(ve1,ve2) ->		getVEVarList ve2 (getVEVarList ve1 res) 
|	VEDotTimes(ve1,ve2) ->	getVEVarList ve2 (getVEVarList ve1 res) 
|	VEDotDiv(ve1,ve2) ->	getVEVarList ve2 (getVEVarList ve1 res) 
|	VEDiv(ve1,ve2) ->		getVEVarList ve2 (getVEVarList ve1 res) 
|	VEPower(ve1,_) ->		getVEVarList ve1 res						(* no slicing of the exponent *) 
|	VEUMinus(ve1) ->		getVEVarList ve1 res 
|	VESqrt(ve1) ->			getVEVarList ve1 res
| 	VEAbs(ve1) ->			getVEVarList ve1 res  
| 	VEMin(ve1) ->			getVEVarList ve1 res  
| 	VEMax(ve1) ->			getVEVarList ve1 res  
| 	VESize(ve1) ->			getVEVarList ve1 res  
|	VEFunCall(f,vel) ->		(match vel with
								[] ->	res
							|	ve'::ves ->	let res' = getVEVarList ve' res 
											in getVEVarList (VEFunCall(f,ves)) res'
							)
|	VEDeltaFun(f,vel1,vel2) ->	
							let vel = vel1@vel2 in
							(match vel with
								[] ->	res
							|	ve'::ves ->	let res' = getVEVarList ve' res 
											in getVEVarList (VEFunCall(f,ves)) res'
							)
							;;


let evalVESpecialMatrix vl env matType = match matType with
	Matrix_Eye ->			let (_,i) = iVal2PairOfInt (List.hd vl) env
							in matrixEye i
|	Matrix_Ones -> 			let (_,i) = iVal2PairOfInt (List.hd vl) env in
							let (_,j) = iVal2PairOfInt (List.hd (List.tl vl)) env
							in matrixOnes i j
|	Matrix_Zeros ->			let (_,i) = iVal2PairOfInt (List.hd vl) env in
							let (_,j) = iVal2PairOfInt (List.hd (List.tl vl)) env  
							in matrixZeros i j
| 	Matrix_Diag ->			let m = getArray (List.hd vl) in
							let (_,i) = iVal2PairOfInt (List.hd (List.tl vl)) env
							in matrixDiag m i
| 	Matrix_Transpose ->		let m = getArray (List.hd vl) 
							in matrixTranspose m
|	Matrix_Rand ->			let (_,i) = iVal2PairOfInt (List.hd vl) env in
							let (_,j) = iVal2PairOfInt (List.hd (List.tl vl)) env
							in matrixRandom i j
|	Matrix_Rand_Errors ->	let (_,i) = iVal2PairOfInt (List.hd vl) env in
							let (_,j) = iVal2PairOfInt (List.hd (List.tl vl)) env 
							in matrixRandomErrors i j
|	Matrix_Linspace ->		let a = List.hd vl in
							let b = List.hd (List.tl vl) in
							let n = List.hd (List.tl (List.tl vl)) in
							let (_,i) = iVal2PairOfInt n env
							in matrixLinspace a b i
|	Matrix_Logspace ->		let a = List.hd vl in
							let b = List.hd (List.tl vl) in
							let n = List.hd (List.tl (List.tl vl)) in
							let (_,i) = iVal2PairOfInt n env
							in matrixLogspace a b i
|	Matrix_Colon_Step ->	let a = List.hd vl in
							let i = List.hd (List.tl vl) in
							let b = List.hd (List.tl (List.tl vl)) 
							in matrixColonStep a b i
;;


(*****************************************************************************)
(*                            SLICING OF VALUES                              *)
(*****************************************************************************)					

					   
let splitFloatPrec x = match x with
	FE((M(a),M(b)),e,p) ->
			let midPoint = FR.mul_prec  p GMP_RNDN (FR.add_prec p GMP_RNDN a b) (FR.from_float_prec p GMP_RNDN 0.5) in
			let r1 = if ((FR.cmp midPoint b)<0) then 
						FVal(FE((M(a),M(midPoint)),e,p))  
					 else
					 	FVal(x)  
			in
			let r2 = if ((FR.cmp a midPoint)<0) then 
						FVal(FE((M(midPoint),M(b)),e,p))  
					 else
					 	FVal(x)  
			in (r1,r2)
| 	_ -> (FVal(x),FVal(x)) ;;

					   
let splitFixedPrec x = match x with
	FX((a,b),ip,fp,e) ->
			let (midPoint,_) = fixedMult (fst (fixedPlus a b ip fp)) (X(true,"0","1"^(String.make (fp-1) '0'))) ip fp in
			let r1 = if (fixedLt midPoint b) then 
						XVal(FX((a,midPoint),ip,fp,e))  
					 else
					 	XVal(x)  
			in
			let r2 = if (fixedLt a midPoint) then 
						XVal(FX((midPoint,b),ip,fp,e))  
					 else
					 	XVal(x)  
			in (r1,r2) ;;


let splitErrPrec x = match x with
	FE(f,(M(a),M(b)),p) ->
			let midPoint = FR.mul_prec !errPrec GMP_RNDN (FR.add_prec !errPrec GMP_RNDN a b) (FR.from_float_prec !errPrec GMP_RNDN 0.5) in
			let r1 = if ((FR.cmp midPoint b)<0) then 
						FVal(FE(f,(M(a),M(midPoint)),p))  
					 else
					 	FVal(x)  
			in
			let r2 = if ((FR.cmp a midPoint)<0) then 
						FVal(FE(f,(M(midPoint),M(b)),p))  
					 else
					 	FVal(x)  
			in (r1,r2)
| 	_ -> (FVal(x),FVal(x)) ;;


let splitXErrPrec x = match x with
	FX(fx,ip,fp,(M(a),M(b))) ->
			let midPoint = FR.mul_prec !errPrec GMP_RNDN (FR.add_prec !errPrec GMP_RNDN a b) (FR.from_float_prec !errPrec GMP_RNDN 0.5) in
			let r1 = if ((FR.cmp midPoint b)<0) then 
						XVal(FX(fx,ip,fp,(M(a),M(midPoint))))  
					 else
					 	XVal(x)  
			in
			let r2 = if ((FR.cmp a midPoint)<0) then 
						XVal(FX(fx,ip,fp,(M(midPoint),M(b))))  
					 else
					 	XVal(x)  
			in (r1,r2)
| 	_ -> (XVal(x),XVal(x)) ;;


let splitInt x = match x with
	(I32(a),I32(b)) ->	(IVal(I32(a),I32(Int32.div (Int32.add a b) (Int32.of_int 2))),
						 IVal(I32(Int32.div (Int32.add a b) (Int32.of_int 2)),I32(b)))
|	(I64(a),I64(b)) ->	(IVal(I64(a),I64(Int64.div (Int64.add a b) (Int64.of_int 2))),
						 IVal(I64(Int64.div (Int64.add a b) (Int64.of_int 2)),I64(b)))
| 	_ -> (IVal(x),IVal(x))
;;


let splitVal v = match v with
	FVal(fe) ->		splitFloatPrec fe
|   XVal(x) ->      splitFixedPrec x 
|	IVal(iVal) ->	splitInt iVal 
| _ ->				(v,v) ;;


let splitErr v = match v with
	FVal(fe) ->		splitErrPrec fe 
|	XVal(x) ->		splitXErrPrec x
|	IVal(iVal) ->	splitInt iVal 
| _ ->				(v,v) ;;
