(*
 * RangeLab 1.0 (a static analyzer for numerical accuracy)
 * Copyright (C) 2011 Matthieu Martel
 * This file is part of RangeLab 1.0.
 *
 *  RangeLab is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  RangeLab is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RangeLab. If not, see <http://www.gnu.org/licenses/>.
 *)

(*
 * The file types.ml contains all the type definitions
 *)


open Gmp ;;
open Int32 ;;
open Int64 ;;


(* Labels attached to the nodes of the syntactic tree *)
type label = Lab of int ;;

(* Rouding mode for interval operations *)
type intervalOpMode =
  NEAREST
| OUTSIDE ;;

(* Domain of Booleans *)
type boolValue =
  B of bool
| BTop
| BBottom ;;

(* Domain of floating-point numbers *)
type mpNumber =
  M of FR.t				(* GMP number *)
| MPInfty				(* Plus infinity *)
| MMInfty				(* Minus infinity *)
| MBottom 				(* Bottom *)
;;

(* Domain of integers *)
type intNumber =
  I32 of int32			(* 32 bits integer *)
| I64 of int64			(* 64 bits integer *)
| IPInfty				(* Plus infinity *)
| IMInfty				(* Minus infinity *)
| IBottom 				(* Bottom *)
;;


type intInterval = intNumber * intNumber ;;  

type mpInterval = mpNumber * mpNumber ;;

type floatWithError = FE of mpInterval * mpInterval * int ;; (* the last int indicates the precision *)


type fixedNumber =
 	X of  bool * string * string  (* sign * integer part * fractional part as a binary string * integer part as a binary string *)
|   XPInfty 
|   XMInfty
|   XBottom ;;


type fixedInterval = fixedNumber * fixedNumber ;;

type fixedWithError = FX of fixedInterval * int * int * mpInterval ;;


(* Domain of values *)
type value = 
	IVal of intInterval
|	FVal of floatWithError
|   XVal of fixedWithError 
| 	BVal of boolValue
|	AVal of value array array
|   ValBottom
|   ValTop 
|   ArrayLines of string		(* used to compute dynamically the size of an array stored in a variable *)
|   ArrayColumns of string 
;;

(* Types *)
type typeDef = 
	Type_Int_32 
| 	Type_Int_64
|	Type_Float_32
|	Type_Float_64
|	Type_Float_128
|   Type_MP of int 									(* number of digits of the mantissa of the multiple precision number*)
|   Type_Fixed of int * int							(* number of digits of the integer and fractional parts *) 
| 	Type_Bool
| 	Type_Array of intNumber * intNumber * typeDef 	(* size * type *) 
|	Type_Fun of typeDef								(* return type *)
| 	UndefinedType 
;;
	
	
type relation = 
	Relation_EQ
| 	Relation_NEQ
|	Relation_LT
|	Relation_LTE
|	Relation_GT
|	Relation_GTE ;;


type specialMatrix =
	Matrix_Eye
|	Matrix_Ones
|	Matrix_Zeros 
|	Matrix_Rand
|	Matrix_Rand_Errors 
|	Matrix_Diag
|	Matrix_Transpose
|	Matrix_Linspace 
|	Matrix_Logspace 
|	Matrix_Colon_Step ;;


(* Arithmetic Expressions *)
type aExpression = 
	Cst of value * label											(* Constants *)
|	Plus of aExpression * aExpression * label						(* Addition *)
|	Minus of aExpression * aExpression * label						(* Subtraction *)
|	Times of aExpression * aExpression * label						(* Product *)
|   DotTimes of aExpression * aExpression * label					(* The .* op: element by element product of matrices *)
|   DotDiv of aExpression * aExpression * label			  			(* The ./ op: element by element division of matrices *)
|	Div of aExpression * aExpression * label						(* Division *)	
|	UMinus of aExpression * label 									(* Unary minus *)
|	Power of aExpression * aExpression * label						(* ae1^ae2 *)
|	Sqrt of aExpression * label										(* Square root *)
|	Abs of aExpression * label										(* Absolute value *)
|   Min of aExpression * label										(* Minimal element of a vector *) 
|   Max of aExpression * label										(* Maximal element of a vector *) 
|   Size of aExpression * label										(* Size of an array, ex: [m,n] *) 
| 	Cast of typeDef * aExpression * label							(* Cast expr to type *)
|	Id of string * label											(* Identifiers *)
|	ArrayExpr of aExpression array array * label					(* Array definition *)
|   ArrayElt of string * aExpression * aExpression * label			(* Array element *)
|   CreateMatrix of aExpression list * specialMatrix * label		(* Special Matrices *)
|	ArrayBlock of string * aExpression * aExpression * 
							aExpression * aExpression * label		(* Block of an array *)
|	FunCall of string * (aExpression list) * label					(* Call to a function *)
|	UndefinedCall of string * (aExpression list) * label			(* Call to function of array indexing *)
;;


(* Boolean Expressions *)
type bExpression =
	Eq of aExpression * aExpression * label			(* Equal *)
|	Neq of aExpression * aExpression * label		(* Not equal *)
|	Lt of aExpression * aExpression * label			(* Less than *)
|	Lte of aExpression * aExpression * label		(* Less than or equal *)
|	Gt of aExpression * aExpression * label			(* Greater than *)
|	Gte of aExpression * aExpression * label		(* Greater than or equal *)
|	And of bExpression * bExpression * label		(* Logical and *)
|	Or of bExpression * bExpression * label			(* Logical or *)
|	Xor of bExpression * bExpression * label		(* Logical xor *)
|	Not of bExpression * label 						(* Logical negation *)
;;


(* Commands *) 
type command = 
|	Nop																			(* No commment: no command *)
|	VarDef of string * typeDef * label  										(* Variable definition *)
|	Assign of string * aExpression * label										(* Assignments *)
|   AssignBool of string * bExpression * label                                  (* Assignments for boolean expressions *)
|	AssignArrayElt of string * aExpression * aExpression * aExpression * label	(* Assignments : ID[expr1]:=expr2 *)
|	AssignArrayBlock of string * aExpression * aExpression * aExpression * aExpression * aExpression * label (* Assignments, ex: A(i,:)=x *)
| 	Seq of command * command * label											(* Sequence *)
|	CondITE of bExpression * command * command * label							(* If then else *)
|	While of bExpression * command * label										(* While loop *)
|	For of string * aExpression * aExpression * aExpression * command * label 	(* For loop: for(id,min,max,step,cmd,lab) *)
;;


type fct = Fun of (string list) * (string list) * command * label		(* declaration of function : arguments * outputs * command *) 


(* Formal expressions used as values *)
type valexpr = 
	VECst of value
| 	VECast of typeDef * valexpr
| 	VEId of string * int
| 	VEArrayElt of string * valexpr * valexpr 
|	VECreateMatrix of (valexpr list) * specialMatrix 
|  	VEArrayBlock of string * valexpr * valexpr * valexpr * valexpr 
|	VEFunCall of string * (valexpr list) 
|	VEDeltaFun of string * (valexpr list) * (valexpr list) 
| 	VEPlus of valexpr * valexpr
| 	VEMinus of valexpr * valexpr
| 	VETimes of valexpr * valexpr 
| 	VEDotTimes of valexpr * valexpr 
| 	VEDotDiv of valexpr * valexpr 
| 	VEDiv of valexpr * valexpr
| 	VEPower of valexpr * valexpr
| 	VEUMinus of valexpr
| 	VESqrt of valexpr
|  	VEAbs of valexpr 
| 	VEMin of valexpr
|   VEMax of valexpr 
|   VESize of valexpr 
;;


type operator = OP_PLUS | OP_MINUS | OP_MULT | OP_DOT_MULT | OP_MAT_MULT | OP_DIV | OP_UMINUS | OP_POW | OP_DOT_DIV ;;


